
/*
 * Computer Algebra Kit (c) 1998 by David Stes.  All Rights Reserved.
 */

#include "cakit.h"

static int ca_vardnsmonom_size(ca_vardnsmonom_args args)
{
    return args.nvars; /* size to add to int pointer */
}

static int ca_vardnsmonom_bsize(ca_vardnsmonom_args args)
{
    return args.nvars * sizeof(int); /* byte size exponent array */
}

static int ca_vardnsmonom_check(ca_vardnsmonom_t self,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) assert(*self++ >= 0);
    return 1;
}

static ca_vardnsmonom_t ca_vardnsmonom_id2t(id object)
{
    return [object ca_vardnsmonom_value];
}

static id ca_vardnsmonom_t2id(id proto,ca_vardnsmonom_t f)
{
    return [proto ca_vardnsmonom_value:f];
}

static ca_vardnsmonom_args ca_vardnsmonom_getargs(id self)
{
    ca_vardnsmonom_args res;
    res.nvars = [[self symbols] size];
    return res; 
}

static ca_vardnsmonom_t ca_vardnsmonom_alloc(ca_vardnsmonom_args args)
{
    return (ca_vardnsmonom_t)OC_Malloc(ca_vardnsmonom_bsize(args));
}

static void ca_vardnsmonom_init(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;while (nvars--) *c++ = 0;
}

static void ca_vardnsmonom_copy(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;while (nvars--) *c++ = *a++;
}

static void ca_vardnsmonom_clear(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    /* nothing to clear... */
}

static void ca_vardnsmonom_free(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    OC_Free(c);
}

static void ca_vardnsmonom_args_fileOut(id aFiler,ca_vardnsmonom_args *args)
{
    [aFiler fileOut:&args->nvars type:'i'];
}

static void ca_vardnsmonom_args_fileIn(id aFiler,ca_vardnsmonom_args *args)
{
    [aFiler fileIn:&args->nvars type:'i'];
}

static void ca_vardnsmonom_fileOut(id aFiler,ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int n = args.nvars;
    while (n--) [aFiler fileOut:c++ type:'i'];
}

static void ca_vardnsmonom_fileIn(id aFiler,ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int n = args.nvars;
    while (n--) [aFiler fileIn:c++ type:'i'];
}

static unsigned ca_vardnsmonom_hash(ca_vardnsmonom_t self,ca_vardnsmonom_args args)
{
    assert(args.nvars>0);return self[0];
}

static int ca_vardnsmonom_deg(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int sum = 0;
    int nvars = args.nvars;
    while (nvars--) sum += *self++;
    return sum;
}

static int ca_vardnsmonom_numterms(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int numterms = 0;
    int nvars = args.nvars;
    while (nvars--) if (*self++) numterms++;
    return numterms;
}

static int ca_vardnsmonom_isone(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if (*self) return 0; self++; }
    return 1;
}

static int ca_vardnsmonom_iseq(ca_vardnsmonom_t a,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if ((*a) != (*b)) return 0; a++;b++; }
    return 1;
}

static int ca_vardnsmonom_iseqmul(ca_vardnsmonom_t a,ca_vardnsmonom_t x,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if ((*a) != (*x) + (*b)) return 0; a++;x++;b++; }
    return 1;
}

static int ca_vardnsmonom_cmp(ca_vardnsmonom_t a,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;

    /* lexicographic cmp *MUST* work from right to left
     * because last symbol is largest for vardns order
     */

    while (nvars--) {
	if ((*a) < (*b)) return -1;
	if ((*a) > (*b)) return +1;
	a++;b++;
    }
    
    return 0;
}

static int ca_vardnsmonom_cmpmul(ca_vardnsmonom_t a,ca_vardnsmonom_t x,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;

    /* lexicographic cmp *MUST* work from right to left
     * because last symbol is largest for vardns order
     */

    while (nvars--) {
	int sum = *x + *b;
	if ((*a) < sum) return -1;
	if ((*a) > sum) return +1;
	a++;x++;b++;
    }
    
    return 0;
}

static void ca_vardnsmonom_oomul(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { (*c) = (*a) + (*b); c++; a++; b++; }
}

static void ca_vardnsmonom_inmul(ca_vardnsmonom_c *c,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { (*c) += (*b); c++; b++; }
}

static BOOL ca_vardnsmonom_oodiv(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) {
    	int q = *a - *b;
	if (q<0) return FAILURE;
	*c = q; c++; a++; b++;
    }
    
    return SUCCESS;
}

static void ca_vardnsmonom_indiv(ca_vardnsmonom_c *c,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) {
    	int q = *c - *b;
	if (q<0) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
	*c = q; c++; b++;
    }
}
static ca_vardns_args ca_vardns_getargs(id self)
{
    ca_vardns_args res;
    res.nvars = [[self symbols] size];
    return res;
}

static ca_vardns_t ca_vardns_id2t(id object)
{
    return [object ca_vardns_value];
}

static ca_vardns_c *ca_vardns_id2c(id object)
{
    return [object ca_vardns_reference];
}

static id ca_vardns_t2id(id proto,ca_vardns_t f)
{
    return [proto ca_vardns_value:f];
}

static id ca_vardns_c2id(id proto,ca_vardns_c *r)
{
    return [proto ca_vardns_reference:r];
}

static int ca_vardns_check(ca_vardns_t self,ca_vardns_args args)
{
    assert(0 < args.nvars);
    return ca_vardnsmonom_check(self,args);
}

static void ca_vardns_init(ca_vardns_c *self,ca_vardns_args args)
{
    *self = ca_vardnsmonom_alloc(args);
    ca_vardnsmonom_init(*self,args);
}

static void ca_vardns_copy(ca_vardns_c *c,ca_vardns_t a,ca_vardns_args args)
{
    *c = ca_vardnsmonom_alloc(args);
    ca_vardnsmonom_copy(*c,a,args);
}

static void ca_vardns_move(ca_vardns_c *c,ca_vardns_c *a,ca_vardns_args args)
{
    *c = *a;*a = NULL;
}

static void ca_vardns_copymonom(ca_vardns_c *c,ca_vardnsmonom_t a,ca_vardns_args args)
{
    *c = ca_vardnsmonom_alloc(args);
    ca_vardnsmonom_copy(*c,a,args);
}

static void ca_vardns_clear(ca_vardns_c *self,ca_vardns_args args)
{
    ca_vardnsmonom_free(*self,args);*self = NULL;
}

static int ca_vardns_deg(ca_vardns_t self,ca_vardns_args args)
{
    return ca_vardnsmonom_deg(self,args);
}

static int ca_vardns_hash(ca_vardns_t self,ca_vardns_args args)
{
    return ca_vardnsmonom_hash(self,args);
}

static int ca_vardns_numterms(ca_vardns_t self,ca_vardns_args args)
{
    return ca_vardnsmonom_numterms(self,args);
}

static int ca_vardns_iseq(ca_vardns_t a,ca_vardns_t b,ca_vardns_args args)
{
    return ca_vardnsmonom_iseq(a,b,args);
}

static int ca_vardns_cmp(ca_vardns_t a,ca_vardns_t b,ca_vardns_args args)
{
    return ca_vardnsmonom_cmp(a,b,args);
}

static int ca_vardns_findfirst(ca_vardns_t self,ca_vardns_args args)
{
    return -1;
}

static int ca_vardns_expat(ca_vardns_t self,int i,ca_vardns_args args)
{
    assert(0 <= i && i < args.nvars);return self[i];
}

static void ca_vardns_setexpat(ca_vardns_t self,int i,int e,ca_vardns_args args)
{
    assert(0 <= i && i < args.nvars);self[i] = e;
}


static void ca_vardns_args_fileOut(id aFiler,ca_vardns_args *args)
{
    ca_vardnsmonom_args_fileOut(aFiler,args);
}

static void ca_vardns_args_fileIn(id aFiler,ca_vardns_args *args) 
{
    ca_vardnsmonom_args_fileIn(aFiler,args);
}

static void ca_vardns_fileOut(id aFiler,ca_vardns_c *self ,ca_vardns_args args)
{
    ca_vardnsmonom_fileOut(aFiler,*self,args);
}

static void ca_vardns_fileIn(id aFiler,ca_vardns_c *self,ca_vardns_args args)
{
    *self = ca_vardnsmonom_alloc(args);
    ca_vardnsmonom_fileIn(aFiler,*self,args);
}

static void ca_vardns_ix_init(ca_vardns_ixt ix,id self)
{
    ix->n = [[self symbols] size];ix->i = -1;
}

static void ca_vardns_ix_clear(ca_vardns_ixt ix)
{
    /* nothing to clear */
}

static int ca_vardns_ix_count(ca_vardns_ixt ix)
{
    return ix->n; /* total number, not just nonzeroes */ 
}

static int ca_vardns_ix_isempty(ca_vardns_ixt ix)
{
    return ix->n == 0; 
}

static int ca_vardns_ix_ismatch(ca_vardns_ixt ix)
{
    return (0 <= ix->i && ix->i < ix->n);
}

static void ca_vardns_ix_tofirst(ca_vardns_ixt ix)
{
    ix->i = -1;
}

static void ca_vardns_ix_tolast(ca_vardns_ixt ix)
{
    ix->i = ix->n;
}

static void ca_vardns_ix_to(ca_vardns_ixt ix,int i)
{
    ix->i = i-1;
}

static void ca_vardns_ix_next(ca_vardns_ixt ix)
{
    if (ix->i == ix->n) {
	/* remain on position just after last term */
    } else {
	ix->i = ix->i + 1;
    }
}

static void ca_vardns_ix_prev(ca_vardns_ixt ix)
{
    if (ix->i == -1) {
	/* remain on position just before first term */
    } else {
	ix->i = ix->i - 1;
    }
}

@implementation ca_vardns_monomial
- check
{
    [super check];
    assert(termOne != nil && symbols != nil);
    [termOne check];[symbols check];
    ca_vardnsmonom_check(value,ca_vardns_getargs(self));return self;
}

- _setUpScalarOne:aOne symbols:aCltn
{
    assert([aOne isOne] && [aCltn size] > 0);
    termOne = [Term scalar:aOne symbol:nil exponent:0];
    symbols = aCltn;
    ca_vardns_init(&value,ca_vardns_getargs(self));
    assert([self check]);return self;
}

+ scalarOne:aOne symbols:aCltn
{
    return [[super new] _setUpScalarOne:aOne symbols:aCltn];
}

- copy
{
    ca_vardns_c c;assert([self check]);
    ca_vardns_copy(&c,value,ca_vardns_getargs(self));
    return ca_vardns_c2id(self,&c);
}

- deepCopy
{
    ca_vardns_c c;assert([self check]);
    ca_vardns_copy(&c,value,ca_vardns_getargs(self));
    return ca_vardns_c2id(self,&c);
}

- clear
{
    termOne = [termOne free];symbols = [symbols free];
    ca_vardns_clear(&value,ca_vardns_getargs(self));
    return [super clear];
}

- clone
{
    self = [super clone];
    termOne = [termOne cheapCopy];symbols = [symbols cheapCopy];
    return self;
}

- empty
{
    ca_vardns_c c;assert([self check]);
    ca_vardns_init(&c,ca_vardns_getargs(self));
    return ca_vardns_c2id(self,&c);
}

- (ca_vardns_t) ca_vardns_value
{
    return value;
}

- ca_vardns_value:(ca_vardns_t)aValue
{
    self = [self clone];
    ca_vardns_copy(&value,aValue,ca_vardns_getargs(self));
    assert([self check]);return self;
}

- (ca_vardns_c *) ca_vardns_reference
{
    return &value;
}

- ca_vardns_reference:(ca_vardns_c *)aReference
{
    self = [self clone];
    ca_vardns_move(&value,aReference,ca_vardns_getargs(self));
    assert([self check]);return self;
}

- (ca_vardnsmonom_t) ca_vardnsmonom_value
{
    return value;
}

- ca_vardnsmonom_value:(ca_vardnsmonom_t)aValue
{
    self = [self clone];
    ca_vardns_copymonom(&value,aValue,ca_vardns_getargs(self));
    assert([self check]);return self;
}

- (ca_vardnsmonom_c *) ca_vardnsmonom_reference
{
    return value;
}

- ca_vardnsmonom_reference:(ca_vardnsmonom_c *)aReference
{
    return [self error:""];
}

- termOne { return termOne; }
- symbols { return symbols; }
- (BOOL)isVariableSparse { return NO;  }
- (BOOL)isVariableDense  { return YES; }

- (BOOL) sameClass:b
{
    return [super sameClass:b] && [symbols isEqual:[b symbols]];
}

- (int) numTerms
{
    return ca_vardns_numterms(value,ca_vardns_getargs(self));
}

- (BOOL) isEmpty
{
    return [self numTerms] == 0;
}

- (BOOL) isOne
{
    return [self numTerms] == 0;
}

- (int) compare:b
{
    return ca_vardns_cmp(value,[b ca_vardns_value],ca_vardns_getargs(self));
}

- (BOOL) isEqual:b
{
    return ca_vardns_iseq(value,[b ca_vardns_value],ca_vardns_getargs(self));
}

- (unsigned) hash
{
    return ca_vardns_hash(value,ca_vardns_getargs(self));
}

- eachTerm
{
    id aSequence = [ca_vardns_sequence content:[self cheapCopy]];
    return [CASequence over:aSequence];
}

- elt_ca_vardns_ix:(ca_vardns_ixt)ix
{
    if (ca_vardns_ix_ismatch(ix)) {
	id c,s;int e;
	c = [[termOne coefficient] cheapCopy];
	e = ca_vardns_expat(value,ix->i,ca_vardns_getargs(self));
	s = (e)?[[symbols at:ix->i] cheapCopy]:nil;
	return [termOne coefficient:c symbol:s exponent:e];
    } else {
	return nil;
    }
}


- removeTerm
{
    int i = ca_vardns_findfirst(value,ca_vardns_getargs(self));
    
    if (i == -1) {
	return nil;
    } else {
	id lt,c,s;int e;
	assert([self checkRefCountOne] == self && [self check] == self);
	c = [[termOne coefficient] cheapCopy];
	e = ca_vardns_expat(value,i,ca_vardns_getargs(self));assert(e>0);
	s = (e)?[[symbols at:i] cheapCopy]:nil;
	lt = [termOne coefficient:c symbol:s exponent:e];
	ca_vardns_setexpat(value,i,0,ca_vardns_getargs(self));
	[self invalidate];return lt;
    }
}

- insertTerm:aTerm
{
    int i;
    
    [self checkRefCountOne];
    assert([self check] && [[aTerm coefficient] isOne]);
    if ([aTerm exponent] == 0) { [aTerm free];return self; }

    i = [symbols offsetOf:[aTerm symbol]];
    
    if (i == -1) {
	return [self error:"-insertTerm: illegal symbol"];
    } else {
	int e = ca_vardns_expat(value,i,ca_vardns_getargs(self));
	e += [aTerm exponent];
	if (e < 0) {
	    return nil;
	} else {
	    ca_vardns_setexpat(value,i,e,ca_vardns_getargs(self));
	    assert([self check]);return [self invalidate];
	}
    }
}


- fileOutOn:aFiler
{
    ca_vardns_args args;
    [super fileOutOn:aFiler];
    args = ca_vardns_getargs(self);
    ca_vardns_args_fileOut(aFiler,&args);
    ca_vardns_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_vardns_args args;
    [super fileInFrom:aFiler];
    ca_vardns_args_fileIn(aFiler,&args);
    ca_vardns_fileIn(aFiler,&value,args);
    return self;
}

@end

