#include "cakit.h"

#define VAL_MIN -32500
#define VAL_MAX +32500

static int ca_intobj_check(ca_intobj_t self,ca_intobj_args args)
{
    assert(args.proto);[args.proto check];
    
    assert(VAL_MIN == - VAL_MAX); /* for opposite tests */
    assert(VAL_MAX * VAL_MAX + VAL_MAX * VAL_MAX < INT_MAX);
    assert(VAL_MIN * VAL_MAX - VAL_MIN * VAL_MAX > INT_MIN);

    if (self->hit) {
	assert(VAL_MIN <= self->val && self->val <= VAL_MAX);
	if (self->obj) {
	    id obj = [args.proto intValue:self->val];
	    assert([obj isEqual:self->obj]);
	    [obj free];
	}
    } else {
	int val;id obj = self->obj;
	assert([obj check]);val = [obj intValue];
	assert(val < VAL_MIN || VAL_MAX < val);
    }
    
    return 1;
}

/*
 * return new reference to object or create it if necessary
 */

static id ca_intobj_obj(ca_intobj_t self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return [self->obj cheapCopy];
    } else {
	id obj;assert(self->hit);
	obj = [args.proto intValue:self->val];
	assert([obj intValue] == self->val);
	return obj;
    }
}

/*
 * works inplace on "self"; creates object if it doesn't exists; returns it.
 */

static id ca_intobj_makeobj(ca_intobj_c *self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return self->obj;
    } else {
	assert(self->hit);return self->obj = [args.proto intValue:self->val];
    }
}

/*
 * initializing "c" by object
 */

static void ca_intobj_ooobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by object result of inplace operation
 * used like :
 *
 *		ca_intobj_makeobj(c,args);
 *		ca_intobj_inobj(c,[c->obj negateSelf],args)
 */

static void ca_intobj_inobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    /* c->obj is invalid reference now... replace it by result inplace op
     * we cannot check "c" here (as the object is invalid...)
     */
    assert(c->obj != nil);
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * initializing "c" by int value
 */

static void ca_intobj_ooval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = nil;
    } else {
	c->hit = 0;c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by int value
 */

static void ca_intobj_inval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;
	if (c->obj) c->obj = [c->obj free];
    } else {
	c->hit = 0;
	if (c->obj) c->obj = [c->obj free];
	c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_oozero(ca_intobj_c *c,ca_intobj_args args)
{
    c->hit = 1;c->val = 0;c->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_copy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj cheapCopy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_deepcopy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj copy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_move(ca_intobj_c *c,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    c->hit = a->hit;
    c->val = a->val;a->val = SHRT_MAX;
    c->obj = a->obj;a->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_clear(ca_intobj_c *c,ca_intobj_args args)
{
    c->val = SHRT_MAX;if (c->obj) c->obj = [c->obj free];
}

static ca_intobj_args ca_intobj_getargs(id self)
{
    ca_intobj_args args;args.proto = self;return args;
}

static int ca_intobj_size(ca_intobj_args args)
{
    return 1;				/* pointer increment */
}

static int ca_intobj_bsize(ca_intobj_args args)
{
    return sizeof(ca_intobj_c);		/* byte size carrier */
}

#if 0
static ca_intobj_t  ca_intobj_id2t(id object) { /* not needed */ }
static ca_intobj_c *ca_intobj_id2t(id object) { /* not needed */ }
#endif

/*
 * initialize scratch space "ref" with a read-only value 
 */

static ca_intobj_t ca_intobj_id2ref(id object,ca_intobj_c *ref)
{
    ca_intobj_args args = ca_intobj_getargs(object);	/* use object as proto */
    
    ca_intobj_ooobj(ref,object,args);return ref;
}

/*
 * return a new object, containing a copy of "t"
 */

static id ca_intobj_t2id(id proto,ca_intobj_t t)
{
    ca_intobj_args args = ca_intobj_getargs(proto);
    return ca_intobj_obj(t,args);
}

/*
 * return a new object, containing "t" itself (moved into object)
 */

static id ca_intobj_c2id(id proto,ca_intobj_c *c)
{
    id res;
    ca_intobj_args args = ca_intobj_getargs(proto);
    
    res = ca_intobj_obj(c,args);ca_intobj_clear(c,args);
     
    return res;
}

static unsigned ca_intobj_hash(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit)?a->val:[a->obj hash];
}

static int ca_intobj_iseq(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return YES;
    } else {
	if (a->hit) {
	    return (b->hit)?a->val==b->val:NO;
	} else {
	    return (b->hit)?NO:[a->obj isEqual:b->obj];
	}
    }
}

static int ca_intobj_sgn(ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	short int sgn = a->val;return (sgn==0)?0:((sgn<0)?-1:+1);
    } else {
	return [a->obj sign];
    }
}

static int ca_intobj_cmp(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    int cmp;
    
    if (a==b) {
	cmp = 0;
    } else {
	if (a->hit && b->hit) {
	    cmp = a->val - b->val;
	    cmp = (cmp==0)?0:((cmp<0)?-1:+1);
	} else {
	    id A = ca_intobj_obj(a,args);
	    id B = ca_intobj_obj(b,args);
	    cmp = [A compare:B];[A free];[B free];
	}
    }
    
    return cmp;
}

static int ca_intobj_iszero(ca_intobj_t a,ca_intobj_args args)
{
    int iszero = (a->hit && a->val == 0);
    assert(iszero || (a->hit&&a->val!=0) || (a->hit==0&&[a->obj notZero]));
    return iszero;
}

static int ca_intobj_isop(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return NO;
    } else {
	if (a->hit) {
	    assert(VAL_MAX + VAL_MIN == 0);
	    assert(VAL_MIN <= a->val && a->val <= VAL_MAX);
	    assert(VAL_MIN <= b->val && b->val <= VAL_MAX);
	    return (b->hit)?a->val+b->val==0:NO;
	} else {
	    return (b->hit)?NO:[a->obj isOpposite:b->obj];
	}
    }
}

static void ca_intobj_ooneg(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = - a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj negate],args);
    }
    
    assert(ca_intobj_isop(c,a,args));
}

static void ca_intobj_inneg(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = - c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj negateSelf],args);
    }
}

static void ca_intobj_oodbl(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_args args)
{
    if (a->hit) {
	int C = 2 * a->val;C = (v==+1)?+C:-C;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj _double:v],args);
    }
}

static void ca_intobj_indbl(ca_intobj_c *c,int v,ca_intobj_args args)
{
    if (c->hit) {
	int C = 2 * c->val;C = (v==+1)?+C:-C;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj _doubleSelf:v],args);
    }
}

static void ca_intobj_ooadd(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _add:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inadd(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B:w],args);
	[B free];
    }
}

static int ca_intobj_isone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == +1);
}

static int ca_intobj_isminusone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == -1);
}

static void ca_intobj_oosqr(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = a->val * a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj square],args);
    }
}

static void ca_intobj_insqr(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = c->val * c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj squareSelf],args);
    }
}

static void ca_intobj_oomul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A multiply:B];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmul(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj multiplySelf:B],args);
	[B free];
    }
}

static void ca_intobj_oomulv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val * v;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _multiply:B:v];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmulv(ca_intobj_c *c,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val * v;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:B:v],args);
	[B free];
    }
}

static void ca_intobj_ooaddmul(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && b->hit && t->hit) {
	int C = a->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	T = ca_intobj_obj(t,args);
	C = [A _add:v:B multiply:T:w];[A free];[B free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddmul(ca_intobj_c *c,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && b->hit && t->hit) {
	int C = c->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B multiply:T:w],args);
	[B free];[T free];
    }
}

static void ca_intobj_oomuladd(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit) {
	int C = a->val * s->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	C = [A _multiply:S:v add:B:w];
	[A free];[B free];[S free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladd(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit) {
	int C = c->val * s->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B:w],args);
	[B free];[S free];
    }
}

static void ca_intobj_oomuladdmul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit && t->hit) {
	int C = a->val * s->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	T = ca_intobj_obj(t,args);
	C = [A _multiply:S:v add:B multiply:T:w];
	[A free];[B free];[S free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladdmul(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit && t->hit) {
	int C = c->val * s->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B multiply:T:w],args);
	[B free];[S free];[T free];
    }
}

static void ca_intobj_ooaddsqr(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _addSquare:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddsqr(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelfSquare:v:B:w],args);
	[B free];
    }
}

static BOOL ca_intobj_oodiv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C;
	if (b->val == 0 || a->val % b->val != 0) {
	    return FAILURE;
	} else {
	    C = a->val / b->val;
	}
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A divide:B];[A free];[B free];
	if (C==nil) return FAILURE;
	ca_intobj_ooobj(c,C,args);
    }
    
    return SUCCESS;
}

static void ca_intobj_indiv(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C;
	if (b->val == 0 || c->val % b->val != 0) {
	    fprintf(stderr,"-exact division failed\n");abort();
	} else {
	    C = c->val / b->val;
	}
	ca_intobj_inval(c,C,args);
    } else {
	id C,B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	C = [c->obj divideSelf:B];
	if (C==nil) {fprintf(stderr,"-exact division failed\n");abort();}
	ca_intobj_inobj(c,C,args);
	[B free];
    }
}

static void ca_intobj_args_fileOut(id aFile,ca_intobj_args *args) {}
static void ca_intobj_args_fileIn(id aFile,ca_intobj_args *args) {}

static void ca_intobj_fileOut(id aFiler,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(a,args));
    [aFiler fileOut:&a->hit type:'S']; 
    if (a->hit) {
	[aFiler fileOut:&a->val type:'s']; /* signed ! */
    } else {
	[aFiler fileOut:&a->obj type:'@'];
    }
}

static void ca_intobj_fileIn(id aFiler,ca_intobj_c *c,ca_intobj_args args)
{
    [aFiler fileIn:&c->hit type:'S'];
    if (c->hit) {
	[aFiler fileIn:&c->val type:'s'];
        c->obj = nil;
    } else {
	[aFiler fileIn:&c->obj type:'@'];
    }
}


/*
 * Computer Algebra Kit (c) 1993,98 by Comp.Alg.Objects.  All Rights Reserved.
 */

static int ca_vardnsmonom_size(ca_vardnsmonom_args args)
{
    return args.nvars; /* size to add to int pointer */
}

static int ca_vardnsmonom_bsize(ca_vardnsmonom_args args)
{
    return args.nvars * sizeof(int); /* byte size exponent array */
}

static int ca_vardnsmonom_check(ca_vardnsmonom_t self,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) assert(*self++ >= 0);
    return 1;
}

static ca_vardnsmonom_t ca_vardnsmonom_id2t(id object)
{
    return [object ca_vardnsmonom_value];
}

static id ca_vardnsmonom_t2id(id proto,ca_vardnsmonom_t f)
{
    return [proto ca_vardnsmonom_value:f];
}

static ca_vardnsmonom_args ca_vardnsmonom_getargs(id self)
{
    ca_vardnsmonom_args res;
    res.nvars = [[self symbols] size];
    return res; 
}

static ca_vardnsmonom_t ca_vardnsmonom_alloc(ca_vardnsmonom_args args)
{
    return (ca_vardnsmonom_t)OC_Malloc(ca_vardnsmonom_bsize(args));
}

static void ca_vardnsmonom_init(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;while (nvars--) *c++ = 0;
}

static void ca_vardnsmonom_copy(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;while (nvars--) *c++ = *a++;
}

static void ca_vardnsmonom_clear(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    /* nothing to clear... */
}

static void ca_vardnsmonom_free(ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    OC_Free(c);
}

static void ca_vardnsmonom_args_fileOut(id aFiler,ca_vardnsmonom_args *args)
{
    [aFiler fileOut:&args->nvars type:'i'];
}

static void ca_vardnsmonom_args_fileIn(id aFiler,ca_vardnsmonom_args *args)
{
    [aFiler fileIn:&args->nvars type:'i'];
}

static void ca_vardnsmonom_fileOut(id aFiler,ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int n = args.nvars;
    while (n--) [aFiler fileOut:c++ type:'i'];
}

static void ca_vardnsmonom_fileIn(id aFiler,ca_vardnsmonom_c *c,ca_vardnsmonom_args args)
{
    int n = args.nvars;
    while (n--) [aFiler fileIn:c++ type:'i'];
}

static unsigned ca_vardnsmonom_hash(ca_vardnsmonom_t self,ca_vardnsmonom_args args)
{
    assert(args.nvars>0);return self[0];
}

static int ca_vardnsmonom_deg(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int sum = 0;
    int nvars = args.nvars;
    while (nvars--) sum += *self++;
    return sum;
}

static int ca_vardnsmonom_numterms(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int numterms = 0;
    int nvars = args.nvars;
    while (nvars--) if (*self++) numterms++;
    return numterms;
}

static int ca_vardnsmonom_isone(ca_vardnsmonom_t self ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if (*self) return 0; self++; }
    return 1;
}

static int ca_vardnsmonom_iseq(ca_vardnsmonom_t a,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if ((*a) != (*b)) return 0; a++;b++; }
    return 1;
}

static int ca_vardnsmonom_iseqmul(ca_vardnsmonom_t a,ca_vardnsmonom_t x,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { if ((*a) != (*x) + (*b)) return 0; a++;x++;b++; }
    return 1;
}

static int ca_vardnsmonom_cmp(ca_vardnsmonom_t a,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;

    /* lexicographic cmp *MUST* work from right to left
     * because last symbol is largest for vardns order
     */

    while (nvars--) {
	if ((*a) < (*b)) return -1;
	if ((*a) > (*b)) return +1;
	a++;b++;
    }
    
    return 0;
}

static int ca_vardnsmonom_cmpmul(ca_vardnsmonom_t a,ca_vardnsmonom_t x,ca_vardnsmonom_t b ,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;

    /* lexicographic cmp *MUST* work from right to left
     * because last symbol is largest for vardns order
     */

    while (nvars--) {
	int sum = *x + *b;
	if ((*a) < sum) return -1;
	if ((*a) > sum) return +1;
	a++;x++;b++;
    }
    
    return 0;
}

static void ca_vardnsmonom_oomul(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { (*c) = (*a) + (*b); c++; a++; b++; }
}

static void ca_vardnsmonom_inmul(ca_vardnsmonom_c *c,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) { (*c) += (*b); c++; b++; }
}

static BOOL ca_vardnsmonom_oodiv(ca_vardnsmonom_c *c,ca_vardnsmonom_t a,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) {
    	int q = *a - *b;
	if (q<0) return FAILURE;
	*c = q; c++; a++; b++;
    }
    
    return SUCCESS;
}

static void ca_vardnsmonom_indiv(ca_vardnsmonom_c *c,ca_vardnsmonom_t b,ca_vardnsmonom_args args)
{
    int nvars = args.nvars;
    while (nvars--) {
    	int q = *c - *b;
	if (q<0) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
	*c = q; c++; b++;
    }
}
/*
 * scalar is accessed by the C "->" operator since having variable sized
 * scalars _and_ monomials gets difficult on machines like the DEC alpha
 * (where we'd have to manually pad to get pointer aligned structures).
 * 
 * the monomial (exponent) is however not accessed by the C "->" operator
 * but rather by functions that can take offsets as arguments
 * this is important the monomials in variable dense representation
 * (fixarray of exp)
 *
 * stes - mar31, 1997 - CHANGES: for DEC alpha coefficient needs to be aligned
 * properly, put it before the monomial and drop the idea of variable sized
 * coefficients.
 */

static ca_vardnsmonom_c *ca_intobj_vardnsexpdegspsterm_monom_cval(ca_intobj_vardnsexpdegspsterm_t self)
{
    return &self->data;
}

static ca_vardnsmonom_t ca_intobj_vardnsexpdegspsterm_monom(ca_intobj_vardnsexpdegspsterm_t self)
{
#if ca_vardnsmonom_pervalue
    return self->data;
#else
    return  ca_intobj_vardnsexpdegspsterm_monom_cval(self);
#endif
}

/* stes - mar31,1997
 * change for DEC alpha to exclude varsized scalars
 */

static ca_intobj_c *ca_intobj_vardnsexpdegspsterm_scal_cval(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
	return &self->scal;
}

static ca_intobj_t ca_intobj_vardnsexpdegspsterm_scal(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
#if ca_intobj_pervalue
	return self->scal;
#else
    return  ca_intobj_vardnsexpdegspsterm_scal_cval(self,args);
#endif
}

static int ca_intobj_vardnsexpdegspsterm_check(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_check(ca_intobj_vardnsexpdegspsterm_scal(self,args),args.sargs));
    assert(!ca_intobj_iszero(ca_intobj_vardnsexpdegspsterm_scal(self,args),args.sargs));
    assert(ca_vardnsmonom_check(ca_intobj_vardnsexpdegspsterm_monom(self),args.margs));
    return 1;
}

static int ca_intobj_vardnsexpdegspsterm_bsize(ca_intobj_vardnsexpdegspsterm_args args)
{
	/* this is sizeof(ca_vardnsmonom_c) too much but why bother ? */
    return sizeof(struct ca_intobj_vardnsexpdegspsterm) + ca_vardnsmonom_bsize(args.margs);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_alloc(ca_intobj_vardnsexpdegspsterm_args args)
{
    return (ca_intobj_vardnsexpdegspsterm_t)OC_Malloc(ca_intobj_vardnsexpdegspsterm_bsize(args));
}

static void ca_intobj_vardnsexpdegspsterm_clear(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_vardnsmonom_clear(ca_intobj_vardnsexpdegspsterm_monom_cval(self),args.margs);
    ca_intobj_clear(ca_intobj_vardnsexpdegspsterm_scal_cval(self,args),args.sargs);
}

static void ca_intobj_vardnsexpdegspsterm_free(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
    /* invalidating links not always helps : buggy code might work because
     * links were cleared...        self->nlink = NULL;self->plink = NULL;
     */
    OC_Free(self);
}

static int ca_intobj_vardnsexpdegspsterm_cmp(ca_intobj_vardnsexpdegspsterm_t a,ca_intobj_vardnsexpdegspsterm_t b,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_check(a,args) && ca_intobj_vardnsexpdegspsterm_check(b,args));
    return ca_vardnsmonom_cmp(ca_intobj_vardnsexpdegspsterm_monom(a),ca_intobj_vardnsexpdegspsterm_monom(b),args.margs);
}

static int ca_intobj_vardnsexpdegspsterm_cmpmonom(ca_intobj_vardnsexpdegspsterm_t a,ca_vardnsmonom_t b,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_check(a,args) && ca_vardnsmonom_check(b,args.margs));
    return ca_vardnsmonom_cmp(ca_intobj_vardnsexpdegspsterm_monom(a),b,args.margs);
}

static int ca_intobj_vardnsexpdegspsterm_cmpmul(ca_intobj_vardnsexpdegspsterm_t a,ca_intobj_vardnsexpdegspsterm_t b,ca_vardnsmonom_t m ,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_check(a,args) && ca_intobj_vardnsexpdegspsterm_check(b,args));
    assert(ca_vardnsmonom_check(m,args.margs));
    return ca_vardnsmonom_cmpmul(ca_intobj_vardnsexpdegspsterm_monom(a),ca_intobj_vardnsexpdegspsterm_monom(b),m,args.margs);
}

static int ca_intobj_vardnsexpdegspsterm_deg(ca_intobj_vardnsexpdegspsterm_t a,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_vardnsmonom_deg(ca_intobj_vardnsexpdegspsterm_monom(a),args.margs);
}

static int ca_intobj_vardnsexpdegspsterm_iszero(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_intobj_iszero(ca_intobj_vardnsexpdegspsterm_scal(self,args),args.sargs);
}

static int ca_intobj_vardnsexpdegspsterm_isone(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_intobj_isone(ca_intobj_vardnsexpdegspsterm_scal(self,args),args.sargs) && ca_vardnsmonom_isone(ca_intobj_vardnsexpdegspsterm_monom(self),args.margs);
}

static int ca_intobj_vardnsexpdegspsterm_isminusone(ca_intobj_vardnsexpdegspsterm_t self,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_intobj_isminusone(ca_intobj_vardnsexpdegspsterm_scal(self,args),args.sargs) && ca_vardnsmonom_isone(ca_intobj_vardnsexpdegspsterm_monom(self),args.margs);
}

static int ca_intobj_vardnsexpdegspsterm_iseq(ca_intobj_vardnsexpdegspsterm_t a,ca_intobj_vardnsexpdegspsterm_t b,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_check(a,args) && ca_intobj_vardnsexpdegspsterm_check(b,args));
    return ca_vardnsmonom_iseq(ca_intobj_vardnsexpdegspsterm_monom(a),ca_intobj_vardnsexpdegspsterm_monom(b),args.margs) && ca_intobj_iseq(ca_intobj_vardnsexpdegspsterm_scal(a,args),ca_intobj_vardnsexpdegspsterm_scal(b,args),args.sargs);
}

static int ca_intobj_vardnsexpdegspsterm_isop(ca_intobj_vardnsexpdegspsterm_t a,ca_intobj_vardnsexpdegspsterm_t b,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_check(a,args) && ca_intobj_vardnsexpdegspsterm_check(b,args));
    return ca_vardnsmonom_iseq(ca_intobj_vardnsexpdegspsterm_monom(a),ca_intobj_vardnsexpdegspsterm_monom(b),args.margs) && ca_intobj_isop(ca_intobj_vardnsexpdegspsterm_scal(a,args),ca_intobj_vardnsexpdegspsterm_scal(b,args),args.sargs);
}

static void ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A)
{
    C->nlink = A;A->plink = C;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A)
{
    ca_intobj_vardnsexpdegspsterm_glue(C,A);return A;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_delete(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t C = A->nlink;
    ca_intobj_vardnsexpdegspsterm_glue(A->plink,C);
    ca_intobj_vardnsexpdegspsterm_clear(A,args);
    ca_intobj_vardnsexpdegspsterm_free(A,args);
    return C;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_condcns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t new,ca_intobj_vardnsexpdegspsterm_args args)
{
    if (ca_intobj_vardnsexpdegspsterm_iszero(new,args)) {
	ca_intobj_vardnsexpdegspsterm_clear(new,args);ca_intobj_vardnsexpdegspsterm_free(new,args);return C;
    } else {
	assert(ca_intobj_vardnsexpdegspsterm_check(new,args));return ca_intobj_vardnsexpdegspsterm_cns(C,new);
    }
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_conddelete(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_args args)
{
    if (ca_intobj_vardnsexpdegspsterm_iszero(A,args)) {
	return ca_intobj_vardnsexpdegspsterm_delete(A,args);
    } else {
	return A->nlink;
    }
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_copymonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);
    
    ca_intobj_copy(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_intobj_vardnsexpdegspsterm_check(new,args));return ca_intobj_vardnsexpdegspsterm_cns(C,new);
}

static void ca_intobj_vardnsexpdegspsterm_copymonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_copymonomsc_cns(C,m,s,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_deepcopymonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);
    
    ca_intobj_deepcopy(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),s,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),m,args.margs);
    
    assert(ca_intobj_vardnsexpdegspsterm_check(new,args));return ca_intobj_vardnsexpdegspsterm_cns(C,new);
}

static void ca_intobj_vardnsexpdegspsterm_deepcopymonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_deepcopymonomsc_cns(C,m,s,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_copy_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_intobj_vardnsexpdegspsterm_copymonomsc_cns(C,ca_intobj_vardnsexpdegspsterm_monom(A),ca_intobj_vardnsexpdegspsterm_scal(A,args),args);
}

static void ca_intobj_vardnsexpdegspsterm_copy(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_copy_cns(C,A,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_deepcopy_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_intobj_vardnsexpdegspsterm_deepcopymonomsc_cns(C,ca_intobj_vardnsexpdegspsterm_monom(A),ca_intobj_vardnsexpdegspsterm_scal(A,args),args);
}

static void ca_intobj_vardnsexpdegspsterm_deepcopy(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_deepcopy_cns(C,A,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_ooneg_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO);

    ca_intobj_ooneg(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_cns(C,new);
}

static void ca_intobj_vardnsexpdegspsterm_ooneg(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_ooneg_cns(C,A,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inneg(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO);
    ca_intobj_inneg(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),args.sargs);
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO);
    return A->nlink;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oonegv_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    return (v == -1)?ca_intobj_vardnsexpdegspsterm_ooneg_cns(C,A,args):ca_intobj_vardnsexpdegspsterm_copy_cns(C,A,args);
}

static void ca_intobj_vardnsexpdegspsterm_oonegv(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oonegv_cns(C,A,v,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_innegv(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    return (v == -1)?ca_intobj_vardnsexpdegspsterm_inneg(A,args):A->nlink;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oodbl_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO);

    ca_intobj_oodbl(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),v,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static void ca_intobj_vardnsexpdegspsterm_oodbl(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oodbl_cns(C,A,v,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_indbl(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO);
    ca_intobj_indbl(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),v,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_ooaddsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);
    ca_intobj_ooadd(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),v,B,w,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inaddsc(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_inadd(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),v,B,w,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_ooadd_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == ca_intobj_vardnsexpdegspsterm_cmp(A,B,args));
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO && ca_intobj_vardnsexpdegspsterm_iszero(B,args) == NO);
    return ca_intobj_vardnsexpdegspsterm_ooaddsc_cns(C,A,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),w,args);
}

static void ca_intobj_vardnsexpdegspsterm_ooadd(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_ooadd_cns(C,A,v,B,w,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inadd(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == ca_intobj_vardnsexpdegspsterm_cmp(A,B,args));
    assert(ca_intobj_vardnsexpdegspsterm_iszero(A,args) == NO && ca_intobj_vardnsexpdegspsterm_iszero(B,args) == NO);
    return ca_intobj_vardnsexpdegspsterm_inaddsc(A,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);

    ca_intobj_oomulv(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),s,v,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static void ca_intobj_vardnsexpdegspsterm_oomulsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A, ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulsc(ca_intobj_vardnsexpdegspsterm_t A, ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_inmulv(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),s,v,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);

    ca_intobj_oomulv(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),s,v,args.sargs);
    ca_vardnsmonom_oomul(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),m,args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static void ca_intobj_vardnsexpdegspsterm_oomulmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_inmulv(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),s,v,args.sargs);
    ca_vardnsmonom_inmul(ca_intobj_vardnsexpdegspsterm_monom_cval(A),m,args.margs);

    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_ooaddscmulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);

    ca_intobj_ooaddmul(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),v,B,t,w,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inaddscmulsc(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_inaddmul(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),v,B,t,w,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_ooaddmulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsexpdegspsterm_ooaddscmulsc_cns(C,A,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_ooaddmulmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsexpdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
}

static void ca_intobj_vardnsexpdegspsterm_ooaddmulsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args),CT);
}

static void ca_intobj_vardnsexpdegspsterm_ooaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inaddmulsc(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsexpdegspsterm_inaddscmulsc(A,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t A,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsexpdegspsterm_inaddscmulsc(A,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulscaddsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);

    ca_intobj_oomuladd(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),s,v,B,w,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulscaddsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_inmuladd(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),s,v,B,w,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulscadd_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsexpdegspsterm_oomulscaddsc_cns(C,A,s,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),w,args);
}

static void ca_intobj_vardnsexpdegspsterm_oomulscadd(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulscadd(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsexpdegspsterm_inmulscaddsc(A,s,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulscaddscmulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);
    ca_intobj_oomuladdmul(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),s,v,B,t,w,args.sargs);
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulscaddscmulsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_inmuladdmul(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),s,v,B,t,w,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulscaddmulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsexpdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oomulscaddmulmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsexpdegspsterm_oomulscaddscmulsc_cns(C,A,s,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static void ca_intobj_vardnsexpdegspsterm_oomulscaddmulsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args),CT);
}

static void ca_intobj_vardnsexpdegspsterm_oomulscaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulscaddmulsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmp(A,B,args)));
    return ca_intobj_vardnsexpdegspsterm_inmulscaddscmulsc(A,s,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_inmulscaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_vardnsmonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(0 == (ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args)));
    return ca_intobj_vardnsexpdegspsterm_inmulscaddscmulsc(A,s,v,ca_intobj_vardnsexpdegspsterm_scal(B,args),t,w,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oodivsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    BOOL failed;
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);

    failed = ca_intobj_oodiv(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),s,args.sargs);
    if (failed) {
	ca_intobj_vardnsexpdegspsterm_free(new,args);return NULL;
    }
    
    ca_vardnsmonom_copy(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_intobj_vardnsexpdegspsterm_oodivsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A, ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    if ((C = ca_intobj_vardnsexpdegspsterm_oodivsc_cns(C,A,s,args))) {
	ca_intobj_vardnsexpdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_indivsc(ca_intobj_vardnsexpdegspsterm_t A, ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_indiv(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),s,args.sargs);
    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_oodivmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    BOOL sfailed,mfailed;
    ca_intobj_vardnsexpdegspsterm_t new = ca_intobj_vardnsexpdegspsterm_alloc(args);

    sfailed = ca_intobj_oodiv(ca_intobj_vardnsexpdegspsterm_scal_cval(new,args),ca_intobj_vardnsexpdegspsterm_scal(A,args),s,args.sargs);
    mfailed = ca_vardnsmonom_oodiv(ca_intobj_vardnsexpdegspsterm_monom_cval(new),ca_intobj_vardnsexpdegspsterm_monom(A),m,args.margs);
    
    if (sfailed || mfailed) {
	/* both failures are possible in the multivariate case */
	ca_intobj_vardnsexpdegspsterm_free(new,args);return NULL;
    }
    
    return ca_intobj_vardnsexpdegspsterm_condcns(C,new,args);
}

#if 0 /* unused ! why did I write this ? */
static BOOL ca_intobj_vardnsexpdegspsterm_oodivmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    if ((C = ca_intobj_vardnsexpdegspsterm_oodivmonomsc_cns(C,A,m,s,args))) {
	ca_intobj_vardnsexpdegspsterm_glue(C,CT);return SUCCESS;
    } else {
	return FAILURE;
    }
}
#endif

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_indivmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_vardnsmonom_t m, ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_indiv(ca_intobj_vardnsexpdegspsterm_scal_cval(A,args),s,args.sargs);
    ca_vardnsmonom_indiv(ca_intobj_vardnsexpdegspsterm_monom_cval(A),m,args.margs);

    return ca_intobj_vardnsexpdegspsterm_conddelete(A,args);
}

static void ca_intobj_vardnsexpdegspsterm_args_fileOut(id aFiler,ca_intobj_vardnsexpdegspsterm_args *args)
{
    ca_intobj_args_fileOut(aFiler,&args->sargs);
    ca_vardnsmonom_args_fileOut(aFiler,&args->margs);
}
static void ca_intobj_vardnsexpdegspsterm_args_fileIn(id aFiler,ca_intobj_vardnsexpdegspsterm_args *args)
{
    ca_intobj_args_fileIn(aFiler,&args->sargs);
    ca_vardnsmonom_args_fileIn(aFiler,&args->margs);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_fileOut(id aFiler,ca_intobj_vardnsexpdegspsterm_t self ,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_vardnsexpdegspsterm_check(self,args));
    
    ca_intobj_fileOut(aFiler,ca_intobj_vardnsexpdegspsterm_scal_cval(self,args),args.sargs);
    ca_vardnsmonom_fileOut(aFiler,ca_intobj_vardnsexpdegspsterm_monom_cval(self),args.margs);
    
    return self->nlink;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspsterm_fileIn_cns(id aFiler,ca_intobj_vardnsexpdegspsterm_t C ,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t self = ca_intobj_vardnsexpdegspsterm_alloc(args);
    
    ca_intobj_fileIn(aFiler,ca_intobj_vardnsexpdegspsterm_scal_cval(self,args),args.sargs);
    ca_vardnsmonom_fileIn(aFiler,ca_intobj_vardnsexpdegspsterm_monom_cval(self),args.margs);
    
    return ca_intobj_vardnsexpdegspsterm_cns(C,self);
}

static int ca_intobj_vardnsexpdegspslist_check(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t prev;
    
    for(prev = NULL;A!=AT;prev = A,A = A->nlink) {
	if (ca_intobj_vardnsexpdegspsterm_check(A,args) == 0) return 0;
	if (prev && ca_intobj_vardnsexpdegspsterm_cmp(prev,A,args) <= 0) return 0;
    }
    
    return 1;
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_copy_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) { C = ca_intobj_vardnsexpdegspsterm_copy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_intobj_vardnsexpdegspslist_copy(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_copy_cns(C,A,AT,args),CT);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_deepcopy_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) { C = ca_intobj_vardnsexpdegspsterm_deepcopy_cns(C,A,args);A = A->nlink; }
    return C;
}

static void ca_intobj_vardnsexpdegspslist_deepcopy(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_deepcopy_cns(C,A,AT,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_clear(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) { ca_intobj_vardnsexpdegspsterm_clear(A,args);A = A->nlink; }
}

static void ca_intobj_vardnsexpdegspslist_free(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_t t;for(;A!=AT;A=t) { t = A->nlink;ca_intobj_vardnsexpdegspsterm_free(A,args); }
}

static int ca_intobj_vardnsexpdegspslist_count(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static unsigned ca_intobj_vardnsexpdegspslist_hash(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    return ca_intobj_vardnsexpdegspslist_count(A,AT);
}

static int ca_intobj_vardnsexpdegspslist_iseq(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_t B, ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_vardnsexpdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_intobj_vardnsexpdegspsterm_iseq(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_intobj_vardnsexpdegspslist_isop(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_vardnsexpdegspsterm_args args)
{
    for(;A!=AT&&B!=BT;A=A->nlink,B=B->nlink) {
	if (!ca_intobj_vardnsexpdegspsterm_isop(A,B,args)) return 0;
    }
    return (A==AT && B==BT);
}

static int ca_intobj_vardnsexpdegspslist_deg(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    return (A==AT)?-1:ca_intobj_vardnsexpdegspsterm_deg(A,args);
}

static int ca_intobj_vardnsexpdegspslist_maxdeg(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    int max = (A==AT)?-1:ca_intobj_vardnsexpdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_intobj_vardnsexpdegspsterm_deg(A,args);if (max < d) max = d;
    }
    return max;
}

static int ca_intobj_vardnsexpdegspslist_ord(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    return (A==AT)?-1:ca_intobj_vardnsexpdegspsterm_deg(AT->plink,args);
}

static int ca_intobj_vardnsexpdegspslist_mindeg(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    int min = (A==AT)?-1:ca_intobj_vardnsexpdegspsterm_deg(A,args);
    for(A=A->nlink;A != AT;A = A->nlink) {
	int d = ca_intobj_vardnsexpdegspsterm_deg(A,args);if (min > d) min = d;
    }
    return min;
}

static void ca_intobj_vardnsexpdegspslist_insert(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(ca_intobj_iszero(s,args.sargs) == NO);
    
    /* adding to zero */
    if (A == AT) {
	ca_intobj_vardnsexpdegspsterm_copymonomsc(A,m,s,A,args);return;
    }
    
    /* tail append */
    if (ca_intobj_vardnsexpdegspsterm_cmpmonom(AT->plink,m,args) > 0) {
	ca_intobj_vardnsexpdegspsterm_copymonomsc(AT->plink,m,s,AT,args);return;
    }
    
    /* insertion */
    while (A != AT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmpmonom(A,m,args);
	if (cmp == 0) {
	    ca_intobj_vardnsexpdegspsterm_inaddsc(A,+1,s,+1,args);return;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_copymonomsc(A->plink,m,s,A,args);return;
	}
	A = A->nlink;
    }
 
    fprintf(stderr,"Shouldn't happen\n");
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_ooneg_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsexpdegspsterm_ooneg_cns(C,A,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsexpdegspslist_ooneg(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_ooneg_cns(C,A,AT,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inneg(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_inneg(A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oonegv_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    if (v == -1) {
	return ca_intobj_vardnsexpdegspslist_ooneg_cns(C,A,AT,args);
    } else {
	return ca_intobj_vardnsexpdegspslist_copy_cns(C,A,AT,args);
    }
}

static void ca_intobj_vardnsexpdegspslist_oonegv(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_innegv(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    if (v == -1) ca_intobj_vardnsexpdegspslist_inneg(A,AT,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oodbl_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsexpdegspsterm_oodbl_cns(C,A,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsexpdegspslist_oodbl(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oodbl_cns(C,A,AT,v,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_indbl(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_indbl(A,v,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_ooadd_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsexpdegspsterm_ooadd_cns(C,A,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsexpdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_intobj_vardnsexpdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_ooadd(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_ooadd_cns(C,A,AT,v,B,BT,w,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inadd(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsexpdegspsterm_inadd(A,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsexpdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsexpdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_intobj_vardnsexpdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oomulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsexpdegspslist_oomulsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inmulsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_inmulsc(A,s,v,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oomulmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) {
	C = ca_intobj_vardnsexpdegspsterm_oomulmonomsc_cns(C,A,m,s,v,args);
	A = A->nlink;
    }
    return C;
}

static void ca_intobj_vardnsexpdegspslist_oomulmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m, ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oomulmonomsc_cns(C,A,AT,m,s,v,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inmulmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_inmulmonomsc(A,m,s,v,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_ooaddmulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsexpdegspsterm_ooaddmulsc_cns(C,A,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsexpdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_intobj_vardnsexpdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_ooaddmulsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_ooaddmulsc_cns(C,A,AT,v,B,BT,t,w,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inaddmulsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsexpdegspsterm_inaddmulsc(A,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsexpdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsexpdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_intobj_vardnsexpdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_ooaddmulmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsexpdegspsterm_ooaddmulmonomsc_cns(C,A,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsexpdegspsterm_oonegv_cns(C,A,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsexpdegspslist_oonegv_cns(C,A,AT,v,args);
    if (B != BT) return ca_intobj_vardnsexpdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_ooaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_ooaddmulmonomsc_cns(C,A,AT,v,B,BT,m,t,w,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsexpdegspsterm_inaddmulmonomsc(A,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsexpdegspsterm_innegv(A,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsexpdegspslist_innegv(A,AT,v,args);
    if (B != BT) ca_intobj_vardnsexpdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oomulscadd_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulscadd_cns(C,A,s,v,B,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oonegv_cns(C,B,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_intobj_vardnsexpdegspslist_oonegv_cns(C,B,BT,w,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_oomulscadd(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT, int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oomulscadd_cns(C,A,AT,s,v,B,BT,w,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inmulscadd(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsexpdegspsterm_inmulscadd(A,s,v,B,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_oonegv(A->plink,B,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsexpdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsexpdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_intobj_vardnsexpdegspslist_oonegv(A->plink,B,BT,w,A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oomulscaddmulsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulscaddmulsc_cns(C,A,s,v,B,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,B,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_intobj_vardnsexpdegspslist_oomulsc_cns(C,B,BT,t,w,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_oomulscaddmulsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oomulscaddmulsc_cns(C,A,AT,s,v,B,BT,t,w,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inmulscaddmulsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmp(A,B,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsexpdegspsterm_inmulscaddmulsc(A,s,v,B,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_oomulsc(A->plink,B,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsexpdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsexpdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_intobj_vardnsexpdegspslist_oomulsc(A->plink,B,BT,t,w,A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_oomulscaddmulmonomsc_cns(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulscaddmulmonomsc_cns(C,A,s,v,B,m,t,w,args);
	    A = A->nlink;B = B->nlink;continue;
	}
	if (cmp < 0) {
	    C = ca_intobj_vardnsexpdegspsterm_oomulmonomsc_cns(C,B,m,t,w,args);
	    B = B->nlink;
	} else {
	    C = ca_intobj_vardnsexpdegspsterm_oomulsc_cns(C,A,s,v,args);
	    A = A->nlink;
	}
    }
 
    if (A != AT) return ca_intobj_vardnsexpdegspslist_oomulsc_cns(C,A,AT,s,v,args);
    if (B != BT) return ca_intobj_vardnsexpdegspslist_oomulmonomsc_cns(C,B,BT,m,t,w,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_oomulscaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m, ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_oomulscaddmulmonomsc_cns(C,A,AT,s,v,B,BT,m,t,w,args),CT);
}

static void ca_intobj_vardnsexpdegspslist_inmulscaddmulmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspsterm_t B,ca_intobj_vardnsexpdegspsterm_t BT,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspsterm_args args)
{
    assert(A != B);
    
    while (A != AT && B != BT) {
	int cmp = ca_intobj_vardnsexpdegspsterm_cmpmul(A,B,m,args);
	if (cmp == 0) {
	    A = ca_intobj_vardnsexpdegspsterm_inmulscaddmulmonomsc(A,s,v,B,m,t,w,args);
	    B = B->nlink;continue;
	}
	if (cmp < 0) {
	    ca_intobj_vardnsexpdegspsterm_oomulmonomsc(A->plink,B,m,t,w,A,args);
	    B = B->nlink;
	} else {
	    A = ca_intobj_vardnsexpdegspsterm_inmulsc(A,s,v,args);
	}
    }
 
    if (A != AT) ca_intobj_vardnsexpdegspslist_inmulsc(A,AT,s,v,args);
    if (B != BT) ca_intobj_vardnsexpdegspslist_oomulmonomsc(A->plink,B,BT,m,t,w,A,args);
}

static BOOL ca_intobj_vardnsexpdegspslist_oodivsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_intobj_vardnsexpdegspsterm_t nextC = ca_intobj_vardnsexpdegspsterm_oodivsc_cns(C,A,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_intobj_vardnsexpdegspsterm_glue(C,CT);return failed;
}

static void ca_intobj_vardnsexpdegspslist_indivsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_indivsc(A,s,args);
}

static BOOL ca_intobj_vardnsexpdegspslist_oodivmonomsc(ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m, ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    BOOL failed = NO;
    
    while (A != AT) {
	ca_intobj_vardnsexpdegspsterm_t nextC = ca_intobj_vardnsexpdegspsterm_oodivmonomsc_cns(C,A,m,s,args);
	if (nextC) {
	    C = nextC;
	    A = A->nlink;
	} else {
	    failed = YES;break;
	}
    }
    
    /* in any case, success or failure, have to keep list circular */
    ca_intobj_vardnsexpdegspsterm_glue(C,CT);return failed;
}

static void ca_intobj_vardnsexpdegspslist_indivmonomsc(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspsterm_args args)
{
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_indivmonomsc(A,m,s,args);
}

static void ca_intobj_vardnsexpdegspslist_fileOut(id aFiler,ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT,ca_intobj_vardnsexpdegspsterm_args args)
{
    int n;
    n = ca_intobj_vardnsexpdegspslist_count(A,AT);
    [aFiler fileOut:&n type:'i'];
    while (A != AT) A = ca_intobj_vardnsexpdegspsterm_fileOut(aFiler,A,args);
}

static ca_intobj_vardnsexpdegspsterm_t ca_intobj_vardnsexpdegspslist_fileIn_cns(id aFiler,ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_args args)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    while (n--) C = ca_intobj_vardnsexpdegspsterm_fileIn_cns(aFiler,C,args);
    return C;
}

static void ca_intobj_vardnsexpdegspslist_fileIn(id aFiler,ca_intobj_vardnsexpdegspsterm_t C,ca_intobj_vardnsexpdegspsterm_t CT,ca_intobj_vardnsexpdegspsterm_args args)
{
    ca_intobj_vardnsexpdegspsterm_glue(ca_intobj_vardnsexpdegspslist_fileIn_cns(aFiler,C,args),CT);
}

static void ca_intobj_vardnsexpdegspspol_ix_init(ca_intobj_vardnsexpdegspspol_ixt ix,id self)
{
    ca_intobj_vardnsexpdegspspol_t v = [self ca_intobj_vardnsexpdegspspol_value];
    ix->head = v;ix->ix = v;
}

static void ca_intobj_vardnsexpdegspspol_ix_clear(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    /* nothing to clear */
}

static int ca_intobj_vardnsexpdegspspol_ix_count_aux_aux(ca_intobj_vardnsexpdegspsterm_t A,ca_intobj_vardnsexpdegspsterm_t AT)
{
    int c = 0;for(;A!=AT;A=A->nlink) c++;return c;
}

static int ca_intobj_vardnsexpdegspspol_ix_count_aux(ca_intobj_vardnsexpdegspspol_t self)
{
    return ca_intobj_vardnsexpdegspspol_ix_count_aux_aux(self->nlink,self);
}

static int ca_intobj_vardnsexpdegspspol_ix_count(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    /* merde ! can't include entire genlist for ca_intobj_vardnsexpdegspslist_count */
    return ca_intobj_vardnsexpdegspspol_ix_count_aux(ix->head); 
}

static int ca_intobj_vardnsexpdegspspol_ix_isempty_aux(ca_intobj_vardnsexpdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static int ca_intobj_vardnsexpdegspspol_ix_isempty(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    /* can't include entire genpol for ca_intobj_vardnsexpdegspspol_isempty */
    return ca_intobj_vardnsexpdegspspol_ix_isempty_aux(ix->head); 
}

static int ca_intobj_vardnsexpdegspspol_ix_ismatch(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    return (ix->ix != NULL && ix->ix != ix->head);
}

static void ca_intobj_vardnsexpdegspspol_ix_tofirst(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    ix->ix = ix->head;
}

static void ca_intobj_vardnsexpdegspspol_ix_tolast(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    ix->ix = NULL;
}

static void ca_intobj_vardnsexpdegspspol_ix_to(ca_intobj_vardnsexpdegspspol_ixt ix,int i)
{
    fprintf(stderr,"at: notImplemented\n");
}

static void ca_intobj_vardnsexpdegspspol_ix_next(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    if (ix->ix == NULL) {
	/* remain on position just after last term */
    } else {
	ix->ix = ix->ix->nlink; /* next term or head = position after last */
	if (ix->ix == ix->head) ix->ix = NULL;
    }
}

static void ca_intobj_vardnsexpdegspspol_ix_prev(ca_intobj_vardnsexpdegspspol_ixt ix)
{
    if (ix->ix == ix->head) {
	/* remain on position just before first term */
    } else {
	if (ix->ix == NULL) {
	    ix->ix = ix->head->plink; /* last term or head = position before first */
	} else {
	    ix->ix = ix->ix->plink; /* previous term or head = position before first */
	}
    }
}
static int ca_intobj_vardnsexpdegspspol_check(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    assert (self->data == 0xcafebabe);
    return ca_intobj_vardnsexpdegspslist_check(self->nlink,self,args);
}

static ca_intobj_vardnsexpdegspspol_t ca_intobj_vardnsexpdegspspol_id2t(id object)
{
    return [object ca_intobj_vardnsexpdegspspol_value];
}

static ca_intobj_vardnsexpdegspspol_c *ca_intobj_vardnsexpdegspspol_id2c(id object)
{
    return [object ca_intobj_vardnsexpdegspspol_reference];
}

static id ca_intobj_vardnsexpdegspspol_t2id(id proto,ca_intobj_vardnsexpdegspspol_t f)
{
    return [proto ca_intobj_vardnsexpdegspspol_value:f];
}

static id ca_intobj_vardnsexpdegspspol_c2id(id proto,ca_intobj_vardnsexpdegspspol_c *r)
{
    return [proto ca_intobj_vardnsexpdegspspol_reference:r];
}

static int ca_intobj_vardnsexpdegspspol_isempty(ca_intobj_vardnsexpdegspspol_t self)
{
    if (self->nlink == self) {
	assert(self->plink == self);return YES;
    } else {
	assert(self->plink != self);return NO;
    }
}

static void ca_intobj_vardnsexpdegspspol_init(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    self->data = 0xcafebabe;self->nlink = self;self->plink = self;
    assert(ca_intobj_vardnsexpdegspspol_check(self,args));
}

static void ca_intobj_vardnsexpdegspspol_copy(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_copy(c,a->nlink,a,c,args);
}

static void ca_intobj_vardnsexpdegspspol_deepcopy(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_deepcopy(c,a->nlink,a,c,args);
}

static void ca_intobj_vardnsexpdegspspol_move(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    if (ca_intobj_vardnsexpdegspspol_isempty(a)) {
	ca_intobj_vardnsexpdegspspol_init(c,args);
    } else {
	ca_intobj_vardnsexpdegspspol_init(c,args);
	ca_intobj_vardnsexpdegspsterm_glue(c,a->nlink);ca_intobj_vardnsexpdegspsterm_glue(a->plink,c);
    }
    
    a->data = 0;assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_clear(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_clear(self->nlink,self,args);
    self->data = 0;ca_intobj_vardnsexpdegspslist_free(self->nlink,self,args);
}

static int ca_intobj_vardnsexpdegspspol_count(ca_intobj_vardnsexpdegspspol_t self)
{
    return ca_intobj_vardnsexpdegspslist_count(self->nlink,self);
}

static unsigned ca_intobj_vardnsexpdegspspol_hash(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    return ca_intobj_vardnsexpdegspslist_hash(a->nlink,a,args);
}

static int ca_intobj_vardnsexpdegspspol_iseq(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_t b,ca_intobj_vardnsexpdegspspol_args args)
{
    return (a==b)?YES:ca_intobj_vardnsexpdegspslist_iseq(a->nlink,a,b->nlink,b,args);
}

static int ca_intobj_vardnsexpdegspspol_isone(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    /* if the leading term is one, then it is one */
    return (ca_intobj_vardnsexpdegspspol_isempty(a))?0:ca_intobj_vardnsexpdegspsterm_isone(a->nlink,args);
}

static int ca_intobj_vardnsexpdegspspol_isminusone(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    /* if the leading term is minus one, then it is minus one */
    return (ca_intobj_vardnsexpdegspspol_isempty(a))?0:ca_intobj_vardnsexpdegspsterm_isminusone(a->nlink,args);
}

static int ca_intobj_vardnsexpdegspspol_deg(ca_intobj_vardnsexpdegspspol_t self ,ca_intobj_vardnsexpdegspspol_args args)
{
    return ca_intobj_vardnsexpdegspslist_deg(self->nlink,self,args);
}

static int ca_intobj_vardnsexpdegspspol_ord(ca_intobj_vardnsexpdegspspol_t self ,ca_intobj_vardnsexpdegspspol_args args)
{
    return ca_intobj_vardnsexpdegspslist_ord(self->nlink,self,args);
}

static int ca_intobj_vardnsexpdegspspol_maxdeg(ca_intobj_vardnsexpdegspspol_t self ,ca_intobj_vardnsexpdegspspol_args args)
{
    return ca_intobj_vardnsexpdegspslist_maxdeg(self->nlink,self,args);
}

static int ca_intobj_vardnsexpdegspspol_mindeg(ca_intobj_vardnsexpdegspspol_t self ,ca_intobj_vardnsexpdegspspol_args args)
{
    return ca_intobj_vardnsexpdegspslist_mindeg(self->nlink,self,args);
}

static ca_intobj_t ca_intobj_vardnsexpdegspspol_scalat(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_ixt ix,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_intobj_vardnsexpdegspsterm_scal(ix->ix,args);
}

static ca_vardnsmonom_t ca_intobj_vardnsexpdegspspol_monomat(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_ixt ix,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_ix_ismatch(ix) && ix->head == self);
    return ca_intobj_vardnsexpdegspsterm_monom(ix->ix);
}

static ca_intobj_t ca_intobj_vardnsexpdegspspol_scalfirst(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_isempty(self) == NO);
    return ca_intobj_vardnsexpdegspsterm_scal(self->nlink,args);
}

static ca_vardnsmonom_t ca_intobj_vardnsexpdegspspol_monomfirst(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_isempty(self) == NO);
    return ca_intobj_vardnsexpdegspsterm_monom(self->nlink);
}

static void ca_intobj_vardnsexpdegspspol_remove(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_check(self,args) && ca_intobj_vardnsexpdegspspol_isempty(self) == NO);
    ca_intobj_vardnsexpdegspsterm_delete(self->nlink,args);
    assert(ca_intobj_vardnsexpdegspspol_check(self,args));
}

static void ca_intobj_vardnsexpdegspspol_insert(ca_intobj_vardnsexpdegspspol_t self,ca_vardnsmonom_t m,ca_intobj_t s,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_check(self,args));
    ca_intobj_vardnsexpdegspslist_insert(self->nlink,self,m,s,args);
    assert(ca_intobj_vardnsexpdegspspol_check(self,args));
}

static int ca_intobj_vardnsexpdegspspol_iszero(ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    return ca_intobj_vardnsexpdegspspol_isempty(self);
}

static int ca_intobj_vardnsexpdegspspol_isop(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_t b,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(a != b);return ca_intobj_vardnsexpdegspslist_isop(a->nlink,a,b->nlink,b,args);
}

static void ca_intobj_vardnsexpdegspspol_ooneg(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_ooneg(c,a->nlink,a,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inneg(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inneg(a->nlink,a,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_oodbl(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_oodbl(c,a->nlink,a,v,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_indbl(ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_indbl(a->nlink,a,v,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_ooadd(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_t b,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(a!=b);
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_ooadd(c,a->nlink,a,v,b->nlink,b,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inadd(ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_t b,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(a!=b);
    ca_intobj_vardnsexpdegspslist_inadd(a->nlink,a,v,b->nlink,b,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_oomulsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_oomulsc(c,a->nlink,a,s,+1,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inmulsc(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inmulsc(a->nlink,a,s,+1,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_ooaddmulsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_t b,ca_intobj_t t,int w, ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_ooaddmulsc(c,a->nlink,a,v,b->nlink,b,t,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inaddmulsc(ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_t b,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inaddmulsc(a->nlink,a,v,b->nlink,b,t,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_oomulscadd(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspspol_t b,int w, ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_oomulscadd(c,a->nlink,a,s,v,b->nlink,b,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inmulscadd(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspspol_t b,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inmulscadd(a->nlink,a,s,v,b->nlink,b,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_oomulscaddmulsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspspol_t b,ca_intobj_t t,int w, ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_oomulscaddmulsc(c,a->nlink,a,s,v,b->nlink,b,t,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inmulscaddmulsc(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspspol_t b,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inmulscaddmulsc(a->nlink,a,s,v,b->nlink,b,t,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static BOOL ca_intobj_vardnsexpdegspspol_oodivsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsexpdegspspol_args args)
{
    BOOL failed;
    ca_intobj_vardnsexpdegspspol_init(c,args);
    failed = ca_intobj_vardnsexpdegspslist_oodivsc(c,a->nlink,a,s,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
    if (failed) ca_intobj_vardnsexpdegspspol_clear(c,args);
    return failed;
}

static void ca_intobj_vardnsexpdegspspol_indivsc(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_indivsc(a->nlink,a,s,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_oomulmonomsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_oomulmonomsc(c,b->nlink,b,m,t,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inmulmonomsc(ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inmulmonomsc(b->nlink,b,m,t,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(b,args));
}

static void ca_intobj_vardnsexpdegspspol_ooaddmulmonomsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,int w, ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_ooaddmulmonomsc(c,a->nlink,a,v,b->nlink,b,m,t,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inaddmulmonomsc(ca_intobj_vardnsexpdegspspol_t a,int v,ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inaddmulmonomsc(a->nlink,a,v,b->nlink,b,m,t,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static void ca_intobj_vardnsexpdegspspol_oomulscaddmulmonomsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,int w, ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(c,args);ca_intobj_vardnsexpdegspslist_oomulscaddmulmonomsc(c,a->nlink,a,s,v,b->nlink,b,m,t,w,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
}

static void ca_intobj_vardnsexpdegspspol_inmulscaddmulmonomsc(ca_intobj_vardnsexpdegspspol_t a,ca_intobj_t s,int v,ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,int w,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_inmulscaddmulmonomsc(a->nlink,a,s,v,b->nlink,b,m,t,w,args);
    assert(ca_intobj_vardnsexpdegspspol_check(a,args));
}

static BOOL ca_intobj_vardnsexpdegspspol_oodivmonomsc(ca_intobj_vardnsexpdegspspol_t c,ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t, ca_intobj_vardnsexpdegspspol_args args)
{
    BOOL failed;
    ca_intobj_vardnsexpdegspspol_init(c,args);
    failed = ca_intobj_vardnsexpdegspslist_oodivmonomsc(c,b->nlink,b,m,t,c,args);
    assert(ca_intobj_vardnsexpdegspspol_check(c,args));
    if (failed) ca_intobj_vardnsexpdegspspol_clear(c,args);
    return failed;
}

static void ca_intobj_vardnsexpdegspspol_indivmonomsc(ca_intobj_vardnsexpdegspspol_t b,ca_vardnsmonom_t m,ca_intobj_t t,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspslist_indivmonomsc(b->nlink,b,m,t,args);
    assert(ca_intobj_vardnsexpdegspspol_check(b,args));
}

static void ca_intobj_vardnsexpdegspspol_args_fileOut(id aFiler,ca_intobj_vardnsexpdegspspol_args *args)
{
    ca_intobj_vardnsexpdegspsterm_args_fileOut(aFiler,args);
}
static void ca_intobj_vardnsexpdegspspol_args_fileIn(id aFiler,ca_intobj_vardnsexpdegspspol_args *args)
{
    ca_intobj_vardnsexpdegspsterm_args_fileIn(aFiler,args);
}

static void ca_intobj_vardnsexpdegspspol_fileOut(id aFiler,ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    assert(ca_intobj_vardnsexpdegspspol_check(self,args));
    ca_intobj_vardnsexpdegspslist_fileOut(aFiler,self->nlink,self,args);
}

static void ca_intobj_vardnsexpdegspspol_fileIn(id aFiler,ca_intobj_vardnsexpdegspspol_t self,ca_intobj_vardnsexpdegspspol_args args)
{
    ca_intobj_vardnsexpdegspspol_init(self,args);ca_intobj_vardnsexpdegspslist_fileIn(aFiler,self,self,args);
    assert(ca_intobj_vardnsexpdegspspol_check(self,args));
}

static ca_intobj_vardnsexpdegspspol_args ca_intobj_vardnsexpdegspspol_getargs(id self)
{
    ca_intobj_vardnsexpdegspspol_args res;id monomialZero = [self monomialZero];
    res.sargs = ca_intobj_getargs([monomialZero scalar]);
    res.margs = ca_vardnsmonom_getargs([monomialZero terms]);
    return res; 
}
@implementation ca_intobj_vardnsexpdegsps_polynomial : ca_vardnsexpdegsps_polynomial

- (ca_intobj_vardnsexpdegspspol_t)ca_intobj_vardnsexpdegspspol_value
{
    return &value;
}

- ca_intobj_vardnsexpdegspspol_value:(ca_intobj_vardnsexpdegspspol_t)aValue
{
    self = [self clone];
    ca_intobj_vardnsexpdegspspol_copy(&value,aValue,ca_intobj_vardnsexpdegspspol_getargs(self));
    assert([self check]);return self;
}

- (ca_intobj_vardnsexpdegspspol_t)ca_intobj_vardnsexpdegspspol_reference
{
    return &value;
}

- ca_intobj_vardnsexpdegspspol_reference:(ca_intobj_vardnsexpdegspspol_c *)aReference
{
    self = [self clone];
    ca_intobj_vardnsexpdegspspol_move(&value,aReference,ca_intobj_vardnsexpdegspspol_getargs(self));
    assert([self check]);return self;
}

- copy
{
    ca_intobj_vardnsexpdegspspol_c c;assert([self check]);
    ca_intobj_vardnsexpdegspspol_copy(&c,&value,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- deepCopy
{
    ca_intobj_vardnsexpdegspspol_c c;assert([self check]);
    ca_intobj_vardnsexpdegspspol_deepcopy(&c,&value,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- (BOOL) isEmpty
{
    return ca_intobj_vardnsexpdegspspol_isempty(&value);
}

- (BOOL) isOne
{
    return ca_intobj_vardnsexpdegspspol_isone(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (BOOL) isMinusOne
{
    return ca_intobj_vardnsexpdegspspol_isminusone(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (unsigned) hash
{
    return ca_intobj_vardnsexpdegspspol_hash(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_intobj_vardnsexpdegspspol_iseq(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_id2t(b),ca_intobj_vardnsexpdegspspol_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_intobj_vardnsexpdegspspol_iseq(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_id2t(b),ca_intobj_vardnsexpdegspspol_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_intobj_vardnsexpdegspspol_iszero(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_intobj_vardnsexpdegspspol_iszero(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_intobj_vardnsexpdegspspol_isop(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_id2t(b),ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_intobj_vardnsexpdegspspol_isop(ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_id2t(b),ca_intobj_vardnsexpdegspspol_getargs(self));
}

- negate
{
    ca_intobj_vardnsexpdegspspol_c c;assert([self check]);
    ca_intobj_vardnsexpdegspspol_ooneg(&c,ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_intobj_vardnsexpdegspspol_inneg(ca_intobj_vardnsexpdegspspol_id2c(self),ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_intobj_vardnsexpdegspspol_c c;assert([self check]);
    ca_intobj_vardnsexpdegspspol_oodbl(&c,ca_intobj_vardnsexpdegspspol_id2t(self),v,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_intobj_vardnsexpdegspspol_indbl(ca_intobj_vardnsexpdegspspol_id2c(self),v,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- double
{
    ca_intobj_vardnsexpdegspspol_c c;assert([self check]);
    ca_intobj_vardnsexpdegspspol_oodbl(&c,ca_intobj_vardnsexpdegspspol_id2t(self),+1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_intobj_vardnsexpdegspspol_indbl(ca_intobj_vardnsexpdegspspol_id2c(self),+1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_intobj_vardnsexpdegspspol_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_intobj_vardnsexpdegspspol_ooadd(&c,ca_intobj_vardnsexpdegspspol_id2t(self),v,ca_intobj_vardnsexpdegspspol_id2t(b),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_intobj_vardnsexpdegspspol_inadd(ca_intobj_vardnsexpdegspspol_id2c(self),v,ca_intobj_vardnsexpdegspspol_id2t(b),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_intobj_vardnsexpdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobj_vardnsexpdegspspol_ooadd(&c,ca_intobj_vardnsexpdegspspol_id2t(self),+1,ca_intobj_vardnsexpdegspspol_id2t(b),+1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_intobj_vardnsexpdegspspol_inadd(ca_intobj_vardnsexpdegspspol_id2c(self),+1,ca_intobj_vardnsexpdegspspol_id2t(b),+1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_intobj_vardnsexpdegspspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobj_vardnsexpdegspspol_ooadd(&c,ca_intobj_vardnsexpdegspspol_id2t(self),+1,ca_intobj_vardnsexpdegspspol_id2t(b),-1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_intobj_vardnsexpdegspspol_inadd(ca_intobj_vardnsexpdegspspol_id2c(self),+1,ca_intobj_vardnsexpdegspspol_id2t(b),-1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- (int) leadingDegree
{
    return ca_intobj_vardnsexpdegspspol_deg(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (int) lastDegree
{
    return ca_intobj_vardnsexpdegspspol_ord(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (int) maxDegree
{
    return ca_intobj_vardnsexpdegspspol_maxdeg(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
}

- (int) minDegree
{
    return ca_intobj_vardnsexpdegspspol_mindeg(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
}

- asScalar
{
    if ([self size] == 1) {
	id asScalar,lastTerm,aSequence = [self eachElement];
	asScalar = (lastTerm = [aSequence lastElement])?[lastTerm asScalar]:nil;
	[aSequence free];return asScalar;
    } else {
	return nil;
    }
}

- asSymbol
{
    if ([self size] == 1) {
	id asSymbol,lastTerm,aSequence = [self eachElement];
	asSymbol = (lastTerm = [aSequence lastElement])?[lastTerm asSymbol]:nil;
	[aSequence free];return asSymbol;
    } else {
	return nil;
    }
}

- check
{
    [super check];
    assert(scalarZero && termZero && monomialZero && symbols);
    [scalarZero check];[termZero check];[monomialZero check];
    [symbols check];assert([symbols size] >= 1);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([symbols isEqual:[monomialZero symbols]]);

    ca_intobj_vardnsexpdegspspol_check(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
    return self;
}

- _setUpScalarZero:aScalar symbols:aCltn
{
    assert([aScalar isZero] && [aCltn notEmpty]);
    assert([aScalar class_vardnsexpdegsps_polynomial] == [self class]);

    aScalar  = [aScalar cheapCopy];
    termZero = [Term scalar:aScalar symbol:nil exponent:0];
    aScalar  = [aScalar cheapCopy];aCltn = [aCltn cheapCopy];
    monomialZero = [Monomial scalar:aScalar symbols:aCltn];

    scalarZero = aScalar;
    symbols    = aCltn;

    ca_intobj_vardnsexpdegspspol_init(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
    assert([self check]);return self;
}

+ scalarZero:aScalar symbols:aCltn
{
    return [[super new] _setUpScalarZero:aScalar symbols:aCltn];
}

- clear
{
    ca_intobj_vardnsexpdegspspol_clear(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
    /* free args after clearing polynomial value */
    scalarZero = [scalarZero free];
    termZero = [termZero free];monomialZero = [monomialZero free];
    symbols = [symbols free];return [super clear];
}

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    symbols      = [symbols cheapCopy];
    return self;
}

- empty
{
    ca_intobj_vardnsexpdegspspol_c c;assert([self check]);
    ca_intobj_vardnsexpdegspspol_init(&c,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- scalarZero   { return scalarZero;   }
- termZero     { return termZero;     }
- monomialZero { return monomialZero; }
- symbols      { return symbols;      }

- (int) numMonomials
{
    assert([self check]);return ca_intobj_vardnsexpdegspspol_count(&value);
}

- eachMonomial
{
    id aSequence = [ca_intobj_vardnsexpdegspspol_sequence content:[self cheapCopy]];
    return [CASequence over:aSequence];
}

- elt_ca_intobj_vardnsexpdegspspol_ix:(ca_intobj_vardnsexpdegspspol_ixt)ix
{
    if (ca_intobj_vardnsexpdegspspol_ix_ismatch(ix)) {
	id s,m;ca_intobj_t sc;ca_vardnsmonom_t mc;
	sc = ca_intobj_vardnsexpdegspspol_scalat(&value,ix,ca_intobj_vardnsexpdegspspol_getargs(self));
	mc = ca_intobj_vardnsexpdegspspol_monomat(&value,ix,ca_intobj_vardnsexpdegspspol_getargs(self));
	s  = ca_intobj_t2id([monomialZero scalar],sc);
	m  = ca_vardnsmonom_t2id([monomialZero terms],mc);
	return [monomialZero scalar:s terms:m];
    } else {
	return nil;
    }
}

- removeMonomial
{
    if ([self isEmpty]) {
	return nil;
    } else {
	id lt;
	id m,s;ca_intobj_t sc;ca_vardnsmonom_t mc;
	[self checkRefCountOne];assert([self check]);
	
	sc = ca_intobj_vardnsexpdegspspol_scalfirst(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
	mc = ca_intobj_vardnsexpdegspspol_monomfirst(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
	s  = ca_intobj_t2id([monomialZero scalar],sc);
	m  = ca_vardnsmonom_t2id([monomialZero terms],mc);
	lt = [monomialZero scalar:s terms:m];
	
	ca_intobj_vardnsexpdegspspol_remove(&value,ca_intobj_vardnsexpdegspspol_getargs(self));
	[self invalidate];assert([self check]);return lt;
    }
}

- insertMonomial:aMonomial
{
    if ([aMonomial isZero])  {
	[aMonomial free];return self;
    } else {
	ca_intobj_c c;ca_intobj_t r;ca_vardnsmonom_t m;
	assert([aMonomial check]);[self checkRefCountOne];assert([self check]);
	r = ca_intobj_id2ref([aMonomial scalar],&c);
	m = ca_vardnsmonom_id2t([aMonomial terms]);
	ca_intobj_vardnsexpdegspspol_insert(&value,m,r,ca_intobj_vardnsexpdegspspol_getargs(self));
	assert([aMonomial check]);[aMonomial free];
	[self invalidate];assert([self check]);return self;
    }
}

- multiplyScalar:s
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c sc;
    ca_intobj_vardnsexpdegspspol_oomulsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- multiplySelfScalar:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self multiplyScalar:s]);
    ca_intobj_vardnsexpdegspspol_inmulsc(ca_intobj_vardnsexpdegspspol_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- divideScalar:s
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c sc;
    BOOL failed = ca_intobj_vardnsexpdegspspol_oodivsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsexpdegspspol_getargs(self));
    return (failed)?nil:ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- divideSelfScalar:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self divideScalar:s]);
    ca_intobj_vardnsexpdegspspol_indivsc(ca_intobj_vardnsexpdegspspol_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyScalar:b:(int)w
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c bc;
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobj_vardnsexpdegspspol_ooaddmulsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),v,ca_intobj_vardnsexpdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyScalar:b:(int)w
{
    ca_intobj_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyScalar:b:w]);
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobj_vardnsexpdegspspol_inaddmulsc(ca_intobj_vardnsexpdegspspol_id2c(self),v,ca_intobj_vardnsexpdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B:(int)w
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c ac;
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_intobj_vardnsexpdegspspol_oomulscadd(&c,ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsexpdegspspol_id2t(B),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B:(int)w
{
    ca_intobj_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_intobj_vardnsexpdegspspol_inmulscadd(ca_intobj_vardnsexpdegspspol_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsexpdegspspol_id2t(B),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c ac,bc;
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobj_vardnsexpdegspspol_oomulscaddmulsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsexpdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_intobj_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B multiplyScalar:b:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobj_vardnsexpdegspspol_inmulscaddmulsc(ca_intobj_vardnsexpdegspspol_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobj_vardnsexpdegspspol_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}
- multiplyMonomial:aMonomial
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    ca_vardnsmonom_t m;
    
    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_intobj_id2ref([aMonomial scalar],&sc);
    ca_intobj_vardnsexpdegspspol_oomulmonomsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),m,s,+1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- multiplySelfMonomial:aMonomial
{
    ca_intobj_t s;
    ca_intobj_c sc;
    ca_vardnsmonom_t m;
    if (refCount>1) return idReplace(self,[self multiplyMonomial:aMonomial]);

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_intobj_id2ref([aMonomial scalar],&sc);
    ca_intobj_vardnsexpdegspspol_inmulmonomsc(ca_intobj_vardnsexpdegspspol_id2c(self),m,s,+1,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyMonomial:aMonomial:(int)v
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    ca_vardnsmonom_t m;

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_intobj_id2ref([aMonomial scalar],&sc);
    ca_intobj_vardnsexpdegspspol_oomulmonomsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),m,s,v,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfMonomial:aMonomial:(int)v
{
    ca_intobj_t s;
    ca_intobj_c sc;
    ca_vardnsmonom_t m;
    if (refCount>1) return idReplace(self,[self _multiplyMonomial:aMonomial:v]);

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_intobj_id2ref([aMonomial scalar],&sc);
    ca_intobj_vardnsexpdegspspol_inmulmonomsc(ca_intobj_vardnsexpdegspspol_id2c(self),m,s,v,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- divideMonomial:aMonomial
{
    BOOL failed;
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_t s;
    ca_intobj_c sc;
    ca_vardnsmonom_t m;

    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_intobj_id2ref([aMonomial scalar],&sc);
    failed = ca_intobj_vardnsexpdegspspol_oodivmonomsc(&c,ca_intobj_vardnsexpdegspspol_id2t(self),m,s,ca_intobj_vardnsexpdegspspol_getargs(self));
    return (failed)?nil:ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- divideSelfMonomial:aMonomial
{
    ca_intobj_t s;
    ca_intobj_c sc;
    ca_vardnsmonom_t m;
    if (refCount>1) return idReplace(self,[self divideMonomial:aMonomial]);
    
    m = ca_vardnsmonom_id2t([aMonomial terms]);
    s = ca_intobj_id2ref([aMonomial scalar],&sc);
    ca_intobj_vardnsexpdegspspol_indivmonomsc(ca_intobj_vardnsexpdegspspol_id2c(self),m,s,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyMonomial:b:(int)w
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c bc;
    ca_intobj_t bs;
    ca_vardnsmonom_t bm;

    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_intobj_id2ref([b scalar],&bc);
    ca_intobj_vardnsexpdegspspol_ooaddmulmonomsc(&c,ca_intobj_vardnsexpdegspspol_id2c(self),v,ca_intobj_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyMonomial:b:(int)w
{
    ca_intobj_c bc;
    ca_intobj_t bs;
    ca_vardnsmonom_t bm;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyMonomial:b:w]);
    
    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_intobj_id2ref([b scalar],&bc);
    ca_intobj_vardnsexpdegspspol_inaddmulmonomsc(ca_intobj_vardnsexpdegspspol_id2c(self),v,ca_intobj_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B multiplyMonomial:b:(int)w
{
    ca_intobj_vardnsexpdegspspol_c c;
    ca_intobj_c ac,bc;
    ca_intobj_t as,bs;
    ca_vardnsmonom_t bm;
    
    as = ca_intobj_id2ref(a,&ac);
    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_intobj_id2ref([b scalar],&bc);
    ca_intobj_vardnsexpdegspspol_oomulscaddmulmonomsc(&c,ca_intobj_vardnsexpdegspspol_id2c(self),as,v,ca_intobj_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return ca_intobj_vardnsexpdegspspol_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B multiplyMonomial:b:(int)w
{
    ca_intobj_c ac,bc;
    ca_intobj_t as,bs;
    ca_vardnsmonom_t bm;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B multiplyMonomial:b:w]);
    
    as = ca_intobj_id2ref(a,&ac);
    bm = ca_vardnsmonom_id2t([b terms]);
    bs = ca_intobj_id2ref([b scalar],&bc);
    ca_intobj_vardnsexpdegspspol_inmulscaddmulmonomsc(ca_intobj_vardnsexpdegspspol_id2c(self),as,v,ca_intobj_vardnsexpdegspspol_id2t(B),bm,bs,w,ca_intobj_vardnsexpdegspspol_getargs(self));
    return [self invalidate];
}

- fileOutOn:aFiler
{
    ca_intobj_vardnsexpdegspspol_args args;
    assert([self check]);
    [super fileOutOn:aFiler];
    args = ca_intobj_vardnsexpdegspspol_getargs(self);
    ca_intobj_vardnsexpdegspspol_args_fileOut(aFiler,&args);
    ca_intobj_vardnsexpdegspspol_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_intobj_vardnsexpdegspspol_args args;
    [super fileInFrom:aFiler];
    ca_intobj_vardnsexpdegspspol_args_fileIn(aFiler,&args);
    ca_intobj_vardnsexpdegspspol_fileIn(aFiler,&value,args);
    return self;
}

@end

