
/*
 * Computer Algebra Kit (c) 1993,98 by Comp.Alg.Objects.  All Rights Reserved.
 */

#include "cakit.h"

static ca_modp_args ca_modp_getargs(id self)
{
   ca_modp_args args;args.p = [self modulus];return args;
}

static int ca_modp_size(ca_modp_args args)
{
    return 1;				/* pointer increment */
}

static int ca_modp_bsize(ca_modp_args args)
{
    return sizeof(ca_modp_c);		/* bytes */
}

static ca_modp_t ca_modp_id2t(id object)
{
    return [object ca_modp_value];	/* read-only */
}

static ca_modp_t ca_modp_id2ref(id object,ca_modp_c *ref)
{
    return [object ca_modp_value];	/* read-only, scratch space "ref" */
}

static ca_modp_c *ca_modp_id2c(id object)
{
    return [object ca_modp_reference];	/* write for inplace arithmetic */
}

static id ca_modp_t2id(id proto,ca_modp_t f)
{
    return [proto ca_modp_value:f];	/* new object containing copy of "f" */
}

static id ca_modp_c2id(id proto,ca_modp_c* c)
{
    return [proto ca_modp_reference:c];/* new object containing "c" */
}

static ca_modp_t ca_modp_fromintvalue(int i,ca_modp_args args)
{
    ca_modp_t red;
    
    if (i==0) return 0;
    if (i <0) {
	red = -i;
	red %= args.p;	/* it's dangerous to % int & unsigned */
	red = (red)?args.p-red:0;
    } else {
	red  = i;
	red %= args.p;	/* dito */
    }
    
    return red;
}

static int ca_modp_check(ca_modp_t self,ca_modp_args args)
{
    assert(self < args.p);return 1;
}

static ca_modp_t ca_modp_zero(ca_modp_args args)
{
    return 0;
}

static ca_modp_t ca_modp_one(ca_modp_args args)
{
    return 1;
}

static unsigned ca_modp_hash(ca_modp_t a,ca_modp_args args)
{
    return a;
}

static int ca_modp_iseq(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a == b);
}

static int ca_modp_isop(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a==0)?(b==0):(a+b==args.p);
}

static int ca_modp_iszero(ca_modp_t a,ca_modp_args args)
{
    return a==0;
}

static int ca_modp_isone(ca_modp_t a,ca_modp_args args)
{
    return a==1;
}

static int ca_modp_isminusone(ca_modp_t a,ca_modp_args args)
{
    return (a+1)==args.p;
}

static ca_modp_t ca_modp_neg(ca_modp_t a,ca_modp_args args)
{
    assert(ca_modp_check(a,args));return (a)?args.p-a:0;
}

static ca_modp_t ca_modp_dbl(ca_modp_t a,ca_modp_args args)
{
    ca_modp_t c = a+a;
    assert(args.p != 2 && ca_modp_check(a,args));
    if (c) return (c<args.p)?c:c-args.p; else return 0;
}

static ca_modp_t ca_modp_add(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    ca_modp_t c = a+b;
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return ((c>=args.p)?c-args.p:c); /* compiles better than c<p?c:c-p */
}

static ca_modp_t ca_modp_sub(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a>=b)?(a-b):(args.p-(b-a));
}

static ca_modp_t ca_modp_mul(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    unsigned c = a * b;		/* limitation to 16bits to be able to do this
				 * russian peasant 50% slower
                                 */
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_sqr(ca_modp_t a,ca_modp_args args)
{
    unsigned c = a * a;
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_inv(ca_modp_t x,ca_modp_args args)
{
    unsigned q,s;
    int u1,u2,us;  /* intermediate values can be negative */
    ca_modp_t a,b,inv;
    
    assert(x != 0 && x != 1 && x+1 != args.p);
    
    a = x;b = args.p;u1 = 1;u2 = 0;
    while (b) {
	s  = a%b;q = a/b;
	a  = b;b = s;
	u1 -= q*u2;			
	us  = u1;u1 = u2;u2 = us;
    }
    
    inv = ((u1<0)?u1+args.p:u1); /* keep result positive */
    
    assert(ca_modp_check(inv,args));
    assert(ca_modp_isone(ca_modp_mul(x,inv,args),args));
    return inv;
}

static ca_modp_t ca_modp_div(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(b!=0);
    return ca_modp_mul(a,ca_modp_inv(b,args),args);
}

static ca_modp_t ca_modp_pow(ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_t o;
    
    if (n==0) return ca_modp_one(args);
    if (n==1) return a;
    if (n==2) return ca_modp_sqr(a,args);
    
    o = ca_modp_one(args);
    while (n>1) {
	if (n&1) o = ca_modp_mul(o,a,args);
	a = ca_modp_sqr(a,args);
	n >>= 1;
    }
    
    return ca_modp_mul(o,a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_copy_val.fab should not be included"
#endif

static void ca_modp_clear(ca_modp_c *c,ca_modp_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_modp_t)0xcafebabe;
    #endif
}

static void ca_modp_copy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_deepcopy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_move(ca_modp_c *c,ca_modp_c *a,ca_modp_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_modp_t)0xcafebabe;
    #endif
}

#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static void ca_modp_oozero(ca_modp_c *c,ca_modp_args args)
{
    *c = ca_modp_zero(args);
}

static void ca_modp_inzero(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_zero(args);
}

static void ca_modp_inneg(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_neg(*a,args);
}

static void ca_modp_ooneg(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_neg(a,args);
}

static ca_modp_t ca_modp_dblv(ca_modp_t a,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_dbl(a,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_indbl(ca_modp_c *a,int v,ca_modp_args args)
{
    *a = ca_modp_dblv(*a,v,args);
}

static void ca_modp_oodbl(ca_modp_c *c,ca_modp_t a,int v,ca_modp_args args)
{
    *c = ca_modp_dblv(a,v,args);
}

static ca_modp_t ca_modp_addvw(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    ca_modp_t c;assert(v*v == +1 && w*w == 1);
    
    if (v == +1) {
	if (w == +1) {
	    c = ca_modp_add(a,b,args);
	} else {
	    c = ca_modp_sub(a,b,args);
	}
    } else {
	if (w == +1) {
	    c = ca_modp_sub(b,a,args);
	} else {
	    c = ca_modp_neg(ca_modp_add(a,b,args),args);
	}
    }
    
    return c;
}

static void ca_modp_inadd(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addvw(*a,v,b,w,args);
}

static void ca_modp_ooadd(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addvw(a,v,b,w,args);
}
#if !ca_modp_isvalue
#error "ca_modp_mul_val.fab should not be included"
#endif

static void ca_modp_insqr(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_sqr(*a,args);
}

static void ca_modp_oosqr(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_sqr(a,args);
}

static void ca_modp_inmul(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_mul(*a,b,args);
}

static void ca_modp_oomul(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_mul(a,b,args);
}
#if !ca_modp_isvalue
#error "ca_modp_inv_val.fab should not be included"
#endif

static void ca_modp_ininv(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_inv(*a,args);
}

static void ca_modp_ooinv(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_inv(a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_div_val.fab should not be included"
#endif

static void ca_modp_indiv(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_div(*a,b,args);
}

static BOOL ca_modp_oodiv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_div(a,b,args);return SUCCESS;
}
#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static ca_modp_t ca_modp_mulv(ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_mul(a,b,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_inmulv(ca_modp_c *a,ca_modp_t b,int v,ca_modp_args args)
{
    *a = ca_modp_mulv(*a,b,v,args);
}

static void ca_modp_oomulv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    *c = ca_modp_mulv(a,b,v,args);
}

static ca_modp_t ca_modp_addmul(ca_modp_t a,int v,ca_modp_t b,ca_modp_t s,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_mul(b,s,args),w,args);
}

static ca_modp_t ca_modp_addsqr(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_sqr(b,args),w,args);
}

static ca_modp_t ca_modp_muladd(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,b,w,args);
}

static ca_modp_t ca_modp_muladdmul(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,ca_modp_mul(b,t,args),w,args);
}

static void ca_modp_inaddmul(ca_modp_c *a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_addmul(*a,v,b,t,w,args);
}

static void ca_modp_ooaddmul(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_addmul(a,v,b,t,w,args);
}

static void ca_modp_inaddsqr(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addsqr(*a,v,b,w,args);
}

static void ca_modp_ooaddsqr(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addsqr(a,v,b,w,args);
}

static void ca_modp_inmuladd(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_muladd(*a,s,v,b,w,args);
}

static void ca_modp_oomuladd(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_muladd(a,s,v,b,w,args);
}

static void ca_modp_inmuladdmul(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_muladdmul(*a,s,v,b,t,w,args);
}

static void ca_modp_oomuladdmul(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_muladdmul(a,s,v,b,t,w,args);
}
static void ca_modp_args_fileOut(id aFiler,ca_modp_args *args) {}
static void ca_modp_args_fileIn(id aFiler,ca_modp_args *args) {}

static void ca_modp_fileOut(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileOut:c type:'S'];
}

static void ca_modp_fileIn(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileIn:c type:'S'];
}

@implementation IntegerModp

+ new
{
    return [self error:"You can't create a IntegerModp without modulus"];
}

+ modulus:(unsigned short)aPrime
{
    return [self int:0 modulus:aPrime];
}

- _setUpIntValue:(int)intValue modulus:(unsigned short)aPrime
{
    modulus = aPrime;
    value   = ca_modp_fromintvalue(intValue,ca_modp_getargs(self));
    return self;
}

- int:(int)intValue modulus:(unsigned short)aPrime
{
    return [self notImplemented:_cmd];
}

+ int:(int)intValue modulus:(unsigned short)aPrime
{
    unsigned i,m = aPrime;
    
    if (m==2) [self error:"Modulus Should Not Be Equal To 2"];
    if (m>USHRT_MAX_PRIME) [self error:"Modulus Too Big"];

    if (m != USHRT_MAX_PRIME) {
    for(i=2;i<=m/2;i++) if (m%i == 0) return [self error:"Modulus Not Prime"];
    }
    
    return [[super new] _setUpIntValue:intValue modulus:aPrime];
}

- class_vector
{
    return (id)[ca_modp_vector class];
}

- class_vardnsrecdegsps_polynomial
{
    return (id)[ca_modp_vardnsrecdegsps_polynomial class];
}

- class_vardnsrecdegdns_polynomial
{
    return (id)[ca_modp_vardnsrecdegdns_polynomial class];
}

- class_vardnsexpdegsps_polynomial
{
    return (id)[ca_modp_vardnsexpdegsps_polynomial class];
}

- clone
{
    self = [super clone];value = 0;assert(modulus);return self;
}

- (ca_modp_t) ca_modp_value
{
    return value;
}

- ca_modp_value:(ca_modp_t)aValue
{
    self = [self clone];value = aValue;return self;
}

- (ca_modp_t *) ca_modp_reference
{
    return &value;
}

- ca_modp_reference:(ca_modp_t *)aValue
{
    self = [self clone];value = *aValue;return self;
}

- (BOOL)sameClass:b			{ return isa == [b class] && modulus == [b modulus]; }
- (BOOL)inAdditiveSemiGroup		{ return YES; }
- (BOOL)inAdditiveMonoid		{ return YES; }
- (BOOL)inAdditiveGroup			{ return YES; }

- (BOOL)inSemiGroup			{ return YES; }
- (BOOL)inMonoid			{ return YES; }
- (BOOL)inGroup				{ return YES;  }

- (BOOL)inEuclideanDomain		{ return YES; }
- (BOOL)inIntegralDomain		{ return YES; }
- (BOOL)inField				{ return YES; }

- (BOOL)inOrderedSet			{ return NO; }



- (unsigned) hash
{
    return ca_modp_hash(ca_modp_id2t(self),ca_modp_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_modp_iseq(ca_modp_id2t(self),ca_modp_id2t(b),ca_modp_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_modp_iseq(ca_modp_id2t(self),ca_modp_id2t(b),ca_modp_getargs(self));
    }
}

- (unsigned short) modulus
{
    return modulus;
}

- (int) intValue
{
    return value;
}

- intValue:(int)aValue
{
    ca_modp_t a = ca_modp_fromintvalue(aValue,ca_modp_getargs(self));
    return [self ca_modp_value:a];
}


- zero
{
    return [self intValue:0];
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_modp_iszero(ca_modp_id2t(self),ca_modp_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return NO == ca_modp_iszero(ca_modp_id2t(self),ca_modp_getargs(self));
}

- (BOOL) isOpposite:b
{
    [self checkSameClass:b];
    
    if (self == b) {
	return [self isZero] || [self isCharacteristicTwo];
    } else {
	return ca_modp_isop(ca_modp_id2t(self),ca_modp_id2t(b),ca_modp_getargs(self));
    }
}

- (BOOL) notOpposite:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self notZero] && [self notCharacteristicTwo];
    } else {
	return NO == ca_modp_isop(ca_modp_id2t(self),ca_modp_id2t(b),ca_modp_getargs(self));
    }
}

- negate
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self cheapCopy];
    } else {
	ca_modp_c c;assert([self check]);
	ca_modp_ooneg(&c,ca_modp_id2t(self),ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return self;
    } else {
	assert([self check]);
	ca_modp_inneg(ca_modp_id2c(self),ca_modp_getargs(self));
	return [self invalidate];
    }
}

- _double:(int)v
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self zero];
    } else {
	ca_modp_c c;assert([self check]);
	ca_modp_oodbl(&c,ca_modp_id2t(self),v,ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return idReplace(self,[self zero]);
    } else {
	assert([self check]);
	ca_modp_indbl(ca_modp_id2c(self),v,ca_modp_getargs(self));
	return [self invalidate];
    }
}

- double
{
    if ([self isZero] || [self isCharacteristicTwo]) {
	return [self zero];
    } else {
	ca_modp_c c;assert([self check]);
	ca_modp_oodbl(&c,ca_modp_id2t(self),+1,ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);

    if ([self isZero] || [self isCharacteristicTwo]) {
	return idReplace(self,[self zero]);
    } else {
	assert([self check]);
	ca_modp_indbl(ca_modp_id2c(self),+1,ca_modp_getargs(self));
	return [self invalidate];
    }
}

- _add:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1);

    if (self == b) {
	return (v==w)?[self _double:v]:[self zero];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b _negate:w];
	    if ([b isZero]) return [self _negate:v];
	}
	if (1) {
	    ca_modp_c c;
	    ca_modp_ooadd(&c,ca_modp_id2t(self),v,ca_modp_id2t(b),w,ca_modp_getargs(self));
	    return ca_modp_c2id(self,&c);
	}
    }

    return nil; /* to quiet some compilers, e.g. WATCOM */
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);

    if (self == b) {
	return (v==w)?[self _doubleSelf:v]:idReplace(self,[self zero]);
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return [self _negateSelf:v];
	    if ([self isZero]) return idReplace(self,[b _negate:w]);
	}
	ca_modp_inadd(ca_modp_id2c(self),v,ca_modp_id2t(b),w,ca_modp_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- add:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self double];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b cheapCopy];
	    if ([b isZero]) return [self cheapCopy];
	}
	if (1) {
	    ca_modp_c c;
	    ca_modp_ooadd(&c,ca_modp_id2t(self),+1,ca_modp_id2t(b),+1,ca_modp_getargs(self));
	    return ca_modp_c2id(self,&c);
	}
    }

    return nil;
}

- addSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self add:b]);

    if (self == b) {
	return [self doubleSelf];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return self;
	    if ([self isZero]) return idReplace(self,[b cheapCopy]);
	}
	ca_modp_inadd(ca_modp_id2c(self),+1,ca_modp_id2t(b),+1,ca_modp_getargs(self));
	return [self invalidate];
    }

    return nil;
}

- subtract:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self zero];
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return [b negate];
	    if ([b isZero]) return [self cheapCopy];
	}
	if (1) {
	    ca_modp_c c;
	    ca_modp_ooadd(&c,ca_modp_id2t(self),+1,ca_modp_id2t(b),-1,ca_modp_getargs(self));
	    return ca_modp_c2id(self,&c);
	}
    }

    return nil;
}

- subtractSelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self subtract:b]);

    if (self == b) {
	return idReplace(self,[self zero]);
    } else {
	if ([self inAdditiveMonoid]) {
	    if ([b isZero]) return self;
	    if ([self isZero]) return idReplace(self,[b negate]);
	}
	ca_modp_inadd(ca_modp_id2c(self),+1,ca_modp_id2t(b),-1,ca_modp_getargs(self));
	return [self invalidate];
    }

    return nil;
}


- one
{
    return [self intValue:+1];
}

- minusOne
{
    return [self intValue:-1];
}

- (BOOL) isOne
{
    assert([self check]);
    return ca_modp_isone(ca_modp_id2t(self),ca_modp_getargs(self));
}

- (BOOL) notOne
{
    assert([self check]);
    return NO == ca_modp_isone(ca_modp_id2t(self),ca_modp_getargs(self));
}

- (BOOL) isMinusOne
{
    assert([self check]);
    return ca_modp_isminusone(ca_modp_id2t(self),ca_modp_getargs(self));
}

- (BOOL) notMinusOne
{
    assert([self check]);
    return NO == ca_modp_isminusone(ca_modp_id2t(self),ca_modp_getargs(self));
}

- square
{
    if ([self inMonoid] && [self isOne]) return [self cheapCopy];
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return [self one];
	if ([self isZero])     return [self cheapCopy];
    }
    if (1) {
	ca_modp_c c;assert([self check]);
	ca_modp_oosqr(&c,ca_modp_id2t(self),ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }

    return nil;
}

- squareSelf
{
    if (refCount>1) return idReplace(self,[self square]);

    if ([self inMonoid] && [self isOne]) return self;
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return idReplace(self,[self one]);
	if ([self isZero])     return self;
    }
    
    assert([self check]);
    ca_modp_insqr(ca_modp_id2c(self),ca_modp_getargs(self));
    return [self invalidate];
}

- multiply:b
{
    [self checkSameClass:b];

    if (self == b) {
	return [self square];
    } else {
	if ([self inMonoid]) {
	    if ([self isOne]) return [b cheapCopy];
	    if ([b isOne]) return [self cheapCopy];
	}
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]||[b isZero]) return [self zero];
	    if ([self isMinusOne]) return [b negate];
	    if ([b isMinusOne]) return [self negate];
	}
	if (1) {
	    ca_modp_c c;
	    ca_modp_oomul(&c,ca_modp_id2t(self),ca_modp_id2t(b),ca_modp_getargs(self));
	    return ca_modp_c2id(self,&c);
	}
    }

    return nil;
}

- multiplySelf:b
{
    assert([self sameClass:b]);
    if (refCount>1) return idReplace(self,[self multiply:b]);

    if (self == b) {
	return [self squareSelf];
    } else {
	if ([self inMonoid]) {
	    if ([b isOne]) return self;
	    if ([self isOne]) return idReplace(self,[b cheapCopy]);
	}
	if ([self inAdditiveMonoid]) {
	    if ([self isZero]) return self;
	    if ([b isZero]) return idReplace(self,[b cheapCopy]);
	    if ([b isMinusOne]) return [self negateSelf];
	    if ([self isMinusOne]) return idReplace(self,[b negate]);
	}
	ca_modp_inmul(ca_modp_id2c(self),ca_modp_id2t(b),ca_modp_getargs(self));
	return [self invalidate];
    }

    return nil;
}


- inverse
{
    if ([self inMonoid] && [self isOne]) return [self cheapCopy];
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return [self cheapCopy];
	if ([self isZero])     return nil;
    }
    if (1) {
	ca_modp_c c;assert([self check]);
	ca_modp_ooinv(&c,ca_modp_id2t(self),ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }

    return nil;
}

- inverseSelf
{
    if (refCount>1) return idReplace(self,[self inverse]);

    if ([self inMonoid] && [self isOne]) return self;
    if ([self inAdditiveMonoid]) {
	if ([self isMinusOne]) return self;
	if ([self isZero])     return nil;
    }
    
    assert([self check]);
    ca_modp_ininv(ca_modp_id2c(self),ca_modp_getargs(self));
    return [self invalidate];
}

- divide:b
{
    if ([b isZero]) return nil;
    if (self == b) {
	return [self one];
    } else {
	id tmp = [b inverse];
	self = [self multiply:tmp];
	tmp = [tmp free];return self;
    }
}

- _add:(int)v:B multiply:b:(int)w
{
    assert([self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _multiply:b:+w];
    if ([B isZero]) return [self _negate:v];
    if ([b isZero]) return [self _negate:v];

    if ([b isOne]) return [self _add:v:B:+w];
    if ([B isOne]) return [self _add:v:b:+w];
    if ([b isMinusOne]) return [self _add:v:B:-w];
    if ([B isMinusOne]) return [self _add:v:b:-w];

    if (1) {
	ca_modp_c c;
	ca_modp_ooaddmul(&c,ca_modp_id2t(self),v,ca_modp_id2t(B),ca_modp_id2t(b),w,ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }

    return nil;
}

- _addSelf:(int)v:B multiply:b:(int)w
{
    if (refCount>1) return idReplace(self,[self _add:v:B multiply:b:w]);
    assert([self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([B isZero]) return [self _negateSelf:v];
    if ([b isZero]) return [self _negateSelf:v];

    if ([b isOne]) return [self _addSelf:v:B:+w];
    if ([B isOne]) return [self _addSelf:v:b:+w];
    if ([b isMinusOne]) return [self _addSelf:v:B:-w];
    if ([B isMinusOne]) return [self _addSelf:v:b:-w];

    ca_modp_inaddmul(ca_modp_id2c(self),v,ca_modp_id2t(B),ca_modp_id2t(b),w,ca_modp_getargs(self));
    return [self invalidate];
}

- _multiply:a:(int)v add:B:(int)w
{
    assert([self sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _negate:+w];
    if ([a isZero]) return [B _negate:+w];
    if ([B isZero]) return [self _multiply:a:+v];

    if ([self isOne]) return [a _add:+v:B:+w];
    if ([a isOne]) return [self _add:+v:B:+w];
    if ([self isMinusOne]) return [a _add:-v:B:+w];
    if ([a isMinusOne]) return [self _add:-v:B:+w];

    if (1) {
	ca_modp_c c;
	ca_modp_oomuladd(&c,ca_modp_id2t(self),ca_modp_id2t(a),v,ca_modp_id2t(B),w,ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }

    return nil;
}

- _multiplySelf:a:(int)v add:B:(int)w
{
    if (refCount>1) return idReplace(self,[self _multiply:a:v add:B:w]);
    assert([self sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _negate:+w]);
    if ([a isZero]) return idReplace(self,[B _negate:+w]);
    if ([B isZero]) return [self _multiplySelf:a:+v];

    if ([self isOne]) return idReplace(self,[a _add:+v:B:+w]);
    if ([a isOne]) return [self _addSelf:+v:B:+w];
    if ([self isMinusOne]) return idReplace(self,[a _add:-v:B:+w]);
    if ([a isMinusOne]) return [self _addSelf:-v:B:+w];

    ca_modp_inmuladd(ca_modp_id2c(self),ca_modp_id2t(a),v,ca_modp_id2t(B),w,ca_modp_getargs(self));
    return [self invalidate];
}

- _multiply:a:(int)v add:B multiply:b:(int)w
{
    assert([self sameClass:a] && [self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return [B _multiply:b:+w];
    if ([a isZero]) return [B _multiply:b:+w];
    if ([B isZero]) return [self _multiply:a:+v];
    if ([b isZero]) return [self _multiply:a:+v];

    if ([self isOne]) return [a _add:+v:B multiply:b:+w];
    if ([a isOne]) return [self _add:+v:B multiply:b:+w];
    if ([b isOne]) return [self _multiply:a:+v add:B:+w];
    if ([B isOne]) return [self _multiply:a:+v add:b:+w];

    if ([self isMinusOne]) return [a _add:-v:B multiply:b:+w];
    if ([a isMinusOne]) return [self _add:-v:B multiply:b:+w];
    if ([b isMinusOne]) return [self _multiply:a:+v add:B:-w];
    if ([B isMinusOne]) return [self _multiply:a:+v add:b:-w];

    if (1) {
	ca_modp_c c;
	ca_modp_oomuladdmul(&c,ca_modp_id2t(self),ca_modp_id2t(a),v,ca_modp_id2t(B),ca_modp_id2t(b),w,ca_modp_getargs(self));
	return ca_modp_c2id(self,&c);
    }

    return nil;
}

- _multiplySelf:a:(int)v add:B multiply:b:(int)w
{
    if (refCount>1) return idReplace(self,[self _multiply:a:v add:B multiply:b:w]);
    assert([self sameClass:a] && [self sameClass:B] && [self sameClass:b] && v*v == 1 && w*w == 1);

    if ([self isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([a isZero]) return idReplace(self,[B _multiply:b:+w]);
    if ([B isZero]) return [self _multiplySelf:a:+v];
    if ([b isZero]) return [self _multiplySelf:a:+v];

    if ([self isOne]) return idReplace(self,[a _add:+v:B multiply:b:+w]);
    if ([a isOne]) return [self _addSelf:+v:B multiply:b:+w];
    if ([b isOne]) return [self _multiplySelf:a:+v add:B:+w];
    if ([B isOne]) return [self _multiplySelf:a:+v add:b:+w];

    if ([self isMinusOne]) return idReplace(self,[a _add:-v:B multiply:b:+w]);
    if ([a isMinusOne]) return [self _addSelf:-v:B multiply:b:+w];
    if ([b isMinusOne]) return [self _multiplySelf:a:+v add:B:-w];
    if ([B isMinusOne]) return [self _multiplySelf:a:+v add:b:-w];

    ca_modp_inmuladdmul(ca_modp_id2c(self),ca_modp_id2t(a),v,ca_modp_id2t(B),ca_modp_id2t(b),w,ca_modp_getargs(self));
    return [self invalidate];
}



- (int) characteristic
{
    return modulus;
}

- frobenius
{
    return [self cheapCopy];
}

- frobeniusInverse
{
    return [self cheapCopy];
}


- (BOOL) printsLeadingSign
{
    return NO;
}

- printOn:(IOD)aFile
{
    fprintf(aFile,"%i",value);return self;
}


- fileOutOn:aFiler
{
    [super fileOutOn:aFiler];
    [aFiler fileOut:&value type:'S'];
    [aFiler fileOut:&modulus type:'S'];
    return self;
}

- fileInFrom:aFiler
{
    [super fileInFrom:aFiler];
    [aFiler fileIn:&value type:'S'];
    [aFiler fileIn:&modulus type:'S'];
    return self;
}

@end

