#include "cakit.h"

#define VAL_MIN -32500
#define VAL_MAX +32500

static int ca_intobj_check(ca_intobj_t self,ca_intobj_args args)
{
    assert(args.proto);[args.proto check];
    
    assert(VAL_MIN == - VAL_MAX); /* for opposite tests */
    assert(VAL_MAX * VAL_MAX + VAL_MAX * VAL_MAX < INT_MAX);
    assert(VAL_MIN * VAL_MAX - VAL_MIN * VAL_MAX > INT_MIN);

    if (self->hit) {
	assert(VAL_MIN <= self->val && self->val <= VAL_MAX);
	if (self->obj) {
	    id obj = [args.proto intValue:self->val];
	    assert([obj isEqual:self->obj]);
	    [obj free];
	}
    } else {
	int val;id obj = self->obj;
	assert([obj check]);val = [obj intValue];
	assert(val < VAL_MIN || VAL_MAX < val);
    }
    
    return 1;
}

/*
 * return new reference to object or create it if necessary
 */

static id ca_intobj_obj(ca_intobj_t self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return [self->obj cheapCopy];
    } else {
	id obj;assert(self->hit);
	obj = [args.proto intValue:self->val];
	assert([obj intValue] == self->val);
	return obj;
    }
}

/*
 * works inplace on "self"; creates object if it doesn't exists; returns it.
 */

static id ca_intobj_makeobj(ca_intobj_c *self,ca_intobj_args args)
{
    assert(ca_intobj_check(self,args));
    
    if (self->obj) {
	return self->obj;
    } else {
	assert(self->hit);return self->obj = [args.proto intValue:self->val];
    }
}

/*
 * initializing "c" by object
 */

static void ca_intobj_ooobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by object result of inplace operation
 * used like :
 *
 *		ca_intobj_makeobj(c,args);
 *		ca_intobj_inobj(c,[c->obj negateSelf],args)
 */

static void ca_intobj_inobj(ca_intobj_c *c,id obj,ca_intobj_args args)
{
    int val = [obj intValue];
    /* c->obj is invalid reference now... replace it by result inplace op
     * we cannot check "c" here (as the object is invalid...)
     */
    assert(c->obj != nil);
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = obj;
    } else {
	c->hit = 0;c->obj = obj;
    }
    assert(ca_intobj_check(c,args));
}

/*
 * initializing "c" by int value
 */

static void ca_intobj_ooval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;c->obj = nil;
    } else {
	c->hit = 0;c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

/*
 * replacing "c" by int value
 */

static void ca_intobj_inval(ca_intobj_c *c,int val,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    if (VAL_MIN <= val && val <= VAL_MAX) {
	c->hit = 1;c->val = val;
	if (c->obj) c->obj = [c->obj free];
    } else {
	c->hit = 0;
	if (c->obj) c->obj = [c->obj free];
	c->obj = [args.proto intValue:val];
    }
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_oozero(ca_intobj_c *c,ca_intobj_args args)
{
    c->hit = 1;c->val = 0;c->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_copy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj cheapCopy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_deepcopy(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    c->hit = a->hit;
    c->val = a->val;
    c->obj = (a->obj)?[a->obj copy]:nil;
    assert(ca_intobj_check(c,args) && ca_intobj_check(a,args));
}

static void ca_intobj_move(ca_intobj_c *c,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(c,args));
    c->hit = a->hit;
    c->val = a->val;a->val = SHRT_MAX;
    c->obj = a->obj;a->obj = nil;
    assert(ca_intobj_check(c,args));
}

static void ca_intobj_clear(ca_intobj_c *c,ca_intobj_args args)
{
    c->val = SHRT_MAX;if (c->obj) c->obj = [c->obj free];
}

static ca_intobj_args ca_intobj_getargs(id self)
{
    ca_intobj_args args;args.proto = self;return args;
}

static int ca_intobj_size(ca_intobj_args args)
{
    return 1;				/* pointer increment */
}

static int ca_intobj_bsize(ca_intobj_args args)
{
    return sizeof(ca_intobj_c);		/* byte size carrier */
}

#if 0
static ca_intobj_t  ca_intobj_id2t(id object) { /* not needed */ }
static ca_intobj_c *ca_intobj_id2t(id object) { /* not needed */ }
#endif

/*
 * initialize scratch space "ref" with a read-only value 
 */

static ca_intobj_t ca_intobj_id2ref(id object,ca_intobj_c *ref)
{
    ca_intobj_args args = ca_intobj_getargs(object);	/* use object as proto */
    
    ca_intobj_ooobj(ref,object,args);return ref;
}

/*
 * return a new object, containing a copy of "t"
 */

static id ca_intobj_t2id(id proto,ca_intobj_t t)
{
    ca_intobj_args args = ca_intobj_getargs(proto);
    return ca_intobj_obj(t,args);
}

/*
 * return a new object, containing "t" itself (moved into object)
 */

static id ca_intobj_c2id(id proto,ca_intobj_c *c)
{
    id res;
    ca_intobj_args args = ca_intobj_getargs(proto);
    
    res = ca_intobj_obj(c,args);ca_intobj_clear(c,args);
     
    return res;
}

static unsigned ca_intobj_hash(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit)?a->val:[a->obj hash];
}

static int ca_intobj_iseq(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return YES;
    } else {
	if (a->hit) {
	    return (b->hit)?a->val==b->val:NO;
	} else {
	    return (b->hit)?NO:[a->obj isEqual:b->obj];
	}
    }
}

static int ca_intobj_sgn(ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	short int sgn = a->val;return (sgn==0)?0:((sgn<0)?-1:+1);
    } else {
	return [a->obj sign];
    }
}

static int ca_intobj_cmp(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    int cmp;
    
    if (a==b) {
	cmp = 0;
    } else {
	if (a->hit && b->hit) {
	    cmp = a->val - b->val;
	    cmp = (cmp==0)?0:((cmp<0)?-1:+1);
	} else {
	    id A = ca_intobj_obj(a,args);
	    id B = ca_intobj_obj(b,args);
	    cmp = [A compare:B];[A free];[B free];
	}
    }
    
    return cmp;
}

static int ca_intobj_iszero(ca_intobj_t a,ca_intobj_args args)
{
    int iszero = (a->hit && a->val == 0);
    assert(iszero || (a->hit&&a->val!=0) || (a->hit==0&&[a->obj notZero]));
    return iszero;
}

static int ca_intobj_isop(ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a==b) {
	return NO;
    } else {
	if (a->hit) {
	    assert(VAL_MAX + VAL_MIN == 0);
	    assert(VAL_MIN <= a->val && a->val <= VAL_MAX);
	    assert(VAL_MIN <= b->val && b->val <= VAL_MAX);
	    return (b->hit)?a->val+b->val==0:NO;
	} else {
	    return (b->hit)?NO:[a->obj isOpposite:b->obj];
	}
    }
}

static void ca_intobj_ooneg(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = - a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj negate],args);
    }
    
    assert(ca_intobj_isop(c,a,args));
}

static void ca_intobj_inneg(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = - c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj negateSelf],args);
    }
}

static void ca_intobj_oodbl(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_args args)
{
    if (a->hit) {
	int C = 2 * a->val;C = (v==+1)?+C:-C;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj _double:v],args);
    }
}

static void ca_intobj_indbl(ca_intobj_c *c,int v,ca_intobj_args args)
{
    if (c->hit) {
	int C = 2 * c->val;C = (v==+1)?+C:-C;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj _doubleSelf:v],args);
    }
}

static void ca_intobj_ooadd(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _add:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inadd(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B:w],args);
	[B free];
    }
}

static int ca_intobj_isone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == +1);
}

static int ca_intobj_isminusone(ca_intobj_t a,ca_intobj_args args)
{
    return (a->hit && a->val == -1);
}

static void ca_intobj_oosqr(ca_intobj_c *c,ca_intobj_t a,ca_intobj_args args)
{
    if (a->hit) {
	int C = a->val * a->val;
	ca_intobj_ooval(c,C,args);
    } else {
	ca_intobj_ooobj(c,[a->obj square],args);
    }
}

static void ca_intobj_insqr(ca_intobj_c *c,ca_intobj_args args)
{
    if (c->hit) {
	int C = c->val * c->val;
	ca_intobj_inval(c,C,args);
    } else {
	ca_intobj_inobj(c,[c->obj squareSelf],args);
    }
}

static void ca_intobj_oomul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A multiply:B];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmul(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj multiplySelf:B],args);
	[B free];
    }
}

static void ca_intobj_oomulv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * b->val * v;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _multiply:B:v];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmulv(ca_intobj_c *c,ca_intobj_t b,int v,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * b->val * v;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:B:v],args);
	[B free];
    }
}

static void ca_intobj_ooaddmul(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && b->hit && t->hit) {
	int C = a->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	T = ca_intobj_obj(t,args);
	C = [A _add:v:B multiply:T:w];[A free];[B free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddmul(ca_intobj_c *c,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && b->hit && t->hit) {
	int C = c->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelf:v:B multiply:T:w],args);
	[B free];[T free];
    }
}

static void ca_intobj_oomuladd(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit) {
	int C = a->val * s->val * v + b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	C = [A _multiply:S:v add:B:w];
	[A free];[B free];[S free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladd(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit) {
	int C = c->val * s->val * v + b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B:w],args);
	[B free];[S free];
    }
}

static void ca_intobj_oomuladdmul(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (a->hit && s->hit && b->hit && t->hit) {
	int C = a->val * s->val * v + b->val * t->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,S,T,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	S = ca_intobj_obj(s,args);
	T = ca_intobj_obj(t,args);
	C = [A _multiply:S:v add:B multiply:T:w];
	[A free];[B free];[S free];[T free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inmuladdmul(ca_intobj_c *c,ca_intobj_t s,int v,ca_intobj_t b,ca_intobj_t t,int w,ca_intobj_args args)
{
    if (c->hit && s->hit && b->hit && t->hit) {
	int C = c->val * s->val * v + b->val * t->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	id S = ca_intobj_obj(s,args);
	id T = ca_intobj_obj(t,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _multiplySelf:S:v add:B multiply:T:w],args);
	[B free];[S free];[T free];
    }
}

static void ca_intobj_ooaddsqr(ca_intobj_c *c,ca_intobj_t a,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C = a->val * v + b->val * b->val * w;
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A _addSquare:v:B:w];[A free];[B free];
	ca_intobj_ooobj(c,C,args);
    }
}

static void ca_intobj_inaddsqr(ca_intobj_c *c,int v,ca_intobj_t b,int w,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C = c->val * v + b->val * b->val * w;
	ca_intobj_inval(c,C,args);
    } else {
	id B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	ca_intobj_inobj(c,[c->obj _addSelfSquare:v:B:w],args);
	[B free];
    }
}

static BOOL ca_intobj_oodiv(ca_intobj_c *c,ca_intobj_t a,ca_intobj_t b,ca_intobj_args args)
{
    if (a->hit && b->hit) {
	int C;
	if (b->val == 0 || a->val % b->val != 0) {
	    return FAILURE;
	} else {
	    C = a->val / b->val;
	}
	ca_intobj_ooval(c,C,args);
    } else {
	id A,B,C;
	A = ca_intobj_obj(a,args);
	B = ca_intobj_obj(b,args);
	C = [A divide:B];[A free];[B free];
	if (C==nil) return FAILURE;
	ca_intobj_ooobj(c,C,args);
    }
    
    return SUCCESS;
}

static void ca_intobj_indiv(ca_intobj_c *c,ca_intobj_t b,ca_intobj_args args)
{
    if (c->hit && b->hit) {
	int C;
	if (b->val == 0 || c->val % b->val != 0) {
	    fprintf(stderr,"-exact division failed\n");abort();
	} else {
	    C = c->val / b->val;
	}
	ca_intobj_inval(c,C,args);
    } else {
	id C,B = ca_intobj_obj(b,args);
	ca_intobj_makeobj(c,args);
	C = [c->obj divideSelf:B];
	if (C==nil) {fprintf(stderr,"-exact division failed\n");abort();}
	ca_intobj_inobj(c,C,args);
	[B free];
    }
}

static void ca_intobj_args_fileOut(id aFile,ca_intobj_args *args) {}
static void ca_intobj_args_fileIn(id aFile,ca_intobj_args *args) {}

static void ca_intobj_fileOut(id aFiler,ca_intobj_c *a,ca_intobj_args args)
{
    assert(ca_intobj_check(a,args));
    [aFiler fileOut:&a->hit type:'S']; 
    if (a->hit) {
	[aFiler fileOut:&a->val type:'s']; /* signed ! */
    } else {
	[aFiler fileOut:&a->obj type:'@'];
    }
}

static void ca_intobj_fileIn(id aFiler,ca_intobj_c *c,ca_intobj_args args)
{
    [aFiler fileIn:&c->hit type:'S'];
    if (c->hit) {
	[aFiler fileIn:&c->val type:'s'];
        c->obj = nil;
    } else {
	[aFiler fileIn:&c->obj type:'@'];
    }
}


/*
 * Computer Algebra Kit (c) 1993,98 by Comp.Alg.Objects.  All Rights Reserved.
 */

#if ca_intobj_pervalue
static ca_intobj_t ca_intobjseq_tval(ca_intobjseq_t self)
{
    return *self;
}
static ca_intobj_c *ca_intobjseq_cval(ca_intobjseq_t self)
{
    return self;
}
#else
static ca_intobj_t ca_intobjseq_tval(ca_intobjseq_t self)
{
    return self;
}
static ca_intobj_c *ca_intobjseq_cval(ca_intobjseq_t self)
{
    return self;
}
#endif

static int ca_intobjseq_check(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    while (n--) {
	ca_intobj_check(ca_intobjseq_tval(self),args);
	self += ca_intobj_size(args);
    }
    return 1;
}

static ca_intobjseq_t ca_intobjseq_alloc(int n,ca_intobj_args args)
{
    return (ca_intobjseq_t)OC_Malloc(n * ca_intobj_bsize(args));
}

static ca_intobjseq_t ca_intobjseq_realloc(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    return (ca_intobjseq_t)OC_Realloc(self,n * ca_intobj_bsize(args));
}

static ca_intobjseq_t ca_intobjseq_free(ca_intobjseq_t self,ca_intobj_args args)
{
    OC_Free(self);return NULL;
}

static void ca_intobjseq_oozero(ca_intobjseq_t c,int n,ca_intobj_args args)
{
    while (n--) {
	ca_intobj_oozero(ca_intobjseq_cval(c),args);
	c += ca_intobj_size(args);
    }
}

static void ca_intobjseq_copy(ca_intobjseq_t c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_copy(ca_intobjseq_cval(c),ca_intobjseq_tval(a),args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_deepcopy(ca_intobjseq_t c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_deepcopy(ca_intobjseq_cval(c),ca_intobjseq_tval(a),args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

#if ca_intobj_isvalue
static void ca_intobjseq_clear(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    /* nothing to clear */
}
#else
static void ca_intobjseq_clear(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    while (n--) {
	ca_intobj_clear(self,args);
	self+=ca_intobj_size(args);
    }
}
#endif /* ca_intobj_isvalue */

static int ca_intobjseq_iseq(ca_intobjseq_t a,ca_intobjseq_t b,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_intobj_iseq(ca_intobjseq_tval(a),ca_intobjseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_intobj_size(args);b+=ca_intobj_size(args);
	}
    }
    return 1;
}

static unsigned ca_intobjseq_hash(ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    return (n)?ca_intobj_hash(ca_intobjseq_tval(a),args):0;
}

static void ca_intobjseq_place(ca_intobjseq_t self,ca_intobj_t a,int i,ca_intobj_args args)
{
    ca_intobj_c *c = ca_intobjseq_cval(self + i * ca_intobj_size(args));
    ca_intobj_clear(c,args);ca_intobj_copy(c,a,args);
}

static int ca_intobjseq_insertlast(ca_intobjseq_t self,ca_intobj_t a,int n,ca_intobj_args args)
{
    ca_intobj_c *c;
    assert(n >= 0);
    c = ca_intobjseq_cval(self + n * ca_intobj_size(args));
    ca_intobj_copy(c,a,args);return ++n;
}

static int ca_intobjseq_insertfirst(ca_intobjseq_t self,ca_intobj_t a,int n,ca_intobj_args args)
{
    int j;
    assert(n > 0);
    self += n * ca_intobj_size(args);
    j=n;while(j--) {
	ca_intobjseq_t t;
	t = self - ca_intobj_size(args);
	ca_intobj_move(ca_intobjseq_cval(self),ca_intobjseq_cval(t),args);
	self = t;
    }
    ca_intobj_copy(ca_intobjseq_cval(self),a,args);
    return ++n;
}

static int ca_intobjseq_insertat(ca_intobjseq_t self,ca_intobj_t a,int i,int n,ca_intobj_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_intobj_size(args);
    return i + ca_intobjseq_insertfirst(self,a,n-i,args);
}

static int ca_intobjseq_freelast(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    ca_intobj_c *c;assert(0 < n);
    --n;c = ca_intobjseq_cval(self + n * ca_intobj_size(args));
    ca_intobj_clear(c,args);return n;
}

static int ca_intobjseq_freefirst(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    int j;
    assert(0 < n);
    ca_intobj_clear(ca_intobjseq_cval(self),args);
    --n;j=n;while(j--) {
	ca_intobjseq_t t;
	t = self+ca_intobj_size(args);
	ca_intobj_move(ca_intobjseq_cval(self),ca_intobjseq_cval(t),args);
	self = t;
    }
    return n;
}

static int ca_intobjseq_freeat(ca_intobjseq_t self,int i,int n,ca_intobj_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_intobj_size(args);
    return i + ca_intobjseq_freefirst(self,n-i,args);
}

static ca_intobj_t ca_intobjseq_at(ca_intobjseq_t self,int i,ca_intobj_args args)
{
    ca_intobj_t res;assert(0 <= i);
    res = ca_intobjseq_tval(self + i * ca_intobj_size(args));
    assert(ca_intobj_check(res,args));
    return res;
}

static ca_intobj_t ca_intobjseq_last(ca_intobjseq_t self,int n,ca_intobj_args args)
{
    return ca_intobjseq_at(self,n-1,args);
}

static ca_intobj_c *ca_intobjseq_cvalat(ca_intobjseq_t self,int i,ca_intobj_args args)
{
    assert(0 <= i);return ca_intobjseq_cval(self + i * ca_intobj_size(args));
}

static int ca_intobjseq_iszero(ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	if (ca_intobj_iszero(ca_intobjseq_tval(a),args)) {
	    a+=ca_intobj_size(args);
	} else {
	    return 0;
	}
    }
    return 1;
}

static int ca_intobjseq_isop(ca_intobjseq_t a,ca_intobjseq_t b,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_intobj_isop(ca_intobjseq_tval(a),ca_intobjseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_intobj_size(args);b+=ca_intobj_size(args);
	}
    }
    return 1;
}

static void ca_intobjseq_ooneg(ca_intobjseq_t c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_ooneg(ca_intobjseq_cval(c),ca_intobjseq_tval(a),args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inneg(ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_inneg(ca_intobjseq_cval(a),args);
	a += ca_intobj_size(args);
    }
}

static void ca_intobjseq_oonegv(ca_intobjseq_t c,ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(v*v == 1);if (v == -1) ca_intobjseq_ooneg(c,a,n,args);
    else ca_intobjseq_copy(c,a,n,args);
}

static void ca_intobjseq_innegv(ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(v*v == 1);if (v == -1) ca_intobjseq_inneg(a,n,args);
}

static void ca_intobjseq_oodbl(ca_intobjseq_t c,ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_oodbl(ca_intobjseq_cval(c),ca_intobjseq_tval(a),v,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_indbl(ca_intobjseq_t a,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_indbl(ca_intobjseq_cval(a),v,args);
	a += ca_intobj_size(args);
    }
}

static void ca_intobjseq_ooadd(ca_intobjseq_t c,ca_intobjseq_t a,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_ooadd(ca_intobjseq_cval(c),ca_intobjseq_tval(a),v,ca_intobjseq_tval(b),w,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inadd(ca_intobjseq_t a,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inadd(ca_intobjseq_cval(a),v,ca_intobjseq_tval(b),w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulsc(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_oomul(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulsc(ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_inmul(ca_intobjseq_cval(a),s,args);
	a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulscv(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_oomulv(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,v,args);
	c+=ca_intobj_size(args);a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulscv(ca_intobjseq_t a,ca_intobj_t s,int v,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_intobj_inmulv(ca_intobjseq_cval(a),s,v,args);
	a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_ooaddmulsc(ca_intobjseq_t c,ca_intobjseq_t a,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_ooaddmul(ca_intobjseq_cval(c),ca_intobjseq_tval(a),v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);c+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulscadd(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_oomuladd(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,v,ca_intobjseq_tval(b),w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);c+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oomulscaddmulsc(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_oomuladdmul(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);c+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inaddmulsc(ca_intobjseq_t a,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inaddmul(ca_intobjseq_cval(a),v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulscadd(ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inmuladd(ca_intobjseq_cval(a),s,v,ca_intobjseq_tval(b),w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_inmulscaddmulsc(ca_intobjseq_t a,ca_intobj_t s,int v,ca_intobjseq_t b,ca_intobj_t t,int w,int n,ca_intobj_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_intobj_inmuladdmul(ca_intobjseq_cval(a),s,v,ca_intobjseq_tval(b),t,w,args);
	a+=ca_intobj_size(args);b+=ca_intobj_size(args);
    }
}

static BOOL ca_intobjseq_oodivsc(ca_intobjseq_t c,ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    int i;
    BOOL failed;
    ca_intobjseq_t orgc = c;
    assert(n >= 0);
   
    for(i=0;i<n;i++) {
    
	failed = ca_intobj_oodiv(ca_intobjseq_cval(c),ca_intobjseq_tval(a),s,args);
	
	if (failed) {
	    ca_intobjseq_clear(orgc,i - 1,args);
	    return FAILURE;
	} else {
	    c+=ca_intobj_size(args);a+=ca_intobj_size(args);
	}
    }
    
    return SUCCESS;
}

static void ca_intobjseq_indivsc(ca_intobjseq_t a,ca_intobj_t s,int n,ca_intobj_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_intobj_indiv(ca_intobjseq_cval(a),s,args);
	a+=ca_intobj_size(args);
    }
}

static void ca_intobjseq_oodotsqr(ca_intobj_c *c,ca_intobjseq_t a,int n,ca_intobj_args args)
{
    assert(n >= 0);
    ca_intobj_oozero(c,args);
    while (n--) {
	ca_intobj_inaddsqr(c,+1,ca_intobjseq_tval(a),+1,args);a += ca_intobj_size(args);
    }
}

static void ca_intobjseq_oodotmul(ca_intobj_c *c,ca_intobjseq_t a,ca_intobjseq_t b,int n,ca_intobj_args args)
{
    assert(n >= 0);
    ca_intobj_oozero(c,args);
    while (n--) {
	ca_intobj_inaddmul(c,+1,ca_intobjseq_tval(a),ca_intobjseq_tval(b),+1,args);
	a += ca_intobj_size(args);b += ca_intobj_size(args);
    }
}

static void ca_intobjseq_args_fileOut(id aFiler,ca_intobj_args *args)
{
    ca_intobj_args_fileOut(aFiler,args);
}
static void ca_intobjseq_args_fileIn(id aFiler,ca_intobj_args *args)
{
    ca_intobj_args_fileIn(aFiler,args);
}

static void ca_intobjseq_fileOut(id aFiler,ca_intobjseq_t self,int n,ca_intobj_args args)
{
    assert(n>=0);
    while (n--) {
	ca_intobj_fileOut(aFiler,ca_intobjseq_cval(self),args);
	self += ca_intobj_size(args);
    }
}

static void ca_intobjseq_fileIn(id aFiler,ca_intobjseq_t self,int n,ca_intobj_args args)
{
    assert(n>=0);
    while (n--) {
	ca_intobj_fileIn(aFiler,ca_intobjseq_cval(self),args);
	self += ca_intobj_size(args);
    }
}

static ca_intobjvec_args ca_intobjvec_getargs(id self)
{
    ca_intobjvec_args args;
    args.sargs = ca_intobj_getargs([self scalarZero]);
    return args;
}

static int ca_intobjvec_check(ca_intobjvec_t self,ca_intobjvec_args args)
{
    assert(0 <= self->n && self->n <= self->c);
    ca_intobjseq_check(self->ptr,self->n,args.sargs);return 1;
}

static ca_intobjvec_t ca_intobjvec_id2t(id object)
{
    return [object ca_intobjvec_value];
}

static ca_intobjvec_c *ca_intobjvec_id2c(id object)
{
    return [object ca_intobjvec_reference];
}

static id ca_intobjvec_t2id(id proto,ca_intobjvec_t f)
{
    return [proto ca_intobjvec_value:f];
}

static id ca_intobjvec_c2id(id proto,ca_intobjvec_c *r)
{
    return [proto ca_intobjvec_reference:r];
}

static void ca_intobjvec_init(ca_intobjvec_t self,int n,ca_intobjvec_args args)
{
    self->n   = n;
    self->c   = n;
    self->ptr = ca_intobjseq_alloc(n,args.sargs);
}

static void ca_intobjvec_initcount(ca_intobjvec_t c,int n,ca_intobjvec_args args)
{
    ca_intobjvec_init(c,n,args);
    ca_intobjseq_oozero(c->ptr,n,args.sargs);
}

static void ca_intobjvec_initcapacity(ca_intobjvec_t self,int n,ca_intobjvec_args args)
{
    self->n   = 0;
    self->c   = n;
    self->ptr = ca_intobjseq_alloc(n,args.sargs);
}

static void ca_intobjvec_copy(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobjvec_args args)
{
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_copy(c->ptr,a->ptr,a->n,args.sargs);
}

static void ca_intobjvec_deepcopy(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobjvec_args args)
{
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_deepcopy(c->ptr,a->ptr,a->n,args.sargs);
}

static void ca_intobjvec_move(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobjvec_args args)
{
    c->n   = a->n;
    c->c   = a->c;
    c->ptr = a->ptr;a->ptr = NULL;
}

static void ca_intobjvec_destroy(ca_intobjvec_t self,ca_intobjvec_args args)
{
    /*
     * clearing memory allocated by ca_intobjvec_init after a failed operation
     * can't do a ca_intobjseq_clear in that case as the handler doesnt know
     * what elements have been allocated
     */
     
    self->ptr = ca_intobjseq_free(self->ptr,args.sargs);
}

static void ca_intobjvec_clear(ca_intobjvec_t self,ca_intobjvec_args args)
{
    ca_intobjseq_clear(self->ptr,self->n,args.sargs);
    ca_intobjvec_destroy(self,args);
}

static int ca_intobjvec_count(ca_intobjvec_t self)
{
    return self->n;
}

static unsigned ca_intobjvec_hash(ca_intobjvec_t a,ca_intobjvec_args args)
{
    return ca_intobjseq_hash(a->ptr,a->n,args.sargs);
}

static int ca_intobjvec_iseq(ca_intobjvec_t a,ca_intobjvec_t b,ca_intobjvec_args args)
{
    assert(a != b);
    return (a->n == b->n && ca_intobjseq_iseq(a->ptr,b->ptr,a->n,args.sargs));
}

static void ca_intobjvec_expand(ca_intobjvec_t self,ca_intobjvec_args args)
{
    assert(self->n == self->c);
    self->c   = self->n + self->c + 1;
    self->ptr = ca_intobjseq_realloc(self->ptr,self->c,args.sargs);
    assert(self->n < self->c);
}

static void ca_intobjvec_insertfirst(ca_intobjvec_t self,ca_intobj_t a,ca_intobjvec_args args)
{
    assert(self->n <= self->c);
    if (self->n == self->c) ca_intobjvec_expand(self,args);
    self->n = ca_intobjseq_insertlast(self->ptr,a,self->n,args.sargs);
}

static void ca_intobjvec_insertat(ca_intobjvec_t self,ca_intobj_t a,int i,ca_intobjvec_args args)
{
    assert(0 < i && i <= self->n && self->n <= self->c);
    if (self->n == self->c) ca_intobjvec_expand(self,args);
    
    if (i == self->n) {
    self->n = ca_intobjseq_insertfirst(self->ptr,a,self->n,args.sargs);
    } else {
    self->n = ca_intobjseq_insertat(self->ptr,a,self->n-i-1,self->n,args.sargs);
    }
}

static void ca_intobjvec_place(ca_intobjvec_t self,ca_intobj_t a,int i,ca_intobjvec_args args)
{
    assert(0 <= i && i < self->n);
    ca_intobjseq_place(self->ptr,a,self->n - i - 1,args.sargs);
}

static void ca_intobjvec_freefirst(ca_intobjvec_t self,ca_intobjvec_args args)
{
    assert(self->n > 0);
    self->n = ca_intobjseq_freelast(self->ptr,self->n,args.sargs);
}

static void ca_intobjvec_freeat(ca_intobjvec_t self,int i,ca_intobjvec_args args)
{
    assert(0 < i && i < self->n);
    self->n = ca_intobjseq_freeat(self->ptr,self->n-i-1,self->n,args.sargs);
}


static ca_intobj_t ca_intobjvec_at(ca_intobjvec_t self,int ix,ca_intobjvec_args args)
{
    assert(0 <= ix && ix < self->n);
    return ca_intobjseq_at(self->ptr,self->n - ix - 1,args.sargs);
}

static ca_intobj_t ca_intobjvec_first(ca_intobjvec_t self,ca_intobjvec_args args)
{
    assert(self->n > 0);
    return ca_intobjseq_last(self->ptr,self->n,args.sargs);
}

static ca_intobj_t ca_intobjvec_elt(ca_intobjvec_t self,ca_intobjvec_ixt ix,ca_intobjvec_args args)
{
    assert(ix->n == self->n);
    return ca_intobjvec_at(self,ix->i,args);
}

static int ca_intobjvec_iszero(ca_intobjvec_t self,ca_intobjvec_args args)
{
    return ca_intobjseq_iszero(self->ptr,self->n,args.sargs);
}

static int ca_intobjvec_isop(ca_intobjvec_t a,ca_intobjvec_t b,ca_intobjvec_args args)
{
    assert(a != b);
    return (a->n == b->n && ca_intobjseq_isop(a->ptr,b->ptr,a->n,args.sargs));
}

static void ca_intobjvec_ooneg(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobjvec_args args)
{
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_ooneg(c->ptr,a->ptr,a->n,args.sargs);
}

static void ca_intobjvec_inneg(ca_intobjvec_t a,ca_intobjvec_args args)
{
    ca_intobjseq_inneg(a->ptr,a->n,args.sargs);
}

static void ca_intobjvec_oodbl(ca_intobjvec_t c,ca_intobjvec_t a,int v,ca_intobjvec_args args)
{
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_oodbl(c->ptr,a->ptr,v,a->n,args.sargs);
}

static void ca_intobjvec_indbl(ca_intobjvec_t a,int v,ca_intobjvec_args args)
{
    ca_intobjseq_indbl(a->ptr,v,a->n,args.sargs);
}

static void ca_intobjvec_checkcount(int a,int b)
{
    if (a != b) { fprintf(stderr,"Vector : Dimensions Not Equal\n");abort(); }
}

static void ca_intobjvec_ooadd(ca_intobjvec_t c,ca_intobjvec_t a,int v,ca_intobjvec_t b,int w,ca_intobjvec_args args)
{
    assert(a!=b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_ooadd(c->ptr,a->ptr,v,b->ptr,w,a->n,args.sargs);
}

static void ca_intobjvec_inadd(ca_intobjvec_t a,int v,ca_intobjvec_t b,int w,ca_intobjvec_args args)
{
    assert(a!=b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjseq_inadd(a->ptr,v,b->ptr,w,a->n,args.sargs);
}

static void ca_intobjvec_oomulsc(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobj_t s,ca_intobjvec_args args)
{
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_oomulsc(c->ptr,a->ptr,s,a->n,args.sargs);
}

static void ca_intobjvec_inmulsc(ca_intobjvec_t a,ca_intobj_t s,ca_intobjvec_args args)
{
    ca_intobjseq_inmulsc(a->ptr,s,a->n,args.sargs);
}

static void ca_intobjvec_ooaddmulsc(ca_intobjvec_t c,ca_intobjvec_t a,int v,ca_intobjvec_t b,ca_intobj_t s,int w,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_ooaddmulsc(c->ptr,a->ptr,v,b->ptr,s,w,a->n,args.sargs);
}

static void ca_intobjvec_inaddmulsc(ca_intobjvec_t a,int v,ca_intobjvec_t b,ca_intobj_t s,int w,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjseq_inaddmulsc(a->ptr,v,b->ptr,s,w,a->n,args.sargs);
}

static void ca_intobjvec_oomulscadd(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobj_t s,int v,ca_intobjvec_t b,int w,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_oomulscadd(c->ptr,a->ptr,s,v,b->ptr,w,a->n,args.sargs);
}

static void ca_intobjvec_inmulscadd(ca_intobjvec_t a,ca_intobj_t s,int v,ca_intobjvec_t b,int w,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjseq_inmulscadd(a->ptr,s,v,b->ptr,w,a->n,args.sargs);
}

static void ca_intobjvec_oomulscaddmulsc(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobj_t s,int v,ca_intobjvec_t b,ca_intobj_t t,int w,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjvec_init(c,a->n,args);
    ca_intobjseq_oomulscaddmulsc(c->ptr,a->ptr,s,v,b->ptr,t,w,a->n,args.sargs);
}

static void ca_intobjvec_inmulscaddmulsc(ca_intobjvec_t a,ca_intobj_t s,int v,ca_intobjvec_t b,ca_intobj_t t,int w,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjseq_inmulscaddmulsc(a->ptr,s,v,b->ptr,t,w,a->n,args.sargs);
}

static BOOL ca_intobjvec_oodivsc(ca_intobjvec_t c,ca_intobjvec_t a,ca_intobj_t s,ca_intobjvec_args args)
{
    BOOL failed;
    
    ca_intobjvec_init(c,a->n,args);
    failed = ca_intobjseq_oodivsc(c->ptr,a->ptr,s,a->n,args.sargs);
    if (failed) ca_intobjvec_destroy(c,args);
    
    return failed;
}

static void ca_intobjvec_indivsc(ca_intobjvec_t a,ca_intobj_t s,ca_intobjvec_args args)
{
    ca_intobjseq_indivsc(a->ptr,s,a->n,args.sargs);
}

static void ca_intobjvec_oodotmul(ca_intobj_c *c,ca_intobjvec_t a,ca_intobjvec_t b,ca_intobjvec_args args)
{
    assert(a != b);ca_intobjvec_checkcount(a->n,b->n);
    ca_intobjseq_oodotmul(c,a->ptr,b->ptr,a->n,args.sargs);
}

static void ca_intobjvec_oodotsqr(ca_intobj_c *c,ca_intobjvec_t a,ca_intobjvec_args args)
{
    ca_intobjseq_oodotsqr(c,a->ptr,a->n,args.sargs);
}

static void ca_intobjvec_args_fileOut(id aFiler,ca_intobjvec_args *args)
{
    ca_intobjseq_args_fileOut(aFiler,&args->sargs);
}
static void ca_intobjvec_args_fileIn(id aFiler,ca_intobjvec_args *args)
{
    ca_intobjseq_args_fileIn(aFiler,&args->sargs);
}

static void ca_intobjvec_fileOut(id aFiler,ca_intobjvec_t self,ca_intobjvec_args args)
{
    [aFiler fileOut:&self->n type:'i'];
    ca_intobjseq_fileOut(aFiler,self->ptr,self->n,args.sargs);
}

static void ca_intobjvec_fileIn(id aFiler,ca_intobjvec_t self,ca_intobjvec_args args)
{
    int n;
    [aFiler fileIn:&n type:'i']; 
    ca_intobjvec_init(self,n,args);
    ca_intobjseq_fileIn(aFiler,self->ptr,self->n,args.sargs);
}

static void ca_intobjvec_ix_init(ca_intobjvec_ixt ix,id self)
{
    ca_intobjvec_t v = [self ca_intobjvec_value];
    ix->n = v->n;
}

static void ca_intobjvec_ix_copy(ca_intobjvec_ixt c,ca_intobjvec_ixt a)
{
    c->i = a->i;c->n = a->n;
}

static void ca_intobjvec_ix_clear(ca_intobjvec_ixt ix)
{
    /* nothing to clear */
}

static int ca_intobjvec_ix_count(ca_intobjvec_ixt ix)
{
    return ix->n;
}

static int ca_intobjvec_ix_isempty(ca_intobjvec_ixt ix)
{
    return ix->n == 0;
}

static int ca_intobjvec_ix_ismatch(ca_intobjvec_ixt ix)
{
    assert(-1 <= ix->i && ix->i <= ix->n);
    return 0 <= ix->i && ix->i < ix->n;
}

static void ca_intobjvec_ix_tofirst(ca_intobjvec_ixt ix)
{
    ix->i = -1;	/* next -> first */
}

static void ca_intobjvec_ix_tolast(ca_intobjvec_ixt ix)
{
    ix->i = ix->n; /* prev -> last */
}

static void ca_intobjvec_ix_to(ca_intobjvec_ixt ix,int i)
{
    /* the actual position is n-i-1; conversion done in elting routine... */
    if (i < 0) ix->i = -1;else if (i > ix->n) ix->i = ix->n;
    else ix->i = i-1; /* next -> at i */
}

static void ca_intobjvec_ix_next(ca_intobjvec_ixt ix)
{
    ix->i = (ix->i < ix->n)?ix->i + 1:ix->n;
}

static void ca_intobjvec_ix_prev(ca_intobjvec_ixt ix)
{
    ix->i = (0 <= ix->i)?ix->i - 1:-1;
}
@implementation ca_intobj_vector : ca_vector

- check
{
    [super check];
    [scalarZero check];ca_intobjvec_check(&value,ca_intobjvec_getargs(self));
    return self;
}


- _setUpScalarZero:aScalarZero numScalars:(int)numScalars
{
    scalarZero = aScalarZero;
    ca_intobjvec_initcount(&value,numScalars,ca_intobjvec_getargs(self));
    assert([self check]);return self;
}

+ scalarZero:aScalarZero numScalars:(int)numScalars
{
    return [[super new] _setUpScalarZero:aScalarZero numScalars:numScalars];
}

- copy
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_copy(&c,&value,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- deepCopy
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_deepcopy(&c,&value,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- clear
{
    scalarZero = [scalarZero free];
    ca_intobjvec_clear(&value,ca_intobjvec_getargs(self));
    return [super clear];
}

- (ca_intobjvec_t)ca_intobjvec_value
{
    return &value;
}

- ca_intobjvec_value:(ca_intobjvec_t)aValue
{
    self = [self clone];
    ca_intobjvec_copy(&value,aValue,ca_intobjvec_getargs(self));
    assert([self check]);return self;
}

- (ca_intobjvec_t)ca_intobjvec_reference
{
    return &value;
}

- ca_intobjvec_reference:(ca_intobjvec_c *)aReference
{
    self = [self clone];
    ca_intobjvec_move(&value,aReference,ca_intobjvec_getargs(self));
    assert([self check]);return self;
}

- clone
{
    self = [super clone];
    scalarZero = [scalarZero cheapCopy];
    return self;
}

- capacity:(int)aCapacity
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_initcapacity(&c,aCapacity,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- numScalars:(int)numScalars
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_initcount(&c,numScalars,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}


- scalarZero
{
    return scalarZero;
}

- (int) numScalars
{
    assert([self check]);return ca_intobjvec_count(&value);
}


- insertScalar:aScalar
{
    ca_intobj_c c;
    assert([aScalar check]);[self checkRefCountOne];assert([self check]);
    ca_intobjvec_insertfirst(&value,ca_intobj_id2ref(aScalar,&c),ca_intobjvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);[self invalidate];return self;
}

- insertScalar:aScalar at:(int)i
{
    ca_intobj_c c;
    assert(aScalar);[self checkRefCountOne];assert([self check]);
    ca_intobjvec_insertat(&value,ca_intobj_id2ref(aScalar,&c),i,ca_intobjvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);[self invalidate];return self;
}


- removeScalar
{
    if ([self numScalars]) {
	id res;
	ca_intobj_t t = ca_intobjvec_first(&value,ca_intobjvec_getargs(self));
	assert([self check]);[self checkRefCountOne];
	res = ca_intobj_t2id(scalarZero,t);
	ca_intobjvec_freefirst(&value,ca_intobjvec_getargs(self));
	assert([self check]);assert([res check]);[self invalidate];return res;
    } else {
	return nil;
    }
}

- removeScalarAt:(int)i
{
    id res;
    assert([self check]);[self checkRefCountOne];
    res = ca_intobj_t2id(scalarZero,ca_intobjvec_at(&value,i,ca_intobjvec_getargs(self)));
    ca_intobjvec_freeat(&value,i,ca_intobjvec_getargs(self));
    assert([self check]);assert([res check]);[self invalidate];return res;
}


- placeScalar:aScalar at:(int)i
{
    ca_intobj_c c;
    assert([aScalar check]);assert([self check]);[self checkRefCountOne];
    ca_intobjvec_place(&value,ca_intobj_id2ref(aScalar,&c),i,ca_intobjvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);[self invalidate];return self;
}

- replaceScalarAt:(int)i with:aScalar
{
    id res;
    ca_intobj_c c;
    ca_intobj_t t = ca_intobjvec_at(&value,i,ca_intobjvec_getargs(self));
    assert([self check]);[self checkRefCountOne];
    res = ca_intobj_t2id(scalarZero,t);
    ca_intobjvec_place(&value,ca_intobj_id2ref(aScalar,&c),i,ca_intobjvec_getargs(self));
    assert([aScalar check]);[aScalar free];
    assert([self check]);assert([res check]);[self invalidate];return res;
}


- eachScalar
{
    id aSequence = [ca_intobjvec_sequence content:[self cheapCopy]];
    return [CASequence over:aSequence];
}

- elt_ca_intobjvec_ix:(ca_intobjvec_ixt)ix
{
    if (ca_intobjvec_ix_ismatch(ix)) {
	ca_intobj_t t = ca_intobjvec_elt(&value,ix,ca_intobjvec_getargs(self));
	return ca_intobj_t2id(scalarZero,t);
    } else {
	return nil;
    }
}


- dotMultiply:b
{
    ca_intobj_c c;	/* carrier for scalar */
    ca_intobjvec_oodotmul(&c,&value,[b ca_intobjvec_value],ca_intobjvec_getargs(self));
    /* now promote the carrier to object */
    /* all of this used to be very difficult in 0.6 */
    return ca_intobj_c2id(scalarZero,&c);
}

- dotSquare
{
    ca_intobj_c c;	/* carrier for scalar */
    ca_intobjvec_oodotsqr(&c,&value,ca_intobjvec_getargs(self));
    /* now promote the carrier to object */
    /* all of this used to be very difficult in 0.6 */
    return ca_intobj_c2id(scalarZero,&c);
}


- fileOutOn:aFiler
{
    ca_intobjvec_args args;
    assert([self check]);
    [super fileOutOn:aFiler];
    args = ca_intobjvec_getargs(self);
    ca_intobjvec_args_fileOut(aFiler,&args);
    ca_intobjvec_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_intobjvec_args args;
    [super fileInFrom:aFiler];
    ca_intobjvec_args_fileIn(aFiler,&args);
    ca_intobjvec_fileIn(aFiler,&value,args);
    return self;
}

- (unsigned) hash
{
    return ca_intobjvec_hash(ca_intobjvec_id2t(self),ca_intobjvec_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_intobjvec_iseq(ca_intobjvec_id2t(self),ca_intobjvec_id2t(b),ca_intobjvec_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_intobjvec_iseq(ca_intobjvec_id2t(self),ca_intobjvec_id2t(b),ca_intobjvec_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_intobjvec_iszero(ca_intobjvec_id2t(self),ca_intobjvec_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_intobjvec_iszero(ca_intobjvec_id2t(self),ca_intobjvec_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_intobjvec_isop(ca_intobjvec_id2t(self),ca_intobjvec_id2t(b),ca_intobjvec_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_intobjvec_isop(ca_intobjvec_id2t(self),ca_intobjvec_id2t(b),ca_intobjvec_getargs(self));
}

- negate
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_ooneg(&c,ca_intobjvec_id2t(self),ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_intobjvec_inneg(ca_intobjvec_id2c(self),ca_intobjvec_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_oodbl(&c,ca_intobjvec_id2t(self),v,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_intobjvec_indbl(ca_intobjvec_id2c(self),v,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- double
{
    ca_intobjvec_c c;assert([self check]);
    ca_intobjvec_oodbl(&c,ca_intobjvec_id2t(self),+1,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_intobjvec_indbl(ca_intobjvec_id2c(self),+1,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_intobjvec_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_intobjvec_ooadd(&c,ca_intobjvec_id2t(self),v,ca_intobjvec_id2t(b),w,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_intobjvec_inadd(ca_intobjvec_id2c(self),v,ca_intobjvec_id2t(b),w,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_intobjvec_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobjvec_ooadd(&c,ca_intobjvec_id2t(self),+1,ca_intobjvec_id2t(b),+1,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_intobjvec_inadd(ca_intobjvec_id2c(self),+1,ca_intobjvec_id2t(b),+1,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_intobjvec_c c;
    assert([self checkSameClass:b] && self != b);
    ca_intobjvec_ooadd(&c,ca_intobjvec_id2t(self),+1,ca_intobjvec_id2t(b),-1,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_intobjvec_inadd(ca_intobjvec_id2c(self),+1,ca_intobjvec_id2t(b),-1,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- multiplyScalar:s
{
    ca_intobjvec_c c;
    ca_intobj_c sc;
    ca_intobjvec_oomulsc(&c,ca_intobjvec_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- multiplySelfScalar:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self multiplyScalar:s]);
    ca_intobjvec_inmulsc(ca_intobjvec_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobjvec_getargs(self));
    return [self invalidate];
}

- divideScalar:s
{
    ca_intobjvec_c c;
    ca_intobj_c sc;
    BOOL failed = ca_intobjvec_oodivsc(&c,ca_intobjvec_id2t(self),ca_intobj_id2ref(s,&sc),ca_intobjvec_getargs(self));
    return (failed)?nil:ca_intobjvec_c2id(self,&c);
}

- divideSelfScalar:s
{
    ca_intobj_c sc;
    if (refCount>1) return idReplace(self,[self divideScalar:s]);
    ca_intobjvec_indivsc(ca_intobjvec_id2c(self),ca_intobj_id2ref(s,&sc),ca_intobjvec_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyScalar:b:(int)w
{
    ca_intobjvec_c c;
    ca_intobj_c bc;
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobjvec_ooaddmulsc(&c,ca_intobjvec_id2t(self),v,ca_intobjvec_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyScalar:b:(int)w
{
    ca_intobj_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyScalar:b:w]);
    assert([self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobjvec_inaddmulsc(ca_intobjvec_id2c(self),v,ca_intobjvec_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B:(int)w
{
    ca_intobjvec_c c;
    ca_intobj_c ac;
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_intobjvec_oomulscadd(&c,ca_intobjvec_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobjvec_id2t(B),w,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B:(int)w
{
    ca_intobj_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && v*v == 1 && w*w == 1);
    ca_intobjvec_inmulscadd(ca_intobjvec_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobjvec_id2t(B),w,ca_intobjvec_getargs(self));
    return [self invalidate];
}

- _multiplyScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_intobjvec_c c;
    ca_intobj_c ac,bc;
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobjvec_oomulscaddmulsc(&c,ca_intobjvec_id2t(self),ca_intobj_id2ref(a,&ac),v,ca_intobjvec_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobjvec_getargs(self));
    return ca_intobjvec_c2id(self,&c);
}

- _multiplySelfScalar:a:(int)v add:B multiplyScalar:b:(int)w
{
    ca_intobj_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyScalar:a:v add:B multiplyScalar:b:w]);
    assert([scalarZero sameClass:a] && [self sameClass:B] && [scalarZero sameClass:b] && v*v == 1 && w*w == 1);
    ca_intobjvec_inmulscaddmulsc(ca_intobjvec_id2c(self),ca_intobj_id2ref(a,&ac),v,ca_intobjvec_id2t(B),ca_intobj_id2ref(b,&bc),w,ca_intobjvec_getargs(self));
    return [self invalidate];
}
@end

