#include "cakit.h"

static ca_modp_args ca_modp_getargs(id self)
{
   ca_modp_args args;args.p = [self modulus];return args;
}

static int ca_modp_size(ca_modp_args args)
{
    return 1;				/* pointer increment */
}

static int ca_modp_bsize(ca_modp_args args)
{
    return sizeof(ca_modp_c);		/* bytes */
}

static ca_modp_t ca_modp_id2t(id object)
{
    return [object ca_modp_value];	/* read-only */
}

static ca_modp_t ca_modp_id2ref(id object,ca_modp_c *ref)
{
    return [object ca_modp_value];	/* read-only, scratch space "ref" */
}

static ca_modp_c *ca_modp_id2c(id object)
{
    return [object ca_modp_reference];	/* write for inplace arithmetic */
}

static id ca_modp_t2id(id proto,ca_modp_t f)
{
    return [proto ca_modp_value:f];	/* new object containing copy of "f" */
}

static id ca_modp_c2id(id proto,ca_modp_c* c)
{
    return [proto ca_modp_reference:c];/* new object containing "c" */
}

static ca_modp_t ca_modp_fromintvalue(int i,ca_modp_args args)
{
    ca_modp_t red;
    
    if (i==0) return 0;
    if (i <0) {
	red = -i;
	red %= args.p;	/* it's dangerous to % int & unsigned */
	red = (red)?args.p-red:0;
    } else {
	red  = i;
	red %= args.p;	/* dito */
    }
    
    return red;
}

static int ca_modp_check(ca_modp_t self,ca_modp_args args)
{
    assert(self < args.p);return 1;
}

static ca_modp_t ca_modp_zero(ca_modp_args args)
{
    return 0;
}

static ca_modp_t ca_modp_one(ca_modp_args args)
{
    return 1;
}

static unsigned ca_modp_hash(ca_modp_t a,ca_modp_args args)
{
    return a;
}

static int ca_modp_iseq(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a == b);
}

static int ca_modp_isop(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a==0)?(b==0):(a+b==args.p);
}

static int ca_modp_iszero(ca_modp_t a,ca_modp_args args)
{
    return a==0;
}

static int ca_modp_isone(ca_modp_t a,ca_modp_args args)
{
    return a==1;
}

static int ca_modp_isminusone(ca_modp_t a,ca_modp_args args)
{
    return (a+1)==args.p;
}

static ca_modp_t ca_modp_neg(ca_modp_t a,ca_modp_args args)
{
    assert(ca_modp_check(a,args));return (a)?args.p-a:0;
}

static ca_modp_t ca_modp_dbl(ca_modp_t a,ca_modp_args args)
{
    ca_modp_t c = a+a;
    assert(args.p != 2 && ca_modp_check(a,args));
    if (c) return (c<args.p)?c:c-args.p; else return 0;
}

static ca_modp_t ca_modp_add(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    ca_modp_t c = a+b;
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return ((c>=args.p)?c-args.p:c); /* compiles better than c<p?c:c-p */
}

static ca_modp_t ca_modp_sub(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(ca_modp_check(a,args) && ca_modp_check(b,args));
    return (a>=b)?(a-b):(args.p-(b-a));
}

static ca_modp_t ca_modp_mul(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    unsigned c = a * b;		/* limitation to 16bits to be able to do this
				 * russian peasant 50% slower
                                 */
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_sqr(ca_modp_t a,ca_modp_args args)
{
    unsigned c = a * a;
    assert(sizeof(unsigned) == 2 * sizeof(ca_modp_t));
    return c % args.p;
}

static ca_modp_t ca_modp_inv(ca_modp_t x,ca_modp_args args)
{
    unsigned q,s;
    int u1,u2,us;  /* intermediate values can be negative */
    ca_modp_t a,b,inv;
    
    assert(x != 0 && x != 1 && x+1 != args.p);
    
    a = x;b = args.p;u1 = 1;u2 = 0;
    while (b) {
	s  = a%b;q = a/b;
	a  = b;b = s;
	u1 -= q*u2;			
	us  = u1;u1 = u2;u2 = us;
    }
    
    inv = ((u1<0)?u1+args.p:u1); /* keep result positive */
    
    assert(ca_modp_check(inv,args));
    assert(ca_modp_isone(ca_modp_mul(x,inv,args),args));
    return inv;
}

static ca_modp_t ca_modp_div(ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    assert(b!=0);
    return ca_modp_mul(a,ca_modp_inv(b,args),args);
}

static ca_modp_t ca_modp_pow(ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_t o;
    
    if (n==0) return ca_modp_one(args);
    if (n==1) return a;
    if (n==2) return ca_modp_sqr(a,args);
    
    o = ca_modp_one(args);
    while (n>1) {
	if (n&1) o = ca_modp_mul(o,a,args);
	a = ca_modp_sqr(a,args);
	n >>= 1;
    }
    
    return ca_modp_mul(o,a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_copy_val.fab should not be included"
#endif

static void ca_modp_clear(ca_modp_c *c,ca_modp_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_modp_t)0xcafebabe;
    #endif
}

static void ca_modp_copy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_deepcopy(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = a;
}

static void ca_modp_move(ca_modp_c *c,ca_modp_c *a,ca_modp_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_modp_t)0xcafebabe;
    #endif
}

#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static void ca_modp_oozero(ca_modp_c *c,ca_modp_args args)
{
    *c = ca_modp_zero(args);
}

static void ca_modp_inzero(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_zero(args);
}

static void ca_modp_inneg(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_neg(*a,args);
}

static void ca_modp_ooneg(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_neg(a,args);
}

static ca_modp_t ca_modp_dblv(ca_modp_t a,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_dbl(a,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_indbl(ca_modp_c *a,int v,ca_modp_args args)
{
    *a = ca_modp_dblv(*a,v,args);
}

static void ca_modp_oodbl(ca_modp_c *c,ca_modp_t a,int v,ca_modp_args args)
{
    *c = ca_modp_dblv(a,v,args);
}

static ca_modp_t ca_modp_addvw(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    ca_modp_t c;assert(v*v == +1 && w*w == 1);
    
    if (v == +1) {
	if (w == +1) {
	    c = ca_modp_add(a,b,args);
	} else {
	    c = ca_modp_sub(a,b,args);
	}
    } else {
	if (w == +1) {
	    c = ca_modp_sub(b,a,args);
	} else {
	    c = ca_modp_neg(ca_modp_add(a,b,args),args);
	}
    }
    
    return c;
}

static void ca_modp_inadd(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addvw(*a,v,b,w,args);
}

static void ca_modp_ooadd(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addvw(a,v,b,w,args);
}
#if !ca_modp_isvalue
#error "ca_modp_mul_val.fab should not be included"
#endif

static void ca_modp_insqr(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_sqr(*a,args);
}

static void ca_modp_oosqr(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_sqr(a,args);
}

static void ca_modp_inmul(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_mul(*a,b,args);
}

static void ca_modp_oomul(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_mul(a,b,args);
}
#if !ca_modp_isvalue
#error "ca_modp_inv_val.fab should not be included"
#endif

static void ca_modp_ininv(ca_modp_c *a,ca_modp_args args)
{
    *a = ca_modp_inv(*a,args);
}

static void ca_modp_ooinv(ca_modp_c *c,ca_modp_t a,ca_modp_args args)
{
    *c = ca_modp_inv(a,args);
}

#if !ca_modp_isvalue
#error "ca_modp_div_val.fab should not be included"
#endif

static void ca_modp_indiv(ca_modp_c *a,ca_modp_t b,ca_modp_args args)
{
    *a = ca_modp_div(*a,b,args);
}

static BOOL ca_modp_oodiv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,ca_modp_args args)
{
    *c = ca_modp_div(a,b,args);return SUCCESS;
}
#if !ca_modp_isvalue
#error "ca_modp_add_val.fab should not be included"
#endif

static ca_modp_t ca_modp_mulv(ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    ca_modp_t c = ca_modp_mul(a,b,args);return (v==+1)?c:ca_modp_neg(c,args);
}

static void ca_modp_inmulv(ca_modp_c *a,ca_modp_t b,int v,ca_modp_args args)
{
    *a = ca_modp_mulv(*a,b,v,args);
}

static void ca_modp_oomulv(ca_modp_c *c,ca_modp_t a,ca_modp_t b,int v,ca_modp_args args)
{
    *c = ca_modp_mulv(a,b,v,args);
}

static ca_modp_t ca_modp_addmul(ca_modp_t a,int v,ca_modp_t b,ca_modp_t s,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_mul(b,s,args),w,args);
}

static ca_modp_t ca_modp_addsqr(ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(a,v,ca_modp_sqr(b,args),w,args);
}

static ca_modp_t ca_modp_muladd(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,b,w,args);
}

static ca_modp_t ca_modp_muladdmul(ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    return ca_modp_addvw(ca_modp_mul(a,s,args),v,ca_modp_mul(b,t,args),w,args);
}

static void ca_modp_inaddmul(ca_modp_c *a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_addmul(*a,v,b,t,w,args);
}

static void ca_modp_ooaddmul(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_addmul(a,v,b,t,w,args);
}

static void ca_modp_inaddsqr(ca_modp_c *a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_addsqr(*a,v,b,w,args);
}

static void ca_modp_ooaddsqr(ca_modp_c *c,ca_modp_t a,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_addsqr(a,v,b,w,args);
}

static void ca_modp_inmuladd(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *a = ca_modp_muladd(*a,s,v,b,w,args);
}

static void ca_modp_oomuladd(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,int w,ca_modp_args args)
{
    *c = ca_modp_muladd(a,s,v,b,w,args);
}

static void ca_modp_inmuladdmul(ca_modp_c *a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *a = ca_modp_muladdmul(*a,s,v,b,t,w,args);
}

static void ca_modp_oomuladdmul(ca_modp_c *c,ca_modp_t a,ca_modp_t s,int v,ca_modp_t b,ca_modp_t t,int w,ca_modp_args args)
{
    *c = ca_modp_muladdmul(a,s,v,b,t,w,args);
}
static void ca_modp_args_fileOut(id aFiler,ca_modp_args *args) {}
static void ca_modp_args_fileIn(id aFiler,ca_modp_args *args) {}

static void ca_modp_fileOut(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileOut:c type:'S'];
}

static void ca_modp_fileIn(id aFiler,ca_modp_c *c,ca_modp_args args)
{
    [aFiler fileIn:c type:'S'];
}


/*
 * Computer Algebra Kit (c) 1993,98 by Comp.Alg.Objects.  All Rights Reserved.
 */

static int ca_unimonom_size(ca_unimonom_args args)
{
    return 1;
}

static int ca_unimonom_bsize(ca_unimonom_args args)
{
    return sizeof(ca_unimonom_t);
}

static int ca_unimonom_check(ca_unimonom_t self,ca_unimonom_args args)
{
    assert(self >= 0);return 1;
}

static void ca_unimonom_clear(ca_unimonom_c *c,ca_unimonom_args args)
{
    /* nothing to clear, but the following might help... */
    #ifndef DEBUG
    *c = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_copy(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_args args)
{
    *c = a;
}

static void ca_unimonom_move(ca_unimonom_c *c,ca_unimonom_c *a,ca_unimonom_args args)
{
    *c = *a;
    #ifndef DEBUG
    *a = (ca_unimonom_t)0xcafebabe;
    #endif
}

static void ca_unimonom_args_fileOut(id aFiler,ca_unimonom_args *args) {}
static void ca_unimonom_args_fileIn(id aFiler,ca_unimonom_args *args) {}

static void ca_unimonom_fileOut(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileOut:c type:'i'];
}

static void ca_unimonom_fileIn(id aFiler,ca_unimonom_c *c,ca_unimonom_args args)
{
    [aFiler fileIn:c type:'i'];
}

static unsigned ca_unimonom_hash(ca_unimonom_t self,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_deg(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self;
}

static int ca_unimonom_isone(ca_unimonom_t self ,ca_unimonom_args args)
{
    return self == 0;
}

static int ca_unimonom_iseq(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == b;
}

static int ca_unimonom_iseqmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a == x+b;
}

static int ca_unimonom_cmp(ca_unimonom_t a,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-b;
}

static int ca_unimonom_cmpmul(ca_unimonom_t a,ca_unimonom_t x,ca_unimonom_t b ,ca_unimonom_args args)
{
    return a-x-b;
}

static void ca_unimonom_oomul(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    *c = a + b;
}

static void ca_unimonom_inmul(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    *c += b;
}

static BOOL ca_unimonom_oodiv(ca_unimonom_c *c,ca_unimonom_t a,ca_unimonom_t b,ca_unimonom_args args)
{
    if (a < b) {
	return FAILURE;
    } else {
	*c = a - b;
	return SUCCESS;
    }
}

static void ca_unimonom_indiv(ca_unimonom_c *c,ca_unimonom_t b,ca_unimonom_args args)
{
    if (*c < b) { fprintf(stderr,"Division not Exact\n");abort(); } /* should stay */
    *c -= b;
}

#if ca_modp_pervalue
static ca_modp_t ca_modpseq_tval(ca_modpseq_t self)
{
    return *self;
}
static ca_modp_c *ca_modpseq_cval(ca_modpseq_t self)
{
    return self;
}
#else
static ca_modp_t ca_modpseq_tval(ca_modpseq_t self)
{
    return self;
}
static ca_modp_c *ca_modpseq_cval(ca_modpseq_t self)
{
    return self;
}
#endif

static int ca_modpseq_check(ca_modpseq_t self,int n,ca_modp_args args)
{
    while (n--) {
	ca_modp_check(ca_modpseq_tval(self),args);
	self += ca_modp_size(args);
    }
    return 1;
}

static ca_modpseq_t ca_modpseq_alloc(int n,ca_modp_args args)
{
    return (ca_modpseq_t)OC_Malloc(n * ca_modp_bsize(args));
}

static ca_modpseq_t ca_modpseq_realloc(ca_modpseq_t self,int n,ca_modp_args args)
{
    return (ca_modpseq_t)OC_Realloc(self,n * ca_modp_bsize(args));
}

static ca_modpseq_t ca_modpseq_free(ca_modpseq_t self,ca_modp_args args)
{
    OC_Free(self);return NULL;
}

static void ca_modpseq_oozero(ca_modpseq_t c,int n,ca_modp_args args)
{
    while (n--) {
	ca_modp_oozero(ca_modpseq_cval(c),args);
	c += ca_modp_size(args);
    }
}

static void ca_modpseq_copy(ca_modpseq_t c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_copy(ca_modpseq_cval(c),ca_modpseq_tval(a),args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_deepcopy(ca_modpseq_t c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_deepcopy(ca_modpseq_cval(c),ca_modpseq_tval(a),args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

#if ca_modp_isvalue
static void ca_modpseq_clear(ca_modpseq_t self,int n,ca_modp_args args)
{
    /* nothing to clear */
}
#else
static void ca_modpseq_clear(ca_modpseq_t self,int n,ca_modp_args args)
{
    while (n--) {
	ca_modp_clear(self,args);
	self+=ca_modp_size(args);
    }
}
#endif /* ca_modp_isvalue */

static int ca_modpseq_iseq(ca_modpseq_t a,ca_modpseq_t b,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_modp_iseq(ca_modpseq_tval(a),ca_modpseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_modp_size(args);b+=ca_modp_size(args);
	}
    }
    return 1;
}

static unsigned ca_modpseq_hash(ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    return (n)?ca_modp_hash(ca_modpseq_tval(a),args):0;
}

static void ca_modpseq_place(ca_modpseq_t self,ca_modp_t a,int i,ca_modp_args args)
{
    ca_modp_c *c = ca_modpseq_cval(self + i * ca_modp_size(args));
    ca_modp_clear(c,args);ca_modp_copy(c,a,args);
}

static int ca_modpseq_insertlast(ca_modpseq_t self,ca_modp_t a,int n,ca_modp_args args)
{
    ca_modp_c *c;
    assert(n >= 0);
    c = ca_modpseq_cval(self + n * ca_modp_size(args));
    ca_modp_copy(c,a,args);return ++n;
}

static int ca_modpseq_insertfirst(ca_modpseq_t self,ca_modp_t a,int n,ca_modp_args args)
{
    int j;
    assert(n > 0);
    self += n * ca_modp_size(args);
    j=n;while(j--) {
	ca_modpseq_t t;
	t = self - ca_modp_size(args);
	ca_modp_move(ca_modpseq_cval(self),ca_modpseq_cval(t),args);
	self = t;
    }
    ca_modp_copy(ca_modpseq_cval(self),a,args);
    return ++n;
}

static int ca_modpseq_insertat(ca_modpseq_t self,ca_modp_t a,int i,int n,ca_modp_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_modp_size(args);
    return i + ca_modpseq_insertfirst(self,a,n-i,args);
}

static int ca_modpseq_freelast(ca_modpseq_t self,int n,ca_modp_args args)
{
    ca_modp_c *c;assert(0 < n);
    --n;c = ca_modpseq_cval(self + n * ca_modp_size(args));
    ca_modp_clear(c,args);return n;
}

static int ca_modpseq_freefirst(ca_modpseq_t self,int n,ca_modp_args args)
{
    int j;
    assert(0 < n);
    ca_modp_clear(ca_modpseq_cval(self),args);
    --n;j=n;while(j--) {
	ca_modpseq_t t;
	t = self+ca_modp_size(args);
	ca_modp_move(ca_modpseq_cval(self),ca_modpseq_cval(t),args);
	self = t;
    }
    return n;
}

static int ca_modpseq_freeat(ca_modpseq_t self,int i,int n,ca_modp_args args)
{
    assert(0 <= i && i < n);
    self += i*ca_modp_size(args);
    return i + ca_modpseq_freefirst(self,n-i,args);
}

static ca_modp_t ca_modpseq_at(ca_modpseq_t self,int i,ca_modp_args args)
{
    ca_modp_t res;assert(0 <= i);
    res = ca_modpseq_tval(self + i * ca_modp_size(args));
    assert(ca_modp_check(res,args));
    return res;
}

static ca_modp_t ca_modpseq_last(ca_modpseq_t self,int n,ca_modp_args args)
{
    return ca_modpseq_at(self,n-1,args);
}

static ca_modp_c *ca_modpseq_cvalat(ca_modpseq_t self,int i,ca_modp_args args)
{
    assert(0 <= i);return ca_modpseq_cval(self + i * ca_modp_size(args));
}

static int ca_modpseq_iszero(ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	if (ca_modp_iszero(ca_modpseq_tval(a),args)) {
	    a+=ca_modp_size(args);
	} else {
	    return 0;
	}
    }
    return 1;
}

static int ca_modpseq_isop(ca_modpseq_t a,ca_modpseq_t b,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b);
    while (n--) {
	if (!ca_modp_isop(ca_modpseq_tval(a),ca_modpseq_tval(b),args)) {
	    return 0;
	} else {
	    a+=ca_modp_size(args);b+=ca_modp_size(args);
	}
    }
    return 1;
}

static void ca_modpseq_ooneg(ca_modpseq_t c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_ooneg(ca_modpseq_cval(c),ca_modpseq_tval(a),args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_inneg(ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_inneg(ca_modpseq_cval(a),args);
	a += ca_modp_size(args);
    }
}

static void ca_modpseq_oonegv(ca_modpseq_t c,ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(v*v == 1);if (v == -1) ca_modpseq_ooneg(c,a,n,args);
    else ca_modpseq_copy(c,a,n,args);
}

static void ca_modpseq_innegv(ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(v*v == 1);if (v == -1) ca_modpseq_inneg(a,n,args);
}

static void ca_modpseq_oodbl(ca_modpseq_t c,ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_oodbl(ca_modpseq_cval(c),ca_modpseq_tval(a),v,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_indbl(ca_modpseq_t a,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_indbl(ca_modpseq_cval(a),v,args);
	a += ca_modp_size(args);
    }
}

static void ca_modpseq_ooadd(ca_modpseq_t c,ca_modpseq_t a,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_ooadd(ca_modpseq_cval(c),ca_modpseq_tval(a),v,ca_modpseq_tval(b),w,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_inadd(ca_modpseq_t a,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && a != b && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inadd(ca_modpseq_cval(a),v,ca_modpseq_tval(b),w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulsc(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_oomul(ca_modpseq_cval(c),ca_modpseq_tval(a),s,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulsc(ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_inmul(ca_modpseq_cval(a),s,args);
	a+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulscv(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_oomulv(ca_modpseq_cval(c),ca_modpseq_tval(a),s,v,args);
	c+=ca_modp_size(args);a+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulscv(ca_modpseq_t a,ca_modp_t s,int v,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1);
    while (n--) {
	ca_modp_inmulv(ca_modpseq_cval(a),s,v,args);
	a+=ca_modp_size(args);
    }
}

static void ca_modpseq_ooaddmulsc(ca_modpseq_t c,ca_modpseq_t a,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_ooaddmul(ca_modpseq_cval(c),ca_modpseq_tval(a),v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);c+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulscadd(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_oomuladd(ca_modpseq_cval(c),ca_modpseq_tval(a),s,v,ca_modpseq_tval(b),w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);c+=ca_modp_size(args);
    }
}

static void ca_modpseq_oomulscaddmulsc(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_oomuladdmul(ca_modpseq_cval(c),ca_modpseq_tval(a),s,v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);c+=ca_modp_size(args);
    }
}

static void ca_modpseq_inaddmulsc(ca_modpseq_t a,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inaddmul(ca_modpseq_cval(a),v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulscadd(ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inmuladd(ca_modpseq_cval(a),s,v,ca_modpseq_tval(b),w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static void ca_modpseq_inmulscaddmulsc(ca_modpseq_t a,ca_modp_t s,int v,ca_modpseq_t b,ca_modp_t t,int w,int n,ca_modp_args args)
{
    assert(n >= 0 && v*v == 1 && w*w == 1);
    while (n--) {
	ca_modp_inmuladdmul(ca_modpseq_cval(a),s,v,ca_modpseq_tval(b),t,w,args);
	a+=ca_modp_size(args);b+=ca_modp_size(args);
    }
}

static BOOL ca_modpseq_oodivsc(ca_modpseq_t c,ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    int i;
    BOOL failed;
    ca_modpseq_t orgc = c;
    assert(n >= 0);
   
    for(i=0;i<n;i++) {
    
	failed = ca_modp_oodiv(ca_modpseq_cval(c),ca_modpseq_tval(a),s,args);
	
	if (failed) {
	    ca_modpseq_clear(orgc,i - 1,args);
	    return FAILURE;
	} else {
	    c+=ca_modp_size(args);a+=ca_modp_size(args);
	}
    }
    
    return SUCCESS;
}

static void ca_modpseq_indivsc(ca_modpseq_t a,ca_modp_t s,int n,ca_modp_args args)
{
    assert(n >= 0);
    while (n--) {
	ca_modp_indiv(ca_modpseq_cval(a),s,args);
	a+=ca_modp_size(args);
    }
}

static void ca_modpseq_oodotsqr(ca_modp_c *c,ca_modpseq_t a,int n,ca_modp_args args)
{
    assert(n >= 0);
    ca_modp_oozero(c,args);
    while (n--) {
	ca_modp_inaddsqr(c,+1,ca_modpseq_tval(a),+1,args);a += ca_modp_size(args);
    }
}

static void ca_modpseq_oodotmul(ca_modp_c *c,ca_modpseq_t a,ca_modpseq_t b,int n,ca_modp_args args)
{
    assert(n >= 0);
    ca_modp_oozero(c,args);
    while (n--) {
	ca_modp_inaddmul(c,+1,ca_modpseq_tval(a),ca_modpseq_tval(b),+1,args);
	a += ca_modp_size(args);b += ca_modp_size(args);
    }
}

static void ca_modpseq_args_fileOut(id aFiler,ca_modp_args *args)
{
    ca_modp_args_fileOut(aFiler,args);
}
static void ca_modpseq_args_fileIn(id aFiler,ca_modp_args *args)
{
    ca_modp_args_fileIn(aFiler,args);
}

static void ca_modpseq_fileOut(id aFiler,ca_modpseq_t self,int n,ca_modp_args args)
{
    assert(n>=0);
    while (n--) {
	ca_modp_fileOut(aFiler,ca_modpseq_cval(self),args);
	self += ca_modp_size(args);
    }
}

static void ca_modpseq_fileIn(id aFiler,ca_modpseq_t self,int n,ca_modp_args args)
{
    assert(n>=0);
    while (n--) {
	ca_modp_fileIn(aFiler,ca_modpseq_cval(self),args);
	self += ca_modp_size(args);
    }
}

static int ca_modp_vardnsrecdegdnspolseq_check(ca_modp_vardnsrecdegdnspolseq_t self,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    if (n) {
	ca_modp_t lcf = ca_modpseq_last(self,n,args.sargs);
	assert(!ca_modp_iszero(lcf,args.sargs));
    }
    
    return ca_modpseq_check(self,n,args.sargs);
}

static int ca_modp_vardnsrecdegdnspolseq_copy(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_copy(c,a,n,args.sargs);return n;
}

static int ca_modp_vardnsrecdegdnspolseq_deepcopy(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_deepcopy(c,a,n,args.sargs);return n;
}

static int ca_modp_vardnsrecdegdnspolseq_count(ca_modp_vardnsrecdegdnspolseq_t self,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    int count = 0;
    while (n--) {
	if (!ca_modp_iszero(ca_modpseq_tval(self),args.sargs)) count++;
	self += ca_modp_size(args.sargs);
    }
    return count;
}

static int ca_modp_vardnsrecdegdnspolseq_ord(ca_modp_vardnsrecdegdnspolseq_t self,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    int ord = -1;
    while (n--) {
	if (ca_modp_iszero(ca_modpseq_tval(self),args.sargs)) {
	    ord++;self += ca_modp_size(args.sargs);
	} else {
	    break;
	}
    }
    return ord;
}

static int ca_modp_vardnsrecdegdnspolseq_iseq(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_vardnsrecdegdnspolseq_t b,int bn,ca_modp_vardnsrecdegdnspol_args args)
{
    return (an==bn)?ca_modpseq_iseq(a,b,an,args.sargs):0;
}

static int ca_modp_vardnsrecdegdnspolseq_isop(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_vardnsrecdegdnspolseq_t b,int bn,ca_modp_vardnsrecdegdnspol_args args)
{
    return (an==bn)?ca_modpseq_isop(a,b,an,args.sargs):0;
}

static int ca_modp_vardnsrecdegdnspolseq_hash(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modpseq_hash(a,an,args.sargs);
}

static int ca_modp_vardnsrecdegdnspolseq_isone(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_vardnsrecdegdnspol_args args)
{
    return an == 1 && ca_modp_isone(ca_modpseq_tval(a),args.sargs);
}

static int ca_modp_vardnsrecdegdnspolseq_isminusone(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_vardnsrecdegdnspol_args args)
{
    return an == 1 && ca_modp_isminusone(ca_modpseq_tval(a),args.sargs);
}

static int ca_modp_vardnsrecdegdnspolseq_removezeroes(ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    a += n * ca_modp_size(args.sargs);
    while (n--) {
	a -= ca_modp_size(args.sargs);
	if (ca_modp_iszero(ca_modpseq_tval(a),args.sargs)) {
	    ca_modp_clear(ca_modpseq_cval(a),args.sargs);
	} else {
	    break;
	}
    }
    return n+1;
}

static int ca_modp_vardnsrecdegdnspolseq_delete(ca_modp_vardnsrecdegdnspolseq_t self,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    n = n - 1;assert(n >= 0);
    ca_modp_clear(ca_modpseq_cvalat(self,n,args.sargs),args.sargs);
    return ca_modp_vardnsrecdegdnspolseq_removezeroes(self,n,args);
}

static int ca_modp_vardnsrecdegdnspolseq_insert(ca_modp_vardnsrecdegdnspolseq_t self,int n,int e,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_iszero(s,args.sargs) == 0);
    if (e < n) {
	ca_modp_inadd(ca_modpseq_cvalat(self,e,args.sargs),+1,s,+1,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(self,n,args);
    } else {
	ca_modpseq_oozero(self + n * ca_modp_size(args.sargs),e + 1 - n,args.sargs);
	ca_modp_inadd(ca_modpseq_cvalat(self,e,args.sargs),+1,s,+1,args.sargs);
	return e + 1;
    }
}

static int ca_modp_vardnsrecdegdnspolseq_ooneg(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_ooneg(c,a,n,args.sargs);return n;
}

static int ca_modp_vardnsrecdegdnspolseq_inneg(ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_inneg(a,n,args.sargs);return n;
}

static int ca_modp_vardnsrecdegdnspolseq_oodbl(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,int v,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_oodbl(c,a,n,v,args.sargs);return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,n,args);
}

static int ca_modp_vardnsrecdegdnspolseq_indbl(ca_modp_vardnsrecdegdnspolseq_t a,int n,int v,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_indbl(a,n,v,args.sargs);return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,n,args);
}

static int ca_modp_vardnsrecdegdnspolseq_ooadd(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int an,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_modpseq_ooadd(c,a,v,b,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_ooadd(c,a,v,b,w,an,args.sargs);
	    c += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oonegv(c,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_ooadd(c,a,v,b,w,bn,args.sargs);
	    c += bn*ca_modp_size(args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_oonegv(c,a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_inadd(ca_modp_vardnsrecdegdnspolseq_t a,int an,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_modpseq_inadd(a,v,b,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_inadd(a,v,b,w,an,args.sargs);
	    a += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oonegv(a,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_inadd(a,v,b,w,bn,args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_innegv(a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_oomulsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_oomulsc(c,a,s,n,args.sargs);return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,n,args);
}

static int ca_modp_vardnsrecdegdnspolseq_inmulsc(ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_inmulsc(a,s,n,args.sargs);return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,n,args);
}

static int ca_modp_vardnsrecdegdnspolseq_ooaddmulsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int an,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_modpseq_ooaddmulsc(c,a,v,b,t,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_ooaddmulsc(c,a,v,b,t,w,an,args.sargs);
	    c += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oomulscv(c,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_ooaddmulsc(c,a,v,b,t,w,bn,args.sargs);
	    c += bn*ca_modp_size(args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_oonegv(c,a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_inaddmulsc(ca_modp_vardnsrecdegdnspolseq_t a,int an,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_modpseq_inaddmulsc(a,v,b,t,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_inaddmulsc(a,v,b,t,w,an,args.sargs);
	    a += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oomulscv(a,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_inaddmulsc(a,v,b,t,w,bn,args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_innegv(a,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_oomulscadd(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_modpseq_oomulscadd(c,a,s,v,b,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_oomulscadd(c,a,s,v,b,w,an,args.sargs);
	    c += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oonegv(c,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_oomulscadd(c,a,s,v,b,w,bn,args.sargs);
	    c += bn*ca_modp_size(args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_oomulscv(c,a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_inmulscadd(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_modpseq_inmulscadd(a,s,v,b,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_inmulscadd(a,s,v,b,w,an,args.sargs);
	    a += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oonegv(a,b,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_inmulscadd(a,s,v,b,w,bn,args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_inmulscv(a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_oomulscaddmulsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn) {
	ca_modpseq_oomulscaddmulsc(c,a,s,v,b,t,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_oomulscaddmulsc(c,a,s,v,b,t,w,an,args.sargs);
	    c += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oomulscv(c,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_oomulscaddmulsc(c,a,s,v,b,t,w,bn,args.sargs);
	    c += bn*ca_modp_size(args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_oomulscv(c,a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_inmulscaddmulsc(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);

    if (an == bn) {
	ca_modpseq_inmulscaddmulsc(a,s,v,b,t,w,an,args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn) {
	    ca_modpseq_inmulscaddmulsc(a,s,v,b,t,w,an,args.sargs);
	    a += an*ca_modp_size(args.sargs);
	    b += an*ca_modp_size(args.sargs);
	    ca_modpseq_oomulscv(a,b,t,w,bn-an,args.sargs);
	    return bn;
	} else {
	    ca_modpseq_inmulscaddmulsc(a,s,v,b,t,w,bn,args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    ca_modpseq_inmulscv(a,s,v,an-bn,args.sargs);
	    return an;
	}
    }
}

static BOOL ca_modp_vardnsrecdegdnspolseq_oodivsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modpseq_oodivsc(c,a,s,n,args.sargs);
}

static int ca_modp_vardnsrecdegdnspolseq_indivsc(ca_modp_vardnsrecdegdnspolseq_t a,int n,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_indivsc(a,s,n,args.sargs);return n;
}

static int ca_modp_vardnsrecdegdnspolseq_oomulmonomsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,int m,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_oozero(c,m,args.sargs);
    c += m * ca_modp_size(args.sargs);
    ca_modpseq_oomulscv(c,a,s,v,n,args.sargs);
    return n + m;
}

static BOOL ca_modp_vardnsrecdegdnspolseq_oodivmonomsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int n,int m,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(n >= m);
    a += m * ca_modp_size(args.sargs);
    return ca_modpseq_oodivsc(c,a,s,n - m,args.sargs);
}

static int ca_modp_vardnsrecdegdnspolseq_ooaddmulmonomsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int an,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_modpseq_oonegv(c,a,v,m,args.sargs);
	c += m*ca_modp_size(args.sargs);
	a += m*ca_modp_size(args.sargs);
	assert(an >= bn);ca_modpseq_ooaddmulsc(c,a,v,b,t,w,bn,args.sargs);
	c -= m*ca_modp_size(args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_modpseq_oonegv(c,a,v,an,args.sargs);
		c += an*ca_modp_size(args.sargs);
		a += an*ca_modp_size(args.sargs);
		ca_modpseq_oozero(c,m-an,args.sargs);
		c += m*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(c,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_modpseq_oonegv(c,a,v,m,args.sargs);
		c += m*ca_modp_size(args.sargs);
		a += m*ca_modp_size(args.sargs);
		assert(bn > an-m);ca_modpseq_ooaddmulsc(c,a,v,b,t,w,an-m,args.sargs);
		c += (an-m)*ca_modp_size(args.sargs);
		a += (an-m)*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(c,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_modpseq_oonegv(c,a,v,m,args.sargs);
	    c += m*ca_modp_size(args.sargs);
	    a += m*ca_modp_size(args.sargs);
	    assert(an >= bn);ca_modpseq_ooaddmulsc(c,a,v,b,t,w,bn,args.sargs);
	    c += bn*ca_modp_size(args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    assert(an > bn+m);ca_modpseq_oonegv(c,a,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_inaddmulmonomsc(ca_modp_vardnsrecdegdnspolseq_t a,int an,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_modpseq_innegv(a,v,m,args.sargs);
	a += m*ca_modp_size(args.sargs);
	assert(an >= bn);ca_modpseq_inaddmulsc(a,v,b,t,w,bn,args.sargs);
	a -= m*ca_modp_size(args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_modpseq_innegv(a,v,an,args.sargs);
		a += an*ca_modp_size(args.sargs);
		ca_modpseq_oozero(a,m-an,args.sargs);
		a += m*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(a,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_modpseq_innegv(a,v,m,args.sargs);
		a += m*ca_modp_size(args.sargs);
		assert(bn > an-m);ca_modpseq_inaddmulsc(a,v,b,t,w,an-m,args.sargs);
		a += (an-m)*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(a,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_modpseq_innegv(a,v,m,args.sargs);
	    a += m*ca_modp_size(args.sargs);
	    assert(an >= bn);ca_modpseq_inaddmulsc(a,v,b,t,w,bn,args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    assert(an > bn+m);ca_modpseq_innegv(a,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_oomulscaddmulmonomsc(ca_modp_vardnsrecdegdnspolseq_t c,ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_modpseq_oomulscv(c,a,s,v,m,args.sargs);
	c += m*ca_modp_size(args.sargs);
	a += m*ca_modp_size(args.sargs);
	assert(an >= bn);ca_modpseq_oomulscaddmulsc(c,a,s,v,b,t,w,bn,args.sargs);
	c -= m*ca_modp_size(args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(c,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_modpseq_oomulscv(c,a,s,v,an,args.sargs);
		c += an*ca_modp_size(args.sargs);
		a += an*ca_modp_size(args.sargs);
		ca_modpseq_oozero(c,m-an,args.sargs);
		c += m*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(c,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_modpseq_oomulscv(c,a,s,v,m,args.sargs);
		c += m*ca_modp_size(args.sargs);
		a += m*ca_modp_size(args.sargs);
		assert(bn > an-m);ca_modpseq_oomulscaddmulsc(c,a,s,v,b,t,w,an-m,args.sargs);
		c += (an-m)*ca_modp_size(args.sargs);
		a += (an-m)*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(c,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_modpseq_oomulscv(c,a,s,v,m,args.sargs);
	    c += m*ca_modp_size(args.sargs);
	    a += m*ca_modp_size(args.sargs);
	    assert(an >= bn);ca_modpseq_oomulscaddmulsc(c,a,s,v,b,t,w,bn,args.sargs);
	    c += bn*ca_modp_size(args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    assert(an > bn+m);ca_modpseq_oomulscv(c,a,s,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}

static int ca_modp_vardnsrecdegdnspolseq_inmulscaddmulmonomsc(ca_modp_vardnsrecdegdnspolseq_t a,int an,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspolseq_t b,int bn,int m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(v*v == 1 && w*w == 1);
    
    if (an == bn + m) {
	assert(an >= m);ca_modpseq_inmulscv(a,s,v,m,args.sargs);
	a += m*ca_modp_size(args.sargs);
	assert(an >= bn);ca_modpseq_inmulscaddmulsc(a,s,v,b,t,w,bn,args.sargs);
	a -= m*ca_modp_size(args.sargs);
	return ca_modp_vardnsrecdegdnspolseq_removezeroes(a,an,args);
    } else {
	if (an < bn + m) {
	    if (an < m) {
		ca_modpseq_inmulscv(a,s,v,an,args.sargs);
		a += an*ca_modp_size(args.sargs);
		ca_modpseq_oozero(a,m-an,args.sargs);
		a += m*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(a,b,t,w,bn,args.sargs);
	    } else {
		assert(an >= m);ca_modpseq_inmulscv(a,s,v,m,args.sargs);
		a += m*ca_modp_size(args.sargs);
		assert(bn > an-m);ca_modpseq_inmulscaddmulsc(a,s,v,b,t,w,an-m,args.sargs);
		a += (an-m)*ca_modp_size(args.sargs);
		ca_modpseq_oomulscv(a,b,t,w,bn-(an-m),args.sargs);
	    }
	    return bn + m;
	} else {
	    assert(an >= m);ca_modpseq_inmulscv(a,s,v,m,args.sargs);
	    a += m*ca_modp_size(args.sargs);
	    assert(an >= bn);ca_modpseq_inmulscaddmulsc(a,s,v,b,t,w,bn,args.sargs);
	    a += bn*ca_modp_size(args.sargs);
	    assert(an > bn+m);ca_modpseq_inmulscv(a,s,v,an-bn-m,args.sargs);
	    return an;
	}
    }
}
static void ca_modp_vardnsrecdegdnspol_ix_init(ca_modp_vardnsrecdegdnspol_ixt ix,id self)
{
    ca_modp_vardnsrecdegdnspol_t v = [self ca_modp_vardnsrecdegdnspol_value];
    ix->i = -1;ix->n = v->n;
}

static void ca_modp_vardnsrecdegdnspol_ix_clear(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    /* nothing to clear */
}

static int ca_modp_vardnsrecdegdnspol_ix_count(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    return ix->n; /* total number, zero or nonzero */
}

static int ca_modp_vardnsrecdegdnspol_ix_isempty(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    return ix->n == 0; 
}

static int ca_modp_vardnsrecdegdnspol_ix_ismatch(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    assert(-1 <= ix->i && ix->i <= ix->n);
    return (ix->i != -1 && ix->i != ix->n);
}

static void ca_modp_vardnsrecdegdnspol_ix_tofirst(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    ix->i = -1;	/* next = first */
}

static void ca_modp_vardnsrecdegdnspol_ix_tolast(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    ix->i = ix->n;	/* prev = last */
}

static void ca_modp_vardnsrecdegdnspol_ix_to(ca_modp_vardnsrecdegdnspol_ixt ix,int i)
{
    /* conversion to "n-i-1" done in elting routine... */
    if (i < 0) ix->i = -1;else if (i > ix->n) ix->i = ix->n;
    else ix->i = i-1; /* next -> at i */
}

static void ca_modp_vardnsrecdegdnspol_ix_next(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    if (ix->i == ix->n) {
	/* remain on position just after last term */
    } else {
	++(ix->i);
    }
}

static void ca_modp_vardnsrecdegdnspol_ix_prev(ca_modp_vardnsrecdegdnspol_ixt ix)
{
    if (ix->i == -1) {
	/* remain on position just before first term */
    } else {
	--(ix->i);
    }
}
static int ca_modp_vardnsrecdegdnspol_check(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(self->n <= self->c);
    return ca_modp_vardnsrecdegdnspolseq_check(self->ptr,self->n,args);
}

static ca_modp_vardnsrecdegdnspol_args ca_modp_vardnsrecdegdnspol_getargs(id self)
{
    ca_modp_vardnsrecdegdnspol_args res;
    res.sargs = ca_modp_getargs([[self termZero] coefficient]);
    return res; 
}

static ca_modp_vardnsrecdegdnspol_t ca_modp_vardnsrecdegdnspol_id2t(id object)
{
    return [object ca_modp_vardnsrecdegdnspol_value];
}

static ca_modp_vardnsrecdegdnspol_c *ca_modp_vardnsrecdegdnspol_id2c(id object)
{
    return [object ca_modp_vardnsrecdegdnspol_reference];
}

static id ca_modp_vardnsrecdegdnspol_t2id(id proto,ca_modp_vardnsrecdegdnspol_t f)
{
    return [proto ca_modp_vardnsrecdegdnspol_value:f];
}

static id ca_modp_vardnsrecdegdnspol_c2id(id proto,ca_modp_vardnsrecdegdnspol_c *r)
{
    return [proto ca_modp_vardnsrecdegdnspol_reference:r];
}

static int ca_modp_vardnsrecdegdnspol_isempty(ca_modp_vardnsrecdegdnspol_t self)
{
    return self->n == 0;
}

static void ca_modp_vardnsrecdegdnspol_init(ca_modp_vardnsrecdegdnspol_t self,int c,ca_modp_vardnsrecdegdnspol_args args)
{
    self->n   = 0;
    self->c   = c;
    self->ptr = ca_modpseq_alloc(c,args.sargs);
    assert(ca_modp_vardnsrecdegdnspol_check(self,args));
}

static void ca_modp_vardnsrecdegdnspol_copy(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
    ca_modp_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_modp_vardnsrecdegdnspolseq_copy(c->ptr,a->ptr,a->n,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_deepcopy(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
    ca_modp_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_modp_vardnsrecdegdnspolseq_deepcopy(c->ptr,a->ptr,a->n,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_move(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_args args)
{
    c->ptr = a->ptr;c->n = a->n;c->c = a->c;
    a->ptr = NULL;assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_destroy(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    /* needed e.g. after a failed oodiv operation */
    self->ptr = ca_modpseq_free(self->ptr,args.sargs);
}

static void ca_modp_vardnsrecdegdnspol_clear(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modpseq_clear(self->ptr,self->n,args.sargs);
    ca_modp_vardnsrecdegdnspol_destroy(self,args);
}

static int ca_modp_vardnsrecdegdnspol_count(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspolseq_count(self->ptr,self->n,args); /* num non-zero coefs */
}

static unsigned ca_modp_vardnsrecdegdnspol_hash(ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspolseq_hash(a->ptr,a->n,args);
}

static int ca_modp_vardnsrecdegdnspol_iseq(ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_t b,ca_modp_vardnsrecdegdnspol_args args)
{
    return (a==b)?YES:ca_modp_vardnsrecdegdnspolseq_iseq(a->ptr,a->n,b->ptr,b->n,args);
}

static int ca_modp_vardnsrecdegdnspol_isone(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspolseq_isone(self->ptr,self->n,args);
}

static int ca_modp_vardnsrecdegdnspol_isminusone(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspolseq_isminusone(self->ptr,self->n,args);
}

static int ca_modp_vardnsrecdegdnspol_deg(ca_modp_vardnsrecdegdnspol_t self ,ca_modp_vardnsrecdegdnspol_args args)
{
    return self->n - 1;
}

static int ca_modp_vardnsrecdegdnspol_ord(ca_modp_vardnsrecdegdnspol_t self ,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspolseq_ord(self->ptr,self->n,args);
}

static int ca_modp_vardnsrecdegdnspol_maxdeg(ca_modp_vardnsrecdegdnspol_t self ,ca_modp_vardnsrecdegdnspol_args args)
{
    return self->n - 1;
}

static int ca_modp_vardnsrecdegdnspol_mindeg(ca_modp_vardnsrecdegdnspol_t self ,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspolseq_ord(self->ptr,self->n,args);
}

static ca_modp_t ca_modp_vardnsrecdegdnspol_scalat(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_ixt ix,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_ix_ismatch(ix) && ix->n == self->n);
    return ca_modpseq_at(self->ptr,ix->n - ix->i - 1,args.sargs);
}

static int ca_modp_vardnsrecdegdnspol_expat(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_ixt ix,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_ix_ismatch(ix) && ix->n == self->n);
    return ix->n - ix->i - 1;
}

static ca_modp_t ca_modp_vardnsrecdegdnspol_scalfirst(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_isempty(self) == NO);
    return ca_modpseq_last(self->ptr,self->n,args.sargs);
}

static int ca_modp_vardnsrecdegdnspol_expfirst(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_isempty(self) == NO);
    return self->n - 1;
}

static void ca_modp_vardnsrecdegdnspol_expand(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    self->c   = self->n + self->c + 1;
    self->ptr = ca_modpseq_realloc(self->ptr,self->c,args.sargs);
    assert(self->n < self->c);
}

static void ca_modp_vardnsrecdegdnspol_remove(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_check(self,args) && ca_modp_vardnsrecdegdnspol_isempty(self) == NO);
    self->n = ca_modp_vardnsrecdegdnspolseq_delete(self->ptr,self->n,args);
    assert(ca_modp_vardnsrecdegdnspol_check(self,args));
}

static void ca_modp_vardnsrecdegdnspol_insert(ca_modp_vardnsrecdegdnspol_t self,int e,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(ca_modp_vardnsrecdegdnspol_check(self,args));
    while (self->c < e + 1) ca_modp_vardnsrecdegdnspol_expand(self,args);
    self->n = ca_modp_vardnsrecdegdnspolseq_insert(self->ptr,self->n,e,s,args);
    assert(ca_modp_vardnsrecdegdnspol_check(self,args));
}

static int ca_modp_vardnsrecdegdnspol_iszero(ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    return ca_modp_vardnsrecdegdnspol_isempty(self);
}

static int ca_modp_vardnsrecdegdnspol_isop(ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_t b,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(a != b);return ca_modp_vardnsrecdegdnspolseq_isop(a->ptr,a->n,b->ptr,b->n,args);
}

static void ca_modp_vardnsrecdegdnspol_ooneg(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_modp_vardnsrecdegdnspolseq_ooneg(c->ptr,a->ptr,a->n,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inneg(ca_modp_vardnsrecdegdnspol_t a,ca_modp_vardnsrecdegdnspol_args args)
{
    a->n = ca_modp_vardnsrecdegdnspolseq_inneg(a->ptr,a->n,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_oodbl(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_modp_vardnsrecdegdnspolseq_oodbl(c->ptr,a->ptr,a->n,v,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_indbl(ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspolseq_indbl(a->ptr,a->n,v,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_ooadd(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_t b,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(a!=b);
    ca_modp_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n = ca_modp_vardnsrecdegdnspolseq_ooadd(c->ptr,a->ptr,a->n,v,b->ptr,b->n,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inadd(ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_t b,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    assert(a!=b);
    while (a->c < b->n) ca_modp_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_modp_vardnsrecdegdnspolseq_inadd(a->ptr,a->n,v,b->ptr,b->n,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_oomulsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,a->n,args);
    c->n = ca_modp_vardnsrecdegdnspolseq_oomulsc(c->ptr,a->ptr,a->n,s,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inmulsc(ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    a->n = ca_modp_vardnsrecdegdnspolseq_inmulsc(a->ptr,a->n,s,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_ooaddmulsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_t b,ca_modp_t t,int w, ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n=ca_modp_vardnsrecdegdnspolseq_ooaddmulsc(c->ptr,a->ptr,a->n,v,b->ptr,b->n,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inaddmulsc(ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_t b,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n) ca_modp_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_modp_vardnsrecdegdnspolseq_inaddmulsc(a->ptr,a->n,v,b->ptr,b->n,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_oomulscadd(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_t b,int w, ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n = ca_modp_vardnsrecdegdnspolseq_oomulscadd(c->ptr,a->ptr,a->n,s,v,b->ptr,b->n,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inmulscadd(ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_t b,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n) ca_modp_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_modp_vardnsrecdegdnspolseq_inmulscadd(a->ptr,a->n,s,v,b->ptr,b->n,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_oomulscaddmulsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_t b,ca_modp_t t,int w, ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,MAX(a->n,b->n),args);
    c->n = ca_modp_vardnsrecdegdnspolseq_oomulscaddmulsc(c->ptr,a->ptr,a->n,s,v,b->ptr,b->n,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inmulscaddmulsc(ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_t b,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n) ca_modp_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_modp_vardnsrecdegdnspolseq_inmulscaddmulsc(a->ptr,a->n,s,v,b->ptr,b->n,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static BOOL ca_modp_vardnsrecdegdnspol_oodivsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    BOOL failed;
    ca_modp_vardnsrecdegdnspol_init(c,a->n,args);
    failed = ca_modp_vardnsrecdegdnspolseq_oodivsc(c->ptr,a->ptr,a->n,s,args);
    if (failed) {
	/* the actual scalars are already freed */
	ca_modp_vardnsrecdegdnspol_destroy(c,args);
    } else {
	c->n = a->n;
	assert(ca_modp_vardnsrecdegdnspol_check(c,args));
    }
    return failed;
}

static void ca_modp_vardnsrecdegdnspol_indivsc(ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,ca_modp_vardnsrecdegdnspol_args args)
{
    a->n = ca_modp_vardnsrecdegdnspolseq_indivsc(a->ptr,a->n,s,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_oomulmonomsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,b->n + m,args);
    c->n = ca_modp_vardnsrecdegdnspolseq_oomulmonomsc(c->ptr,b->ptr,b->n,m,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inmulmonomsc(ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_c tmp;ca_modp_vardnsrecdegdnspol_oomulmonomsc(&tmp,b,m,t,w,args);
    ca_modp_vardnsrecdegdnspol_clear(b,args);ca_modp_vardnsrecdegdnspol_move(b,&tmp,args);
}

static void ca_modp_vardnsrecdegdnspol_ooaddmulmonomsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,int w, ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,MAX(a->n,b->n + m),args);
    c->n = ca_modp_vardnsrecdegdnspolseq_ooaddmulmonomsc(c->ptr,a->ptr,a->n,v,b->ptr,b->n,m,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inaddmulmonomsc(ca_modp_vardnsrecdegdnspol_t a,int v,ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    while (a->c < b->n + m) ca_modp_vardnsrecdegdnspol_expand(a,args);
    a->n = ca_modp_vardnsrecdegdnspolseq_inaddmulmonomsc(a->ptr,a->n,v,b->ptr,b->n,m,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static void ca_modp_vardnsrecdegdnspol_oomulscaddmulmonomsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,int w, ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_init(c,MAX(a->n,b->n + m),args);
    c->n = ca_modp_vardnsrecdegdnspolseq_oomulscaddmulmonomsc(c->ptr,a->ptr,a->n,s,v,b->ptr,b->n,m,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
}

static void ca_modp_vardnsrecdegdnspol_inmulscaddmulmonomsc(ca_modp_vardnsrecdegdnspol_t a,ca_modp_t s,int v,ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,int w,ca_modp_vardnsrecdegdnspol_args args)
{
    a->n = ca_modp_vardnsrecdegdnspolseq_inmulscaddmulmonomsc(a->ptr,a->n,s,v,b->ptr,b->n,m,t,w,args);
    assert(ca_modp_vardnsrecdegdnspol_check(a,args));
}

static BOOL ca_modp_vardnsrecdegdnspol_oodivmonomsc(ca_modp_vardnsrecdegdnspol_t c,ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t, ca_modp_vardnsrecdegdnspol_args args)
{
    if (ca_modp_vardnsrecdegdnspol_ord(b,args) < m) {
	return FAILURE;
    } else {
	BOOL failed;
	ca_modp_vardnsrecdegdnspol_init(c,b->n - m,args);
	failed = ca_modp_vardnsrecdegdnspolseq_oodivmonomsc(c->ptr,b->ptr,b->n,m,t,args);
	if (failed) {
	    /* the actual scalars are already freed */
	    ca_modp_vardnsrecdegdnspol_destroy(c,args);
	} else {
	    c->n = b->n - m;
	    assert(ca_modp_vardnsrecdegdnspol_check(c,args));
	}
	return failed;
    }
}

static void ca_modp_vardnsrecdegdnspol_indivmonomsc(ca_modp_vardnsrecdegdnspol_t b,ca_unimonom_t m,ca_modp_t t,ca_modp_vardnsrecdegdnspol_args args)
{
    ca_modp_vardnsrecdegdnspol_c tmp;
    BOOL failed = ca_modp_vardnsrecdegdnspol_oodivmonomsc(&tmp,b,m,t,args);
    if (failed) { fprintf(stderr,"Exact Division Failed");abort(); }
    ca_modp_vardnsrecdegdnspol_clear(b,args);ca_modp_vardnsrecdegdnspol_move(b,&tmp,args);
}

static void ca_modp_vardnsrecdegdnspol_args_fileOut(id aFiler,ca_modp_vardnsrecdegdnspol_args *args)
{
    ca_modpseq_args_fileOut(aFiler,&args->sargs);
}
static void ca_modp_vardnsrecdegdnspol_args_fileIn(id aFiler,ca_modp_vardnsrecdegdnspol_args *args)
{
    ca_modpseq_args_fileIn(aFiler,&args->sargs);
}

static void ca_modp_vardnsrecdegdnspol_fileOut(id aFiler,ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    int n = self->n;
    [aFiler fileOut:&n type:'i'];
    assert(ca_modp_vardnsrecdegdnspol_check(self,args));
    ca_modpseq_fileOut(aFiler,self->ptr,n,args.sargs);
}

static void ca_modp_vardnsrecdegdnspol_fileIn(id aFiler,ca_modp_vardnsrecdegdnspol_t self,ca_modp_vardnsrecdegdnspol_args args)
{
    int n;
    [aFiler fileIn:&n type:'i'];
    ca_modp_vardnsrecdegdnspol_init(self,n,args);
    ca_modpseq_fileIn(aFiler,self->ptr,n,args.sargs);self->n = n;
    assert(self->n == n && ca_modp_vardnsrecdegdnspol_check(self,args));
}

@implementation ca_modp_vardnsrecdegdns_polynomial : ca_vardnsrecdegdns_polynomial

- _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    id aSymbol;
    
    assert([aCoef isZero] && [aScalar isZero]);
    assert([aCoef class_vardnsrecdegdns_polynomial] == self);
     
    aScalar  = [aScalar cheapCopy];
    aSymbol  = [[aCltn lastElement] cheapCopy];
    termZero = [Term scalarZero:aScalar coefficient:aCoef symbol:aSymbol exponent:1];
    aScalar  = [aScalar cheapCopy];aCltn = [aCltn cheapCopy];
    monomialZero = [Monomial scalar:aScalar symbols:aCltn];
    scalarZero   = aScalar;
    symbols      = aCltn;
    
    ca_modp_vardnsrecdegdnspol_init(&value,2 /* capacity */,ca_modp_vardnsrecdegdnspol_getargs(self));
    assert([self check]);return self;
}

+ scalarZero:aScalar coefficientZero:aCoef symbols:aCltn
{
    return [[super new] _setUpScalarZero:aScalar coefficientZero:aCoef symbols:aCltn];
}

- empty
{
    ca_modp_vardnsrecdegdnspol_c c;assert([self check]);
    ca_modp_vardnsrecdegdnspol_init(&c,2 /* capacity */ ,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- (int) numTerms
{
    /* doesn't take getargs in degsps case */
    assert([self check]);return ca_modp_vardnsrecdegdnspol_count(&value ,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (ca_modp_vardnsrecdegdnspol_t)ca_modp_vardnsrecdegdnspol_value
{
    return &value;
}

- ca_modp_vardnsrecdegdnspol_value:(ca_modp_vardnsrecdegdnspol_t)aValue
{
    self = [self clone];
    ca_modp_vardnsrecdegdnspol_copy(&value,aValue,ca_modp_vardnsrecdegdnspol_getargs(self));
    assert([self check]);return self;
}

- (ca_modp_vardnsrecdegdnspol_t)ca_modp_vardnsrecdegdnspol_reference
{
    return &value;
}

- ca_modp_vardnsrecdegdnspol_reference:(ca_modp_vardnsrecdegdnspol_c *)aReference
{
    self = [self clone];
    ca_modp_vardnsrecdegdnspol_move(&value,aReference,ca_modp_vardnsrecdegdnspol_getargs(self));
    assert([self check]);return self;
}

- copy
{
    ca_modp_vardnsrecdegdnspol_c c;assert([self check]);
    ca_modp_vardnsrecdegdnspol_copy(&c,&value,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- deepCopy
{
    ca_modp_vardnsrecdegdnspol_c c;assert([self check]);
    ca_modp_vardnsrecdegdnspol_deepcopy(&c,&value,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- (BOOL) isEmpty
{
    return ca_modp_vardnsrecdegdnspol_isempty(&value);
}

- (BOOL) isOne
{
    return ca_modp_vardnsrecdegdnspol_isone(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) isMinusOne
{
    return ca_modp_vardnsrecdegdnspol_isminusone(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (unsigned) hash
{
    return ca_modp_vardnsrecdegdnspol_hash(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) isEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return YES;
    } else {
    return ca_modp_vardnsrecdegdnspol_iseq(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_id2t(b),ca_modp_vardnsrecdegdnspol_getargs(self));
    }
}

- (BOOL) notEqual:b
{
    [self checkSameClass:b];
    if (self==b) {
    return NO;
    } else {
    return !ca_modp_vardnsrecdegdnspol_iseq(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_id2t(b),ca_modp_vardnsrecdegdnspol_getargs(self));
    }
}

- (BOOL) isZero
{
    assert([self check]);
    return ca_modp_vardnsrecdegdnspol_iszero(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) notZero
{
    assert([self check]);
    return !ca_modp_vardnsrecdegdnspol_iszero(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) isOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return ca_modp_vardnsrecdegdnspol_isop(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_id2t(b),ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (BOOL) notOpposite:b
{
    assert([self checkSameClass:b] && self != b);
    return !ca_modp_vardnsrecdegdnspol_isop(ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_id2t(b),ca_modp_vardnsrecdegdnspol_getargs(self));
}

- negate
{
    ca_modp_vardnsrecdegdnspol_c c;assert([self check]);
    ca_modp_vardnsrecdegdnspol_ooneg(&c,ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- negateSelf
{
    if (refCount>1) return idReplace(self,[self negate]);
    assert([self check]);
    ca_modp_vardnsrecdegdnspol_inneg(ca_modp_vardnsrecdegdnspol_id2c(self),ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _double:(int)v
{
    ca_modp_vardnsrecdegdnspol_c c;assert([self check]);
    ca_modp_vardnsrecdegdnspol_oodbl(&c,ca_modp_vardnsrecdegdnspol_id2t(self),v,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _doubleSelf:(int)v
{
    if (refCount>1) return idReplace(self,[self _double:v]);
    assert([self check]);
    ca_modp_vardnsrecdegdnspol_indbl(ca_modp_vardnsrecdegdnspol_id2c(self),v,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- double
{
    ca_modp_vardnsrecdegdnspol_c c;assert([self check]);
    ca_modp_vardnsrecdegdnspol_oodbl(&c,ca_modp_vardnsrecdegdnspol_id2t(self),+1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- doubleSelf
{
    if (refCount>1) return idReplace(self,[self double]);
    assert([self check]);
    ca_modp_vardnsrecdegdnspol_indbl(ca_modp_vardnsrecdegdnspol_id2c(self),+1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:b:(int)w
{
    ca_modp_vardnsrecdegdnspol_c c;
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    ca_modp_vardnsrecdegdnspol_ooadd(&c,ca_modp_vardnsrecdegdnspol_id2t(self),v,ca_modp_vardnsrecdegdnspol_id2t(b),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _addSelf:(int)v:b:(int)w
{
    assert([self sameClass:b] && v*v == 1 && w*w == 1 && self != b);
    if (refCount>1) return idReplace(self,[self _add:v:b:w]);
    ca_modp_vardnsrecdegdnspol_inadd(ca_modp_vardnsrecdegdnspol_id2c(self),v,ca_modp_vardnsrecdegdnspol_id2t(b),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- add:b
{
    ca_modp_vardnsrecdegdnspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modp_vardnsrecdegdnspol_ooadd(&c,ca_modp_vardnsrecdegdnspol_id2t(self),+1,ca_modp_vardnsrecdegdnspol_id2t(b),+1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- addSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self add:b]);
    ca_modp_vardnsrecdegdnspol_inadd(ca_modp_vardnsrecdegdnspol_id2c(self),+1,ca_modp_vardnsrecdegdnspol_id2t(b),+1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- subtract:b
{
    ca_modp_vardnsrecdegdnspol_c c;
    assert([self checkSameClass:b] && self != b);
    ca_modp_vardnsrecdegdnspol_ooadd(&c,ca_modp_vardnsrecdegdnspol_id2t(self),+1,ca_modp_vardnsrecdegdnspol_id2t(b),-1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- subtractSelf:b
{
    assert([self sameClass:b] && self != b);
    if (refCount>1) return idReplace(self,[self subtract:b]);
    ca_modp_vardnsrecdegdnspol_inadd(ca_modp_vardnsrecdegdnspol_id2c(self),+1,ca_modp_vardnsrecdegdnspol_id2t(b),-1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- (int) leadingDegree
{
    return ca_modp_vardnsrecdegdnspol_deg(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (int) lastDegree
{
    return ca_modp_vardnsrecdegdnspol_ord(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (int) maxDegree
{
    return ca_modp_vardnsrecdegdnspol_maxdeg(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- (int) minDegree
{
    return ca_modp_vardnsrecdegdnspol_mindeg(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
}

- asScalar
{
    if ([self size] == 1) {
	id asScalar,lastTerm,aSequence = [self eachElement];
	asScalar = (lastTerm = [aSequence lastElement])?[lastTerm asScalar]:nil;
	[aSequence free];return asScalar;
    } else {
	return nil;
    }
}

- asSymbol
{
    if ([self size] == 1) {
	id asSymbol,lastTerm,aSequence = [self eachElement];
	asSymbol = (lastTerm = [aSequence lastElement])?[lastTerm asSymbol]:nil;
	[aSequence free];return asSymbol;
    } else {
	return nil;
    }
}

- check
{
    [super check];
    [scalarZero check];[termZero check];[monomialZero check];
    [symbols check];assert([symbols size] >= 1);
    assert([scalarZero isEqual:[monomialZero scalar]]);
    assert([symbols isEqual:[monomialZero symbols]]);
    assert([termZero isZero]&&[termZero exponent]==1&&[termZero symbol]!=nil);
    assert([[symbols lastElement] isEqual:[termZero symbol]]);

    ca_modp_vardnsrecdegdnspol_check(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
    return self;
}

- clear
{
    ca_modp_vardnsrecdegdnspol_clear(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
    /* free args after clearing polynomial value */
    scalarZero = [scalarZero free];
    termZero = [termZero free];monomialZero = [monomialZero free];
    symbols = [symbols free];return [super clear];
}

- clone
{
    self = [super clone];
    scalarZero   = [scalarZero cheapCopy];
    termZero     = [termZero cheapCopy];
    monomialZero = [monomialZero cheapCopy];
    symbols      = [symbols cheapCopy];
    return self;
}

- scalarZero   { return scalarZero;   }
- termZero     { return termZero;     }
- monomialZero { return monomialZero; }
- symbols      { return symbols;      }

- eachTerm
{
    id aSequence = [ca_modp_vardnsrecdegdnspol_sequence content:[self cheapCopy]];
    return [CASequence over:aSequence];
}

- elt_ca_modp_vardnsrecdegdnspol_ix:(ca_modp_vardnsrecdegdnspol_ixt)ix
{
    if (ca_modp_vardnsrecdegdnspol_ix_ismatch(ix)) {
	id c,s;int e;ca_modp_t t;
	t = ca_modp_vardnsrecdegdnspol_scalat(&value,ix,ca_modp_vardnsrecdegdnspol_getargs(self));
	e = ca_modp_vardnsrecdegdnspol_expat(&value,ix,ca_modp_vardnsrecdegdnspol_getargs(self));
	c = ca_modp_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	return [termZero coefficient:c symbol:s exponent:e];
    } else {
	return nil;
    }
}
- removeTerm
{
    if ([self isEmpty]) {
	return nil;
    } else {
	id lt;
	id c,s;int e;ca_modp_t t;
	[self checkRefCountOne];assert([self check]);
	
	t = ca_modp_vardnsrecdegdnspol_scalfirst(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
	e = ca_modp_vardnsrecdegdnspol_expfirst(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
	c = ca_modp_t2id([termZero coefficient],t);
	s = (e)?[[termZero symbol] cheapCopy]:nil;
	lt = [termZero coefficient:c symbol:s exponent:e];
	
	ca_modp_vardnsrecdegdnspol_remove(&value,ca_modp_vardnsrecdegdnspol_getargs(self));
	[self invalidate];assert([self check]);return lt;
    }
}

- insertTerm:aTerm
{
    if ([aTerm isZero])  {
	[aTerm free];return self;
    } else {
	ca_modp_c c;ca_modp_t r;int e;
	assert([aTerm check]);[self checkRefCountOne];assert([self check]);
	r = ca_modp_id2ref([aTerm coefficient],&c);e = [aTerm exponent];
	ca_modp_vardnsrecdegdnspol_insert(&value,e,r,ca_modp_vardnsrecdegdnspol_getargs(self));
	assert([aTerm check]);[aTerm free];
	[self invalidate];assert([self check]);return self;
    }
}

- multiplyCoefficient:s
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c sc;
    ca_modp_vardnsrecdegdnspol_oomulsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- multiplySelfCoefficient:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self multiplyCoefficient:s]);
    ca_modp_vardnsrecdegdnspol_inmulsc(ca_modp_vardnsrecdegdnspol_id2c(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- divideCoefficient:s
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c sc;
    BOOL failed = ca_modp_vardnsrecdegdnspol_oodivsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegdnspol_getargs(self));
    return (failed)?nil:ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- divideSelfCoefficient:s
{
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self divideCoefficient:s]);
    ca_modp_vardnsrecdegdnspol_indivsc(ca_modp_vardnsrecdegdnspol_id2c(self),ca_modp_id2ref(s,&sc),ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c bc;
    ca_modp_vardnsrecdegdnspol_ooaddmulsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),v,ca_modp_vardnsrecdegdnspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyCoefficient:b:(int)w
{
    ca_modp_c bc;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyCoefficient:b:w]);
    ca_modp_vardnsrecdegdnspol_inaddmulsc(ca_modp_vardnsrecdegdnspol_id2c(self),v,ca_modp_vardnsrecdegdnspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B:(int)w
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c ac;
    ca_modp_vardnsrecdegdnspol_oomulscadd(&c,ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegdnspol_id2t(B),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B:(int)w
{
    ca_modp_c ac;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B:w]);
    ca_modp_vardnsrecdegdnspol_inmulscadd(ca_modp_vardnsrecdegdnspol_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegdnspol_id2t(B),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c ac,bc;
    ca_modp_vardnsrecdegdnspol_oomulscaddmulsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegdnspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyCoefficient:b:(int)w
{
    ca_modp_c ac,bc;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyCoefficient:b:w]);
    ca_modp_vardnsrecdegdnspol_inmulscaddmulsc(ca_modp_vardnsrecdegdnspol_id2c(self),ca_modp_id2ref(a,&ac),v,ca_modp_vardnsrecdegdnspol_id2t(B),ca_modp_id2ref(b,&bc),w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}
- multiplyTerm:aTerm
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegdnspol_oomulmonomsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),[aTerm exponent],s,+1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- multiplySelfTerm:aTerm
{
    ca_modp_t s;
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self multiplyTerm:aTerm]);
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegdnspol_inmulmonomsc(ca_modp_vardnsrecdegdnspol_id2c(self),[aTerm exponent],s,+1,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyTerm:aTerm:(int)v
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_t s;
    ca_modp_c sc;
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegdnspol_oomulmonomsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),[aTerm exponent],s,v,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfTerm:aTerm:(int)v
{
    ca_modp_t s;
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self _multiplyTerm:aTerm:v]);
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegdnspol_inmulmonomsc(ca_modp_vardnsrecdegdnspol_id2c(self),[aTerm exponent],s,v,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- divideTerm:aTerm
{
    int e;
    ca_modp_vardnsrecdegdnspol_c c;
    BOOL failed;
    ca_modp_t s;
    ca_modp_c sc;
    
    e = [aTerm exponent];
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    failed=ca_modp_vardnsrecdegdnspol_oodivmonomsc(&c,ca_modp_vardnsrecdegdnspol_id2t(self),e,s,ca_modp_vardnsrecdegdnspol_getargs(self));
    return (failed)?nil:ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- divideSelfTerm:aTerm
{
    ca_modp_t s;
    ca_modp_c sc;
    if (refCount>1) return idReplace(self,[self divideTerm:aTerm]);
    s = ca_modp_id2ref([aTerm coefficient],&sc);
    ca_modp_vardnsrecdegdnspol_indivmonomsc(ca_modp_vardnsrecdegdnspol_id2c(self),[aTerm exponent],s,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _add:(int)v:B multiplyTerm:b:(int)w
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c bc;
    ca_modp_t bs;
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegdnspol_ooaddmulmonomsc(&c,ca_modp_vardnsrecdegdnspol_id2c(self),v,ca_modp_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _addSelf:(int)v:B multiplyTerm:b:(int)w
{
    ca_modp_c bc;
    ca_modp_t bs;
    if (refCount>1) return idReplace(self,[self _add:v:B multiplyTerm:b:w]);
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegdnspol_inaddmulmonomsc(ca_modp_vardnsrecdegdnspol_id2c(self),v,ca_modp_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- _multiplyCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_modp_vardnsrecdegdnspol_c c;
    ca_modp_c ac,bc;
    ca_modp_t as,bs;
    as = ca_modp_id2ref(a,&ac);
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegdnspol_oomulscaddmulmonomsc(&c,ca_modp_vardnsrecdegdnspol_id2c(self),as,v,ca_modp_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return ca_modp_vardnsrecdegdnspol_c2id(self,&c);
}

- _multiplySelfCoefficient:a:(int)v add:B multiplyTerm:b:(int)w
{
    ca_modp_c ac,bc;
    ca_modp_t as,bs;
    if (refCount>1) return idReplace(self,[self _multiplyCoefficient:a:v add:B multiplyTerm:b:w]);
    as = ca_modp_id2ref(a,&ac);
    bs = ca_modp_id2ref([b coefficient],&bc);
    ca_modp_vardnsrecdegdnspol_inmulscaddmulmonomsc(ca_modp_vardnsrecdegdnspol_id2c(self),as,v,ca_modp_vardnsrecdegdnspol_id2t(B),[b exponent],bs,w,ca_modp_vardnsrecdegdnspol_getargs(self));
    return [self invalidate];
}

- fileOutOn:aFiler
{
    ca_modp_vardnsrecdegdnspol_args args;
    [super fileOutOn:aFiler];
    args = ca_modp_vardnsrecdegdnspol_getargs(self); /* can't do this in fileIn case */
    ca_modp_vardnsrecdegdnspol_args_fileOut(aFiler,&args);
    ca_modp_vardnsrecdegdnspol_fileOut(aFiler,&value,args);
    return self;
}

- fileInFrom:aFiler
{
    ca_modp_vardnsrecdegdnspol_args args;
    [super fileInFrom:aFiler];
    ca_modp_vardnsrecdegdnspol_args_fileIn(aFiler,&args);
    ca_modp_vardnsrecdegdnspol_fileIn(aFiler,&value,args);
    return self;
}

@end

