
/*
 * Portable Object Compiler (c) 1997,98.  All Rights Reserved.
 */

#include "Block.h"
#include <setjmp.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

@implementation Block

/*****************************************************************************
 *
 * Creating Blocks
 *
 ****************************************************************************/

/*
 * Compiler emits calls to newBlock() for creation of Blocks.
 */

id OBJCRTEXPORT newBlock(int n,IMP f,void*d) 
{ 
    return [Block new:n:f:d]; 
}

/*
 * Might be nice to have a panel here, on the Macintosh
 * (otherwise the app just disappears and you have to look for 
 * a file "stderr" which contains (somewhere) the error message
 *
 */

static id err_fun(id thisBlock,void *data,id msg,id rcv) 
{
    if (rcv) fprintf(stderr,"%s: ",[rcv name]);
    if (msg) [msg printOn:stderr]; else fprintf(stderr,"(null message)");
    if (rcv!=nil || msg!=nil) fprintf(stderr,"\n");
    abort(); return nil;
}

- setUp:(int)n:(IMP)f:(void*)d
{
    nVars = n;fn = f;data = d;return self;
}

+ new
{
    return [self shouldNotImplement];
}

+ new:(int)n:(IMP)f:(void*)d
{
	return [[super new] setUp:n:f:d];
}

/*****************************************************************************
 *
 * Exception Handling
 *
 ****************************************************************************/

#define MAX_JMP 32
static id  defaultHandler;
static int jmpCount = 0;
static id  jmpHandlers[MAX_JMP];

+ pop
{
    return (jmpCount)?jmpHandlers[--jmpCount]:nil;
}

+ push:aHandler
{
    jmpHandlers[jmpCount] = [aHandler errorGoodHandler];
    if ((++jmpCount)==MAX_JMP) return [self error:"Increase MAX_JMP."];
    return self;
}

+ errorHandler
{
    if (!defaultHandler) defaultHandler = newBlock(2,(IMP)err_fun,NULL);
    return defaultHandler;
}

+ errorHandler:aHandler
{
    id ret = defaultHandler;
    defaultHandler = [aHandler errorGoodHandler];
    return ret;
}


- ifError:aHandler
{
    id returnValue;
    [Block push:aHandler]; returnValue = [self value]; [Block pop];
    return returnValue;
}

- value:anObject ifError:aHandler
{
    id returnValue;
    [Block push:aHandler]; returnValue = [self value:anObject]; [Block pop];
    return returnValue;
}


/*****************************************************************************
 *
 * Evaluating Blocks
 *
 ****************************************************************************/

- errorNumArgs
{
    return [self error:"Block has wrong number of arguments."];
}

- errorGoodHandler
{
    return (nVars == 2)?self:[self errorNumArgs];
} 

- value
{
    return (nVars==0)?(id)(*fn)(self,data):[self errorNumArgs];
}

static int atExitCount = 0;
static id atExitBlocks[32];

static void valueAtExit(void)
{
    if (atExitCount) [atExitBlocks[--atExitCount] value];
}

- atExit
{
    if (nVars == 0) {
	atExitBlocks[atExitCount++] = self;
	atexit(valueAtExit);
	return self;
    } else {
	return [self errorNumArgs];
    }
}

- value:anObject
{
    return (nVars==1)?(id)(*fn)(self,data,anObject):[self errorNumArgs];
}

- value:firstObject value:secondObject
{
    if (nVars==2) {
	return (id)(*fn)(self,data,firstObject,secondObject);
    } else {
	return [self errorNumArgs];
    }
    return nil;
}


/*****************************************************************************
 *
 * Control Flow 
 *
 ****************************************************************************/

- repeatTimes:(int)n
{
    int i;
    for(i=0;i<n;i++) [self value];
    return self;
}

@end

/* some defs needed when cross-compiling with a DIFFERENT compiler */
/* placed this here 'cause we will have a different objcrt.h and Object.h */

#if !defined(__PORTABLE_OBJC__) && !defined(OCLONGLONG)

void *OC_Malloc(size_t n) { return malloc(n); }
void *OC_MallocAtomic(size_t n) { return malloc(n); }
void *OC_Realloc(void *p,size_t n) { return realloc(p,n); }
void *OC_Free(void *p) { free(p);return NULL; }

#endif /* __PORTABLE_OBJC__ */

