
/*
 * Portable Object Compiler (c) 1998.  All Rights Reserved.
 */


#include "objpak.h"

#define DEFAULT_CAPACITY (16)

@implementation IntArray

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/

static void intary_init(intary_t self,int v,int n,int c)
{
    int *p = (int *)OC_MallocAtomic(c * sizeof(int));
    assert(0 <= n && n <= c);
    
    self->count    = n;
    self->capacity = c;
    self->ptr      = p;

    while (n--) *p++ = v;
}

+ new
{
    id newObj = [super new];
    intary_init([newObj intaryValue],0,0,DEFAULT_CAPACITY);
    return newObj;
}

+ new:(unsigned)n
{
    id newObj = [super new];
    intary_init([newObj intaryValue],0,n,n);
    return newObj;
}

+ with:(unsigned)nArgs,...
{
    /* use OC macros for porting to SunOS4 */
    unsigned i;
    OC_VA_LIST vp;
    id newObject = [self new:nArgs];

    /* #if 0 this piece of code if problems with stdarg
     * typically this means the driver is not configured with
     * the right -builtintype
     * or builtinfunction flags (because the macros might expand to these)
     * 
     * alternative solution: check the .P output (-retain) and 
     * do a setenv OBJCOPT -builtinfunction __builtin_foo
     * (and please let me know)
     */

#if 1
    OC_VA_START(vp,nArgs);
    for(i=0;i<nArgs;i++) {
		int anObject = OC_VA_ARG(vp,int);
		[newObject intAt:i put:anObject];
	}
    OC_VA_END(vp);
#endif

    return newObject;
}

static void int_ncpy(int *p,int *q,int n)
{
    while (n--) *p++ = *q++;
}

static void intary_copy(intary_t dst,intary_t src)
{
    int n = src->count;
    
    dst->count    = n;
    dst->capacity = n + 1;
    dst->ptr      = (int *)OC_MallocAtomic(dst->capacity * sizeof(int));

    int_ncpy(dst->ptr,src->ptr,n);
}

- copy
{
    id aCopy = [super copy];
    intary_copy([aCopy intaryValue],[self intaryValue]);
    return aCopy;
}

- deepCopy
{
    return [self copy];
}

static void intary_clear(intary_t self)
{
    OC_Free(self->ptr);self->ptr = NULL;
}

- free
{
    intary_clear([self intaryValue]);return [super free];
}

/*****************************************************************************
 *
 * Interrogation
 *
 ****************************************************************************/

- (intary_t) intaryValue
{
    return &value;
}

- (unsigned) size
{
    return [self intaryValue]->count;
}

static int intary_intat(intary_t self,int i)
{
    if (0 <= i && i < self->count) {
	return (self->ptr)[i];
    } else {
	[IntArray error:"index out of bounds"];
	return 0;
    }
}

- (int) intAt:(unsigned)anOffset
{
    return intary_intat([self intaryValue],anOffset);
}

static int intputint(int *self,int c)
{
    int r = *self;*self = c;return r;
}

static int intary_putintat(intary_t self,int i,int c)
{
    if (0 <= i && i < self->count) {
	return intputint(self->ptr + i,c);
    } else {
	[IntArray error:"index out of bounds"];
	return 0;
    }
}

- (int) intAt:(unsigned)anOffset put:(int)anInt
{
    return intary_putintat([self intaryValue],anOffset,anInt);
}

/*****************************************************************************
 *
 * Printing
 *
 ****************************************************************************/

- printOn:(IOD)aFile
{
    unsigned i,n = [self size];
    if (n) fprintf(aFile,"%i",[self intAt:0]);
    for(i=1;i<n;i++) fprintf(aFile,",%i",[self intAt:i]);
    return self;
}

/*****************************************************************************
 *
 * Archiving
 *
 ****************************************************************************/

#ifdef __PORTABLE_OBJC__
static void int_fileout(id aFiler,int *p,int n)
{
    while (n--) [aFiler fileOut:p++ type:'i'];
}

static void int_filein(id aFiler,int *p,int n)
{
    while (n--) [aFiler fileIn:p++ type:'i'];
}

static void intary_fileout(id aFiler,intary_t self)
{
    [aFiler fileOut:&self->count type:'i'];
    [aFiler fileOut:&self->capacity type:'i'];
    int_fileout(aFiler,self->ptr,self->count);
}

static void intary_filein(id aFiler,intary_t self)
{
    [aFiler fileIn:&self->count type:'i'];
    [aFiler fileIn:&self->capacity type:'i'];
    self->ptr = (int *)OC_MallocAtomic(self->capacity * sizeof(int));
    int_filein(aFiler,self->ptr,self->count);
}

- fileOutOn:aFiler
{
    [super fileOutOn:aFiler];intary_fileout(aFiler,&value);return self;
}

- fileInFrom:aFiler
{
    [super fileInFrom:aFiler];intary_filein(aFiler,&value);return self;
}

#endif /* __PORTABLE_OBJC__ */

@end

