
/*
 * Portable Object Compiler (c) 1997,98.  All Rights Reserved.
 */

#include "objpak.h"

#define DEFAULT_CAPACITY (16)

@implementation OrdCltn

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/

static void objcol_init(objcol_t self,int n,int c)
{
    assert(0 <= n && n <= c);
    self->count    = n;
    self->capacity = c;
    self->ptr      = (id *)OC_Malloc(c * sizeof(id));
}

+ new
{
    return [self new:DEFAULT_CAPACITY];
}

+ new:(unsigned)n
{
    id newObject = [super new];
    objcol_init([newObject objcolValue],0,n);
    return newObject;
}

/*****************************************************************************
 *
 * Creating with contents.
 *
 ****************************************************************************/

+ with:(int)nArgs,...
{
    id newObject;

    /* use OC macros for porting to SunOS4 */
    OC_VA_LIST vp;

    newObject = [self new];

    /* #if 0 this piece of code if problems with stdarg
     * typically this means the driver is not configured with
     * the right -builtintype
     * or builtinfunction flags (because the macros might expand to these)
     * 
     * alternative solution: check the .P output (-retain) and 
     * do a setenv OBJCOPT -builtinfunction __builtin_foo
     * (and please let me know)
     */

#if 1
    OC_VA_START(vp,nArgs);
    while (nArgs-- > 0) {
		id anObject = OC_VA_ARG(vp,id);
		[newObject add:anObject];
	}
    OC_VA_END(vp);
#endif

    return newObject;
}

+ with:firstObject with:nextObject
{
    return [[[self new] add:firstObject] add:nextObject];
}

+ add:firstObject
{
    return [[self new] add:firstObject];
}

static void objptr_copy(id *p,id *q,int c)
{
    while (c--) *p++ = *q++;
}

static void objcol_copy(objcol_t dst,objcol_t src)
{
    objcol_init(dst,src->count,src->count);
    objptr_copy(dst->ptr,src->ptr,src->count);
}

- copy
{
    id aCopy = [super copy];
    objcol_copy([aCopy objcolValue],(&value));
    return aCopy;
}

static void objptr_deepcopy(id *p,id *q,int c)
{
    while (c--) { id obj = *q++;*p++ = (obj)?[obj deepCopy]:nil; }
}

static void objcol_deepcopy(objcol_t dst,objcol_t src)
{
    objcol_init(dst,src->count,src->count);
    objptr_deepcopy(dst->ptr,src->ptr,src->count);
}

- deepCopy
{
    /* not all Object implementations have |deepCopy| */
    /* so we send |copy| to super instead             */

    id aCopy = [super copy];
    objcol_deepcopy([aCopy objcolValue],(&value));
    return aCopy;
}

static void objcol_empty(objcol_t self)
{
    self->count = 0;
}

- emptyYourself
{
    objcol_empty((&value));return self;
}

static void objptr_clear(id *p,int c)
{
    while (c--) { id obj = *p;*p++ = (obj)?[obj free]:nil; }
}

static void objcol_freecontents(objcol_t self)
{
    objptr_clear(self->ptr,self->count);self->count = 0;
}

- freeContents
{
    objcol_freecontents((&value));return self;
}

static void objcol_clear(objcol_t self)
{
    self->count = 0;self->capacity = 0;OC_Free(self->ptr);self->ptr = NULL;
}

- free
{
    objcol_clear((&value));return [super free];
}

/*****************************************************************************
 *
 * Interrogation
 *
 ****************************************************************************/

- (objcol_t) objcolValue
{
    return &value;
}

- (unsigned) size
{
    return (unsigned)((&value)->count);
}

- (BOOL) isEmpty
{
    return (&value)->count == 0;
}

- (unsigned) lastOffset
{
    return (&value)->count - 1;
}

- eachElement
{
    id aCarrier = [CollectionSequence over:self];
	return [Sequence over:aCarrier];
}

static id objptr_first(id *p,int n)
{
    return (n)?p[0]:nil;
}

static id objcol_first(objcol_t self)
{
    return objptr_first(self->ptr,self->count);
}

- firstElement
{
    return objcol_first((&value));
}

static id objptr_last(id *p,int n)
{
    return (n)?p[n-1]:nil;
}

static id objcol_last(objcol_t self)
{
    return objptr_last(self->ptr,self->count);
}

- lastElement
{
    return objcol_last((&value));
}

/*****************************************************************************
 *
 * Comparing
 *
 ****************************************************************************/

- (unsigned) hash
{
    return [super hash]; /* would it pay to improve this? */
}

static BOOL objptr_eq(id *p,id *q,int n)
{
    while (n--) if (![*p++ isEqual:*q++]) return NO;
    return YES;
}

static BOOL objcol_eq(objcol_t a,objcol_t b)
{
    if (a->count == b->count) {
	return objptr_eq(a->ptr,b->ptr,a->count);
    } else {
	return NO;
    }
}

- (BOOL) isEqual:aCltn
{
    return (self==aCltn)?YES:objcol_eq((&value),[aCltn objcolValue]);
}

/*****************************************************************************
 *
 * Adding
 *
 ****************************************************************************/

static BOOL objcol_needsexpand(objcol_t self)
{
    assert(self->count <= self->capacity);return self->count == self->capacity;
}

static void objcol_expand(objcol_t self)
{
    self->capacity = 1 + 2*self->capacity;
    self->ptr      = (id *)OC_Realloc(self->ptr,sizeof(id) * self->capacity);
}

static int objptr_addlast(id *p,id obj,int n)
{
    p[n] = obj;return n+1;
}

static void objcol_addlast(objcol_t self,id obj)
{
    if (objcol_needsexpand(self)) objcol_expand(self);
    self->count = objptr_addlast(self->ptr,obj,self->count);
}

- add:anObject
{
    if (anObject) {
	objcol_addlast((&value),anObject);
	return self;
    } else {
	return self;
    }
}

- addYourself
{
    int i,n = [self size];
    for(i=0;i<n;i++) [self add:[self at:i]];
    return self;
}

static int objptr_addfirst(id *p,id obj,int n)
{
    int m = n;
    p += n;while (m--) { id *q = p - 1;*p = *q;p = q; }
    *p = obj;return n + 1;
}

static void objcol_addfirst(objcol_t self,id obj)
{
    if (objcol_needsexpand(self)) objcol_expand(self);
    self->count = objptr_addfirst(self->ptr,obj,self->count);
}

- addFirst:newObject
{
    if (newObject) {
	objcol_addfirst((&value),newObject);
	return self;
    } else {
	return self;
    }
}

- addLast:newObject
{
    return [self add:newObject];
}

- addIfAbsent:anObject
{
    if ([self find:anObject] == nil) [self add:anObject];
    return self;
}

- addIfAbsentMatching:anObject
{
    if ([self findMatching:anObject] == nil) [self add:anObject];
    return self;
}

/*****************************************************************************
 *
 * Insertion
 *
 ****************************************************************************/

static int objptr_insert(id *p,id obj,int i,int n)
{
    if (i==n) {
	return objptr_addlast(p,obj,n);
    } else {
	return i + objptr_addfirst(p + i,obj,n - i);
    }
}

static void objcol_insert(objcol_t self,id obj,int i)
{
    if (objcol_needsexpand(self)) objcol_expand(self);
    self->count = objptr_insert(self->ptr,obj,i,self->count);
}

- boundsError:(unsigned)i in:(SEL)aSelector
{
    return [self error:"Offset out of bounds."];
}

- at:(unsigned )anOffset insert:anObject
{
    if (anObject) {
	if (anOffset > [self size]) [self boundsError:anOffset in:_cmd];
	objcol_insert((&value),anObject,(int)anOffset);
	return self;
   } else {
	return self;
    }
}

- couldntFind:anObject in:(SEL)aSelector
{
    return [self error:"Could not find object."];
}

- insert:newObject after:oldObject
{
    if (newObject) {
	unsigned offset = [self offsetOf:oldObject];
	if (offset == (unsigned)-1) [self couldntFind:oldObject in:_cmd];
	return [self at:offset+1 insert:newObject];
    } else {
	return self;
    }
}

- insert:newObject before:oldObject
{
    if (newObject) {
	unsigned offset = [self offsetOf:oldObject];
	if (offset == (unsigned)-1) [self couldntFind:oldObject in:_cmd];
	return [self at:offset-1 insert:newObject];
    } else {
	return self;
    }
}

/*****************************************************************************
 *
 * Relative Accessing
 *
 ****************************************************************************/

- after:anObject
{
    unsigned offset = [self offsetOf:anObject];
    if (offset == (unsigned)-1) 
	return [self couldntFind:anObject in:_cmd];
    return (offset == [self lastOffset])?nil:[self at:offset+1];
}

- before:anObject
{
    unsigned offset = [self offsetOf:anObject];
    if (offset == (unsigned)-1) return [self couldntFind:anObject in:_cmd];
    return (offset == 0)?nil:[self at:offset-1];
}

static id objcol_at(objcol_t self,int i)
{
    assert(0 <= i && i < self->count);return (self->ptr)[i];
}

- at:(unsigned )anOffset
{
    if (anOffset >= [self size]) [self boundsError:anOffset in:_cmd];
    return objcol_at((&value),anOffset);
}

static id objcol_atput(objcol_t self,int i,id obj)
{
    id tmp;assert(0 <= i && i < self->count);
    tmp = (self->ptr)[i];(self->ptr)[i] = obj;return tmp;
}

- at:(unsigned )anOffset put:anObject
{
    if (anOffset >= [self size]) [self boundsError:anOffset in:_cmd];
    return objcol_atput((&value),anOffset,anObject);
}

/*****************************************************************************
 *
 * Removing
 *
 ****************************************************************************/

static id objptr_removefirst(id *p,int n)
{
    id obj = *p;
    n--;while (n--) { id *q = p + 1;*p = *q;p = q; }
    return obj;
}

static id objcol_removefirst(objcol_t self)
{
    if (self->count) {
	id obj = objptr_removefirst(self->ptr,self->count);
	self->count--;return obj;
    } else {
	return nil;
    }
}

- removeFirst
{
    return objcol_removefirst((&value));
}

static id objptr_removelast(id *p,int n)
{
    return p[n - 1];
}

static id objcol_removelast(objcol_t self)
{
    if (self->count) {
	id obj = objptr_removelast(self->ptr,self->count);
	self->count--;return obj;
    } else {
	return nil;
    }
}

- removeLast
{
    return objcol_removelast((&value));
}

static id objptr_removeat(id *p,int i,int n)
{
    if (i == n - 1) {
	return objptr_removelast(p,n);
    } else {
	return objptr_removefirst(p + i,n - i);
    }
}

static id objcol_removeat(objcol_t self,int i)
{
    if (self->count) {
	id obj = objptr_removeat(self->ptr,i,self->count);
	self->count--;return obj;
    } else {
	return nil;
    }
}

- removeAt:(unsigned )anOffset
{
    if (anOffset >= [self size]) [self boundsError:anOffset in:_cmd];
    return objcol_removeat((&value),anOffset);
}

- removeAtIndex:(unsigned )anOffset
{
    return [self removeAt:anOffset];
}

- remove:oldObject
{
    unsigned offset = [self offsetOf:oldObject];
    if (offset == (unsigned)-1) return nil;
    return [self removeAt:offset];
}

- remove:oldObject ifAbsent:exceptionBlock
{
    id anObject = [ self remove:oldObject ];
    return (anObject)?anObject:[exceptionBlock value];
}

/*****************************************************************************
 *
 * Testing Contents
 *
 ****************************************************************************/

- (BOOL) includesAllOf:aCol
{
    if (self == aCol) {
	return YES;
    } else {
	id e,seq = [aCol eachElement];
	while ((e=[seq next])) if (![self includes:e]) {
	    [seq free];return NO;
	}
	[seq free];return YES;
    }
}

- (BOOL) includesAnyOf:aCol
{
    if (self == aCol) {
	return YES;
    } else {
	id e,seq = [aCol eachElement];
	while ((e=[seq next])) if ([self includes:e]) {
	    [seq free];return YES;
	}
	[seq free];return NO;
    }
}

/*****************************************************************************
 *
 * Adding and Removing Contents
 *
 ****************************************************************************/

- addContentsTo:aCol
{
    if (self == aCol) {
	[self addYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [self eachElement];
	while ((anElement = [elements next])) [aCol add:anElement];
	elements = [elements free];
    }
	
    return aCol;
}

- addContentsOf:aCol
{
    if (self == aCol) {
	[self addYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [aCol eachElement];
	while ((anElement = [elements next])) [self add:anElement];
	elements = [elements free];
    }
	
    return self;
}
- addAll:aCol
{
    return [self addContentsOf:aCol];
}

- removeContentsOf:aCol
{
    if (self == aCol) {
	return [self emptyYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [aCol eachElement];
	while ((anElement = [elements next])) [self remove:anElement];
	elements = [elements free];
	
	return self;
    }
}

- removeContentsFrom:aCol
{
    if (self == aCol) {
	[self emptyYourself];
    } else {
	id elements;
	id anElement;
	
	elements = [self eachElement];
	while ((anElement = [elements next])) [aCol remove:anElement];
	elements = [elements free];
    }
	
    return self;
}

- removeAll:aCol
{
    return [self removeContentsOf:aCol];
}

/*****************************************************************************
 *
 * Converting to different type of collection.
 *
 ****************************************************************************/

- asSet
{
    /* Stepstone isKindOf takes a id but class returns SHR */
    if ([self isKindOf:(id)[Set class]]) {
	return self;
    } else {
	return [[Set new] addContentsOf:self];
    }
}

- asOrdCltn
{
    /* Stepstone isKindOf takes a id but class returns SHR */
    if ([self isKindOf:(id)[OrdCltn class]]) {
	return self;
    } else {
	return [[OrdCltn new] addContentsOf:self];
    }
}

/*****************************************************************************
 *
 * Block methods
 *
 ****************************************************************************/

- detect:aBlock
{
    id elements;
    id anElement;

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if (([aBlock value:anElement])) {
	    elements = [elements free];
	    return anElement; 
	}
    }

    elements = [elements free];
    return nil;
}

- detect:aBlock ifNone:noneBlock
{
    id elements;
    id anElement;

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if (([aBlock value:anElement])) {
	    elements = [elements free];return anElement; 
	}
    }

    elements = [elements free];
    return [noneBlock value];
}

- select:testBlock
{
    id elements;
    id anElement;
    id newObject = [isa new];

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if (([testBlock value:anElement])) [newObject add:anElement];
    }

    elements = [elements free];
    return newObject;
}

- reject:testBlock
{
    id elements;
    id anElement;
    id newObject = [isa new];

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	if (!([testBlock value:anElement])) [newObject add:anElement];
    }

    elements = [elements free];
    return newObject;
}

- collect:transformBlock
{
    id elements;
    id anElement;
    id newObject = [isa new];

    elements = [self eachElement];

    while ((anElement = [elements next])) {
	id anImage = [ transformBlock value:anElement ];
	if (anImage) [ newObject add:anImage ]; 
    }

    elements = [elements free];
    return newObject;
}

- (unsigned) count:aBlock
{
    id elements;
    id anElement;
    unsigned c = 0;

    elements = [self eachElement];
    while ((anElement = [elements next])) {
	if ([aBlock value:anElement]) c++;
    }
    elements = [elements free];

    return c;
}

/*****************************************************************************
 *
 * Making elements perform actions.
 *
 ****************************************************************************/

- elementsPerform:(SEL)aSelector
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
		[anElement perform:aSelector];
	}
    elements = [elements free];
    
    return self;
}

- elementsPerform:(SEL)aSelector with:anObject
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
		[anElement perform:aSelector with:anObject];
	}
    elements = [elements free];
    
    return self;
}

- elementsPerform:(SEL)aSelector with:anObject with:otherObject
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
		[anElement perform:aSelector with:anObject with:otherObject];
	}
    elements = [elements free];
    
    return self;
}

- elementsPerform:(SEL)aSelector with:anObject with:otherObject with:thirdObj
{
    id elements;
    id anElement;
    
    elements = [self eachElement];
    while ((anElement = [elements next])) {
     [anElement perform:aSelector with:anObject with:otherObject with:thirdObj];
    }
    elements = [elements free];
    
    return self;
}
/*****************************************************************************
 *
 * Do Blocks  (also provide a -reverseDo: here)
 *
 ****************************************************************************/

- do:aBlock
{
    int i = 0;
    int n = [self size];
    for(i=0;i<n;i++) { [aBlock value:[self at:i]]; }
    return self;
}

- do:aBlock until:(BOOL*)flag
{
    int i = 0;
    int n = [self size];
    for(i=0;i<n;i++) { [aBlock value:[self at:i]]; if (*flag) break; }
    return self;
}

- reverseDo:aBlock
{
    unsigned n = [self size];
    while (n--) [aBlock value:[self at:n]];
    return self;
}

/*****************************************************************************
 *
 * Locating
 *
 ****************************************************************************/

static id objptr_find(id *p,id obj,int n)
{
    int i;
    for(i=0;i<n;i++) if (*p == obj) return *p;else p++;
    return nil;
}

static id objcol_find(objcol_t self,id obj)
{
    return objptr_find(self->ptr,obj,self->count);
}

- find:anObject
{
    return objcol_find((&value),anObject);
}

static id objptr_findmatch(id *p,id obj,int n)
{
    int i;
    for(i=0;i<n;i++) if (p[i] == obj || [p[i] isEqual:obj]) return p[i];
    return nil;
}

static id objcol_findmatch(objcol_t self,id obj)
{
    return objptr_findmatch(self->ptr,obj,self->count);
}

- findMatching:anObject
{
    return objcol_findmatch((&value),anObject);
}

- (BOOL) includes:anObject
{
    return [self findMatching:anObject] == nil;
}

static id objptr_findstr(id *p,STR s,int n)
{
    int i;
    for(i=0;i<n;i++) if ([*p++ isEqualSTR:s]) return *p;
    return nil;
}

static id objcol_findstr(objcol_t self,STR s)
{
    return objptr_findstr(self->ptr,s,self->count);
}

- findSTR:(STR )strValue
{
    return objcol_findstr((&value),strValue);
}

- (BOOL) contains:anObject
{
    return [self find:anObject] != nil;
}

static int objptr_offset(id *p,id obj,int n)
{
    int i;
    for(i=0;i<n;i++) if (*p++ == obj) return i;
    return -1;
}

static int objcol_offset(objcol_t self,id obj)
{
    return objptr_offset(self->ptr,obj,self->count);
}

- (unsigned) offsetOf:anObject
{
    return (unsigned)objcol_offset((&value),anObject);
}

/*****************************************************************************
 *
 * Printing
 *
 ****************************************************************************/

- printOn:(IOD)aFile
{
    [[[self eachElement] printOn:aFile] free];return self;
}

/*****************************************************************************
 *
 * Archiving
 *
 ****************************************************************************/

#ifdef __PORTABLE_OBJC__
static void objptr_fileout(id aFiler,id *a,int n)
{
    while (n--) [aFiler fileOut:a++ type:'@'];
}

static void objptr_filein(id aFiler,id *a,int n)
{
    while (n--) [aFiler fileIn:a++ type:'@'];
}

static void objcol_fileout(id aFiler,objcol_t self)
{
    [aFiler fileOut:&self->count type:'i'];
    [aFiler fileOut:&self->capacity type:'i'];
    objptr_fileout(aFiler,self->ptr,self->count);
}

static void objcol_filein(id aFiler,objcol_t self)
{
    int n,c;
    [aFiler fileIn:&n type:'i'];
    [aFiler fileIn:&c type:'i'];
    objcol_init(self,n,c);
    objptr_filein(aFiler,self->ptr,n);
}

- fileOutOn:aFiler
{
    [super fileOutOn:aFiler];objcol_fileout(aFiler,&value);return self;
}

- fileInFrom:aFiler
{
    [super fileInFrom:aFiler]; objcol_filein(aFiler,&value);return self;
}

#endif /* __PORTABLE_OBJC__ */

@end

