
/*
 * Portable Object Compiler (c) 1997,98.  All Rights Reserved.
 */

#include "oclib.h"

id currentBlock; /* inside outputBlockDefs compound */

@implementation BlockExpression
- setUpBlock
{
	static int blockCount = 1;
	count = blockCount++;
	return self;
}

+ new
{
	return [[super new] setUpBlock];
}

- free
{
    return nil;
}

- blockFunctionName
{
	return [String sprintf:"blockFun%i",count];
}

- blockDataName
{
    return [String sprintf:"blockVar%i",count];
}

- blockDataTypeName
{
    return [String sprintf:"struct blockData%i",count];
}

- scope
{
    return scope;
}

- setArguments:aNode
{
	arguments = aNode;
	
	if (aNode) { /* :x :y :z */
	    int i;
	    aNode = [aNode subnodes];
	    argumentNames = [OrdCltn new];
	    argumentTypes = [OrdCltn new];
	    for(i=1;i<[aNode size];i+=2) {
		[argumentNames add:[aNode at:i]];
		[argumentTypes add:[Identifier str:"id"]];
	    }
	}
	
	return self;
}

- setDecls:dNode body:bNode
{
	id lToken = [Token str:"{"];
	id rToken = [Token str:"}"];
	
	/* wrap statement in compound: handle inlinecaches and tempvars */
	compoundStatement = 
	[CompoundStatement lbrace:lToken decls:dNode body:bNode rbrace:rToken];

	return self;
}

- setStatements:aNode
{
    return [self setDecls:nil body:aNode];
}

- setExpression:aNode
{
    /* stes 1/98 dirty C cast so that sortBy:{:x:y|strcmp([x str],[y str])}
     * works ...
     */
    id rNode = [Identifier str:"return (id)("];
    id sNode = [Token str:");"];

    /* wrap return expression in compound */
    aNode = [[[[Node new] add:rNode] add:aNode] add:sNode];

    return [self setStatements:aNode];
}

- (int) numArgsOfBlock
{
    if (argumentNames) {
        return [argumentNames size];
    } else {
	return 0;
    }
}

- addTempVarsTo:aStmt
{
        
	return self;
}

- addBlocksTo:aStmt
{
        [aStmt addBlock:self];
	/* do not descend, ie. don't get subblocks */
	return self;
}

- addMyBlocksTo:aStmt
{
        [aStmt addMyBlock:self];
	/* do not descend, ie. don't get subblocks */
	return self;
}

- addInlineCachesTo:aStmt
{
	return self;
}

- addScope:aScope
{
    if (!scope) scope = [Collection new];
    [scope addIfAbsent:aScope];  /* no dups in -isSame sense */
    return self;
}

- outputBlockDataTypeOn:aStream
{
    int n = [scope size];

    /* some compilers warn about empty structs */

    if (n) {
	int i;
	STR type = [[self blockDataTypeName] str];

	[aStream echo:"%s {\n",type];
	for(i=0;i<n;i++) {
	    STR t,d;id s = [scope at:i];
	    t = [[s scopeDataTypeName] str];
	    d = [[s scopeDataName] str];
	    [aStream echo:"%s *%s;\n",t,d];
	}
	[aStream echo:"};\n"];
	[aStream newline];
    }

    return self;
}

- packBlockDataOn:aStream
{
    /* wrap a blockData around the scope pointers */
    /* this is output-ing a comma expression */

    int n = [scope size];

    if (n) {
	int i;
	STR t = [[self blockDataTypeName] str];
	STR d = [[self blockDataName] str];

	[aStream echo:"%s=(%s *)OC_Malloc(sizeof(%s)),\n",d,t,t];

	for(i=0;i<n;i++) {
	    STR ds,ts;
	    id s = [scope at:i];
	    ds = [[s scopeDataName] str];
	    ts = [[s scopeDataTypeName] str];
	    [aStream echo:"%s->%s=%s,",d,ds,ds];
	}
    }

    return self;
}

- unpackBlockDataOn:aStream
{
    /* sort of opposite of packBlockData */

    int n = [scope size];

    if (n) {
	int i;
	for(i=0;i<n;i++) {
	    STR ds,ts;
	    id s = [scope at:i];
	    ts = [[s scopeDataTypeName] str];
	    ds = [[s scopeDataName] str];
	    [aStream echo:"%s *%s= data->%s;\n",ts,ds,ds];
	}
    }

    return self;
}

- processBlocks
{
    /* now copy what's needed of the enclosing scope */
    /* this depends on the pushScope/popScope mechanism of compounds */

    [self addArgumentsAsLocalsTo:compoundStatement];
    [compoundStatement addReferencedScopeToBlock:self];	

    /* descend for nested blocks */
    [Identifier pushScope:compoundStatement];
    [super processBlocks];
    [Identifier popScope];

    return self;
}

- pushScope
{
    int i,n = [scope size];
    for(i=0;i<n;i++) [Identifier pushScope:[scope at:i]];
    return self;
}

- popScope
{
    int i,n = [scope size];
    for(i=0;i<n;i++) [Identifier popScope];
    return self;
}

- addArgumentsAsLocalsTo:compound
{
    [compound addLocalVars:argumentNames types:argumentTypes];
    return self;
}

- outputBlockDefsOn:aStream
{
	STR tName;
        STR fnName = [[self blockFunctionName] str];

	/* first descend into body for nested blocks */
	[compoundStatement outputBlockDefsOn:aStream];

	[self outputBlockDataTypeOn:aStream];
	[aStream newline];

	/* tName is "void" if there is no data */
	if ([scope size]) {
	    tName = [[self blockDataTypeName] str];
	} else {
	    tName = "void";
	}

	[aStream echo:"static id %s(id thisBlock,%s *data",fnName,tName];

	if (argumentNames) {
	    int i;
	    for(i=0;i<[argumentNames size];i++) {
		[aStream echo:","];
		[[argumentTypes at:i] outputOn:aStream];
		[aStream echo:" "];
		[[argumentNames at:i] outputWithoutTagsOn:aStream];
	    }
	}

	[aStream echo:")\n"];
	[aStream echo:"{\n"];

	[self unpackBlockDataOn:aStream];

	[self pushScope];

	[aStream echo:"if (1)\n"];

	currentBlock = self;

	[LineTag forceLineTag];
	[compoundStatement translateOn:aStream];

	currentBlock = nil;

	[self popScope];

	/* return by default - if not yet returned - the block itself */
	[aStream newline];
	[aStream echo:"return thisBlock;\n"];
	[aStream echo:"}\n"];
	return self;
}

- translateOn:aStream
{
	if ([options blocks]) {
		int n = [self numArgsOfBlock];
		STR data,f = [[self blockFunctionName] str];
		
		/* start comma expression */
		[aStream echo:"("];

		if ([scope size]) {
		    data = [[self blockDataName] str];
		    [self packBlockDataOn:aStream];
		} else {
		    data = "(void*)0";
		}

		[aStream echo:"newBlock(%i,(IMP)%s,%s))",n,f,data];
		[LineTag forceLineTag];

		/* this is for automatic initialization */
		[Module usesEntry:[String str:"newBlock"]];
	} else {
		fatalError("blocks are not allowed with -noBlocks");
		[aStream echo:"(id)0"];
	}
	
	return self;
}

@end

