
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 */

#include "oclib.h"
#include "assert.h"
#include "string.h" /* strtok */
#include "ctype.h" /* isalnum */

@implementation Module

static void externC(id aStream)
{
	/* stes - may 16, 1997 */
	if ([options cplus]) [aStream echo:"extern \"C\" "];
}

/*
 * definition of the Objective C module struct
 */

static char *modType = "\n"
	"struct modDescriptor {\n"
/* keep this compatible with objcc */
	"\tchar  *modName;\n"
	"\tchar  *modVersion;\n"
	"\tlong  modStatus;\n"
	"\tchar  *modMinSel;\n"
	"\tchar  *modMaxSel;\n"
	"\tid    *modClsLst;\n"
	"\tshort modSelRef;\n"
	"\tchar  **modSelTbl;\n"
	"\tstruct methodDescriptor *modMapTbl;\n"
/* extensions */
	"};\n"
;

/*
 * Inline Cache
 * datatype must match objcrt.m !!
 */

+ outputInlineCacheMessengerOn:aStream
{
    [aStream echo:"\n"];

    /* datatype must match objcrt.m !! */
    if ([options cplus]) {
    [aStream echo:"struct objcrt_inlineCache {id cls;id (*imp)(...);};\n"];
    } else {
    [aStream echo:"struct objcrt_inlineCache {id cls;id (*imp)();};\n"];
    }

    /* messages to nil with inlineCache */
    externC(aStream);
    [aStream echo:"id %s _nilHandler(id,char*);\n",[[options bind] str]];
    
    return self;
}

+ outputBindCheckOn:aStream
{
	[aStream echo:"\nextern char *objcrt_bindError(char *);\n"];
	return self;
}

+ outputPrologueOn:aStream
{
        STR e = [[options bind] str];
	STR bName = [[self objcbindName] str];

	if ([options comments]) [aStream echo:"\n/* objc prologue */\n\n"];

	/* Note that "id" is not defined in header files
	* but is actually a token for the compiler
	* allows for compatibility with stepstone and GNU
	* where "id" is also built in, and our def would clash
	*/

	/* Change for Microsoft C : don't define "id" as a pointer to
	* 'struct _PRIVATE' (MSC doesn't do late defined structs)
	* We could define id as void* but let's use a generic def that
	* doesn't hide the "isa" pointer
	* At the same time, this requires a change for generation of
	* instance variable access (self->aVar) needs a struct _PRIVATE
	* cast, now.  This breaks something like anObject->isa, but that's
	* less important IMHO anyhow.
	*/

	[aStream echo:"struct objcrt_id { struct objcrt_id *isa; };\n"];
	[aStream echo:"typedef struct objcrt_id *id;\n"];

	/* define C messenger */
	/* stes - apr 13, 97 - changed to prototyped definition */

#if 0
        /* -kr */
	[aStream echo:"id (* _imp())(), (* _impSuper())();\n"];
#endif
	if ([options cplus]) {
	    externC(aStream); /* in C++ IMP is defined as id *(...) */
	    [aStream echo:"id %s (* _imp(id,char*))(...);\n",e];
	    externC(aStream);
	    [aStream echo:"id %s (* _impSuper(id,char*))(...);\n",e];
	} else {
	    [aStream echo:"extern id %s (* _imp(id,char*))();\n",e];
	    [aStream echo:"extern id %s (* _impSuper(id,char*))();\n",e];
	}

	/* define inline cache C messenger */

	if ([options inlineCache]) [self outputInlineCacheMessengerOn:aStream];	

	/* Table of pointers into __msgSelectors.
	* This table is modified by objcInit
	* is referenced in the source so needs prototype
	*
	* stes 22/7/1997 - force a reference to the OBJCBIND function.
	* This is to prevent some linkers (Metrowerks !) to
	* optimize BIND functions away when the module has no class definition
	* and only message expressions (which refer to selTransTbl).
	*
	* stes 2/1/98 - lcc4.0 was otherwise emitting a warning for this
	* assignment, so introduced -refBind to only do this for mwcc
	*/

	[aStream newline];
	[aStream echo:"extern struct modDescriptor %s *%s(void);\n",e, bName];
	if ([options refBind]) {
	[aStream echo:"static char **selTransTbl = (char **)%s;\n",bName];
	} else {
	[aStream echo:"static char **selTransTbl;\n"];
	}

	/* struct used in sharedType & defined by Stepstone objcc */
	if ([options cplus]) {
	[aStream echo:"\nstruct _SLT {char *_cmd;id (*_imp)(...);};\n"];
	} else {
	[aStream echo:"\nstruct _SLT {char *_cmd;id (*_imp)();};\n"];
	}

	/* Generate a type for Objective C modules 
	* make a forward define, for class impl
	* Microsoft C NEEDS this to come before sharedType (since
	* shared Type contains a struct modDescriptor pointer).
	*/
	[aStream echo:modType];
	[aStream echo:"\nextern struct modDescriptor %s;\n",
	[[self moduleDescName] str]];

	/* The BIND function initializes the selTransTbl.  If it somehow
	* does not get initialized we can (at the cost of some test)
	* issue an error message.
	*/
	if ([options checkBindCalled]) [self outputBindCheckOn:aStream];

	[LineTag forceLineTag];
	return self;
}

static id classImplementation;
static id modClsLstCltn; /* if more than one allowed per file */

+ setTheOnlyClassImplementation:classImpl
{
	assert([options onePerFile]);
	if (classImplementation) {
		fatalError("only one implementation per file allowed");
	}
	classImplementation = classImpl;
	return self;
}

+ addClassImplementation:classImpl
{
	assert(![options onePerFile]);
	if (!modClsLstCltn) modClsLstCltn = [Collection new];
	[modClsLstCltn add:classImpl];
	return self;
}

static id selTransTblCltn;

+ (int) selTransTblOffsetFor:methodName
{
	id aString;
	if (!selTransTblCltn) selTransTblCltn = [Collection new];

	aString = [selTransTblCltn findMatching:methodName];

	if (aString) {
		return [selTransTblCltn offsetOf:aString];
	} else {
		[selTransTblCltn add:methodName];
		return [selTransTblCltn offsetOf:methodName];
	}
}

/*****************************************************************************
 *
 * This is a (slightly) incompatible extension that allows us to have more
 * than one class implementation per file.
 *
 ****************************************************************************/

+ outputModClsLstOn:aStream
{
	int i,n;
	
	n = (modClsLstCltn)?[modClsLstCltn size]:0;

	[aStream newline];
	[aStream echo:"static id _modClsLst[] ={\n"];

	for(i=0;i<n;i++) {
		STR s = [[[modClsLstCltn at:i] idName] str];
		/* be careful : &_Foo is generated here */
		/* that is exactly the value of "id Foo := (id)&_Foo" */
		[aStream echo:"(id)&_%s,\n",s];
	}

	/* must be NULL terminated */
	[aStream echo:"(id)0};\n"];

	return self;
}

/***************************************************************************
 *
 * static char *_selTransTbl[] ={
 * "str:",
 * "printOn:",
 * 0};
 *
 **************************************************************************/

+ outputSelTransTblOn:aStream
{
	int i,n;

	n = (selTransTblCltn)?[selTransTblCltn size]:0;

	[aStream newline];
	[aStream echo:"static char *_selTransTbl[] ={\n"];

	for(i=0;i<n;i++) {
		STR s = [[selTransTblCltn at:i] str];
		[aStream echo:"\"%s\",\n",s];
	}
	
	/* always at least one entry (SGI cc doesn't like empty decls)    */
	[aStream echo:"0\n};\n"];

	return self;
}

/*****************************************************************************
 *
 * static struct modDescriptor _modDesc = {
 *         "testdir/test.m",
 *         "objc1.8.7",
 *
 ****************************************************************************/

#define MOD_MORETHANONE 0x4L /* must match objcrt.m value !! */

+ outputModDescriptorOn:aStream
{
	STR modVersion = "objc1.8.7";
	int selSize = (selTransTblCltn)?[selTransTblCltn size]:0;
	STR modName = [[options srcFileName] str];
	STR modDescName = [[self moduleDescName] str];
	long modStatus = ([options onePerFile])?0:MOD_MORETHANONE;
	
	[aStream newline];

#if 0
	[aStream echo:"static struct modDescriptor _modDesc = {\n"];
#endif
	/* removed static for problem with C++ forward "static" declarations */
	[aStream echo:"struct modDescriptor %s = {\n",modDescName];
	[aStream echo:"\t\"%s\",\n",modName];
	[aStream echo:"\t\"%s\",\n",modVersion];
	[aStream echo:"\t%iL,\n",modStatus]; /* modStatus */

/* min and max Sel are obsoleted, it seems */
	[aStream echo:"\t0,\n"]; /* modMinSel */
	[aStream echo:"\t0,\n"]; /* modMaxSel */
/*
	if (selSize) {
		[aStream echo:"\t_selTransTbl[0],\n"];
		[aStream echo:"\t_selTransTbl[%i],\n",selSize - 1];
	} else {
 */

	if ([options onePerFile]) {
		/* this is the pre 1.4.x code with "one class per file" */
		if (classImplementation) {
			STR cName = [[classImplementation idName] str];
			[aStream echo:"\t&%s,\n",cName]; /* modClsLst */
		} else {
			[aStream echo:"\t0,\n"]; /* modClsLst */
		}
	} else {
		/* default since 1.4.x is to have more than one per file */
		[aStream echo:"\t_modClsLst,\n"]; /* modClsLst */
	}

	if (selSize) {
		[aStream echo:"\t%i,\n",selSize]; /* modSelRef */
		[aStream echo:"\t_selTransTbl,\n"]; /* modSelTbl */
	} else {
		[aStream echo:"\t0,\n"]; /* modSelRef */
		[aStream echo:"\t0,\n"]; /* modSelTbl */
	}

	/* we don't support static references _mapTbl */
	/* this is just here for Stepstone compatibility */
	[aStream echo:"\t0\n};\n"]; /* modMapTbl */

	return self;
}

/*****************************************************************************
 *
 * struct modDescriptor * _BIND_file()
 * 	{ return &_modDesc; }
 *
 ****************************************************************************/

/* This function is currently a UNIX-ism ...
 * Replaced strrchr (which seems non-standard on SunOS4.1.1)
 * Maybe should move this to String object.
 */

static char *strRChr(const char *s,int c)
{
	char *t = (char *)s;while (*t) t++;
	while (t != s) if (*t == c) return t;else t--;
	return (*t == c)?t:NULL;
}

static id BindNameFor(const char *filename)
{
	id s;
	char *p,*pathSep = [options pathSeparator];

	p = strRChr(filename,pathSep[0]);
	s = [String str:(p)?p+1:(char *)filename];
	p = [s strCopy]; /* get malloc'ed String buffer */
/* Original (doesnt work for foo-bar.m filename) :
	strtok(p,".");
   marcel@djo.nl replaced by block below: */
	{
	char *cp;
	if ((cp = strRChr(p, '.')) != NULL) *cp = '\0'; /* strip extension */
	for (cp = p; *cp != '\0'; cp++) if (!isalnum(*cp)) *cp = '_';
	}
	return [String str:p];
}

+ bindName
{
	STR s;
	static id name;

	if (name) return name;

	s = [[options srcFileName] str];
	if (!s) s = [[options inFile] str];
	if (!s) s = "stdin";
	return name = BindNameFor(s);
}

+ objcbindName
{
	return [[String str:"_OBJCBIND_"] concatSTR:[[self bindName] str]];
}

+ moduleDescName
{
	/* before cplus support we just used to call this variable "_modDesc"
	 * and made it static to each module.  static initializers don't
	 * seem to work in C++ as in C, and because we need a reference
	 * from _modDesc to class impl and from class impl to _modDesc, we just
	 * make this variable have a unique global name.
	 */
	 
	id s = [self bindName];
	return [String sprintf:"%s_modDesc",[s str]];
}

+ allClasses
{
    if ([options onePerFile]) {
	return (classImplementation)?
	    [[Collection new] add:classImplementation]:nil;
    } else {
	return modClsLstCltn;
    }

    return nil;
}


+ outputClassGlobalFunCallsOn:aStream
{
    int i,n;
    id all = [self allClasses];

    n = (all)?[all size]:0;
    for(i=0;i<n;i++) [[all at:i] outputClassGlobalFunCallsOn:aStream];

    return self;
}

+ outputEntryPointOn:aStream
{
	id bName = [self objcbindName];
	/* int selSize = [selTransTblCltn size]; */

	[aStream newline];
	
	[aStream echo:"struct modDescriptor "];

	/* for DLL's */
	[aStream echo:[[options bind] str]];

	[aStream echo:" *%s(void)\n{\n",[bName str]];
	/*if(selSize)*/ [aStream echo:"\tselTransTbl = _selTransTbl;\n"];

	/* -noShared case, where we can't have global data shared across
	 * modules
	 */

	if ([options sharedData] == NO)  {
	    [self outputClassGlobalFunCallsOn:aStream];
	}

	[aStream echo:"\treturn &%s;\n}\n",[[self moduleDescName] str]];

	return self;
}

/*****************************************************************************
 *
 * struct useDescriptor {
 *	int processed;
 *	struct useDescriptor *next;
 *	struct useDescriptor ***uses;
 *	struct modDescriptor *(*bind)();
 * };
 *
 ****************************************************************************/

static id usesEntries,definesEntries,staticEntries;

+ staticEntry:aString
{
	if (!staticEntries) staticEntries = [Set new];
	if ([options debugInfo]) printf("static use %s\n",[aString str]);
	assert([aString isKindOf:[String class]]);
	[staticEntries add:aString];
	return self;
}

+ usesEntry:aString
{
	if (!usesEntries) usesEntries = [Set new];
	if ([options debugInfo]) printf("use %s\n",[aString str]);
	assert([aString isKindOf:[String class]]);
	if (![staticEntries contains:aString]) [usesEntries add:aString];
	return self;
}

+ definesEntry:aString
{
	if (!definesEntries) definesEntries = [Set new];
	if ([options debugInfo]) printf("define %s\n",[aString str]);
	assert([aString isKindOf:[String class]]);
	if (![staticEntries contains:aString]) [definesEntries add:aString];
	return self;
}

+ (BOOL) definesMain
{
	return [definesEntries contains:[options mainFunction]];
}

+ outputUsesOn:aStream
{
	id aSeq,entry;

	if ([options comments]) [aStream echo:"\n/* uses */\n"];
	
	aSeq = [usesEntries eachElement];
	while ((entry = [aSeq next])) {
		STR s = [entry str];

		/* we do NOT define the "uses" entries as "extern"
                 *
                 * this is a somewhat less understood feature of K&R C 
		 * and most ANSI C compilers also implement it (under
		 * the name of 'relaxed ANSI' or so)
		 *
		 * other compilers, such as the SGI IRIX, have a flag
		 * for it (such as -common).
		 *
		 * the idea is that uninitialized globals are supposed
		 * to be placed into the 'COMMON' segment per tranlation
		 * unit, instead of immediately placing the symbol in the
		 * BSS segment.  then when linking, multiple COMMON
		 * symbols are allowed (and merged) and if the symbol's
		 * also in the DATA segment, it gets a value.  Otherwise
		 * it goes into the BSS segment (i.e. it will be 0).
                 *
                 * we define the entry in multiple modules
                 * (e.g. OCU_MyObject) and to use the definition that has
                 * an initializer (that is the "defines" entry).
                 *
                 * for some functions (e.g. OCU_printf) there will be no
                 * define, but we will not get an unresolved symbol...
                 *
                 * it is necessary to generate OCU_funcs because we might
                 * have a module, to which our only entry point is a funccall,
                 * and that itself uses Objective C classes that need init...
                 */

		[aStream echo:"struct useDescriptor *OCU_%s;\n",s];
	}
	aSeq = [aSeq free];

	return self;
}

+ outputUseControlOn:aStream
{
	id aSeq,entry;
	id bName = [self objcbindName];

	[aStream newline];
	[aStream echo:"static struct useDescriptor **_useControl[] = {\n"];
	
	aSeq = [usesEntries eachElement];
	while ((entry = [aSeq next])) {
		STR s = [entry str];
		[aStream echo: "&OCU_%s,\n",s];
	}
	aSeq = [aSeq free];

	[aStream echo:"0};\n"];

	[aStream newline];
	
	[aStream echo:"static struct useDescriptor _useDesc = {\n"];
	[aStream echo:"\t0,\n"];
	[aStream echo:"\t0,\n"];
	[aStream echo:"\t_useControl"];
	
	/* stes apr/22/1997 - removed if (selSize) test for case where
	 * _OBJCBIND function was wrongly set to 0 ("super" example in testdir)
	 * we can always use the bind function (since it's defined anyway) and
	 * does itself a if (selSize) test.
	 */
	
	[aStream echo:",\n\t%s",[bName str]];

	[aStream echo:"\n};\n"];

	return self;
}

+ outputDefinesOn:aStream
{
	id aSeq,entry;
	id mf = [options mainFunction];

	if ([options comments]) [aStream echo:"\n/* defines */\n"];
	aSeq = [definesEntries eachElement];
	while ((entry = [aSeq next])) {
		STR f,s = [entry str];
		f = "struct useDescriptor *OCU_%s = &_useDesc;\n";
		[aStream echo:f,s];
	}
	aSeq = [aSeq free];

	/* objcrt references OCU_main but our main function can be OCU_bar */

	if ([self definesMain] && (![mf isEqualSTR:"main"])){
	    [aStream echo:"\nstruct useDescriptor *OCU_main=&_useDesc;\n"];
	}

	return self;
}

+ outputUseDescriptorOn:aStream
{
	[aStream newline];

	[aStream echo:"struct useDescriptor {\n"];
	[aStream echo:"\tint processed;\n"];
	[aStream echo:"\tstruct useDescriptor *next;\n"];
	[aStream echo:"\tstruct useDescriptor ***uses;\n"];
	[aStream echo:"\tstruct modDescriptor *(*bind)();\n"];
	[aStream echo:"};\n"];

	[self outputUsesOn:aStream];
	[self outputUseControlOn:aStream];
	[self outputDefinesOn:aStream];

	/* turn on auto-initialization code by defining the postlink
	 * entry point as zero.
	 * objcrt initialization has an extern to this symbol 
         */

	if ([self definesMain]) {
	    [aStream echo:"\nstruct modEntry *_objcModules = 0;\n"];
	}

	return self;
}

+ outputPostLinkMarkOn:aStream
{
	id bName = [self bindName];

	/* the postLink version doesn't need the OCU entries
	 * however, we want to 'mark' the .o file as being
	 * compiled with -postlink, so that when linking
	 * against a library that was compiled with -postlink
	 * we could automatically detect this (and use postlink).
	 */

	[aStream newline];
	[aStream echo:"int _OBJCPOSTLINK_%s = 1;\n",[bName str]];

	/* turn off auto-initialization code
	 * objcrt auto-initialization has an extern to this symbol
	*/

	if ([self definesMain]) {
	    [aStream echo:"\nstruct useDescriptor *OCU_main = 0;\n"];
	}

	return self;
}

+ outputEpilogueOn:aStream
{
	if ([options comments]) [aStream echo:"\n\n/* objc epilogue */\n"];

	[self outputSelTransTblOn:aStream];
	if (![options onePerFile]) [self outputModClsLstOn:aStream];
	[self outputModDescriptorOn:aStream];
	[self outputEntryPointOn:aStream];

	if ([options postLink]) {
		[self outputPostLinkMarkOn:aStream];
	} else {
		[self outputUseDescriptorOn:aStream];
	}

	return self;
}

@end

